@@ -277,15 +277,53 @@ bool parse_value(type_t type, const char *str, parsed_value_t *value)
       while (map[(uint8_t)*str] == INT_MAX && isspace_iso88591(*str))
          ++str;
 
-      const bool quoted = map['\"'] == INT_MAX && *str == '\"';
-      if (quoted) str++;
+      int base = 0, dbits = 1;
+      bool quoted = false;
+      if (map['\"'] == INT_MAX) {
+         if (str[0] == '\"') {
+            quoted = true;
+            str++;
+         }
+         else if (str[1] == '\"' && toupper_iso88591(str[0]) == 'X') {
+            quoted = true;
+            base = 16;
+            dbits = 4;
+            str += 2;
+         }
+      }
 
-      const size_t max = strlen(str);
+      const size_t max = strlen(str) * dbits;
       enum_array_t *array = xmalloc_flex(sizeof(enum_array_t), max, 1);
 
       int n = 0, m;
-      for (; *str != '\0' && (m = map[(uint8_t)*str]) != INT_MAX; str++, n++)
-         array->values[n] = m;
+      for (; *str != '\0'; str++, n += dbits) {
+         if (base > 0) {
+            const bool extended = (isdigit_iso88591(*str) && *str < '0' + base)
+               || (base > 10 && *str >= 'A' && *str < 'A' + base - 10)
+               || (base > 10 && *str >= 'a' && *str < 'a' + base - 10);
+
+            if (!extended)
+               break;
+
+            int digit = isdigit_iso88591(*str)
+               ? (*str - '0')
+               : 10 + (toupper_iso88591(*str) - 'A');
+
+            bool valid = true;
+            for (int i = dbits - 1; i >= 0; i--) {
+               const uint8_t bit = (digit >> i) & 1 ? '1' : '0';
+               valid &= (m = map[bit]) != INT_MAX;
+               array->values[n + dbits - i - 1] = m;
+            }
+
+            if (!valid)
+               break;
+         }
+         else if ((m = map[(uint8_t)*str]) == INT_MAX)
+            break;
+         else
+            array->values[n] = m;
+      }
 
       assert(n <= max);
       array->count = n;
@@ -695,6 +733,7 @@ bool is_concurrent_block(tree_t t)
    case T_BLOCK:
    case T_IF_GENERATE:
    case T_FOR_GENERATE:
+   case T_CASE_GENERATE:
       return true;
    default:
       return false;
@@ -932,6 +971,7 @@ unsigned dimension_of(type_t type)
    case T_RECORD:
    case T_INCOMPLETE:
    case T_FILE:
+   case T_SIGNATURE:
       return 0;
    case T_INTEGER:
    case T_REAL:
@@ -1104,7 +1144,6 @@ void intern_strings(void)
    id_cache[W_NEVER_WAITS]     = ident_new("NEVER_WAITS");
    id_cache[W_NVC_VERILOG]     = ident_new("NVC.VERILOG");
    id_cache[W_NVC_PSL_SUPPORT] = ident_new("NVC.PSL_SUPPORT");
-   id_cache[W_SHAPE]           = ident_new("shape");
    id_cache[W_INSTANCE_NAME]   = ident_new("instance_name");
    id_cache[W_PATH_NAME]       = ident_new("path_name");
    id_cache[W_VITAL]           = ident_new("VITAL");
@@ -1304,13 +1343,15 @@ type_t std_type(tree_t std, std_type_t which)
 
 type_t ieee_type(ieee_type_t which)
 {
-   static type_t cache[IEEE_STD_LOGIC + 1] = {};
+   static type_t cache[IEEE_STD_LOGIC_VECTOR + 1] = {};
    assert(which < ARRAY_LEN(cache));
 
    if (cache[which] == NULL) {
-      const char *names[] = {
-         "STD_ULOGIC",
-         "STD_LOGIC",
+      static const char *const names[] = {
+         [IEEE_STD_ULOGIC] = "STD_ULOGIC",
+         [IEEE_STD_LOGIC] = "STD_LOGIC",
+         [IEEE_STD_ULOGIC_VECTOR] = "STD_ULOGIC_VECTOR",
+         [IEEE_STD_LOGIC_VECTOR] = "STD_LOGIC_VECTOR",
       };
 
       static tree_t ieee_cache[STD_19 + 1] = {};
@@ -1341,7 +1382,7 @@ type_t verilog_type(verilog_type_t which)
    assert(which < ARRAY_LEN(cache));
 
    if (cache[which] == NULL) {
-      const char *names[] = {
+      static const char *const names[] = {
          [VERILOG_LOGIC] = "T_LOGIC",
          [VERILOG_LOGIC_ARRAY] = "T_LOGIC_ARRAY",
          [VERILOG_INT64] = "T_INT64",
@@ -1370,9 +1411,9 @@ type_t reflection_type(reflect_type_t which)
    assert(which < ARRAY_LEN(cache));
 
    if (cache[which] == NULL) {
-      const char *names[] = {
-         "VALUE_MIRROR",
-         "SUBTYPE_MIRROR",
+      static const char *const names[] = {
+         [REFLECT_VALUE_MIRROR] = "VALUE_MIRROR",
+         [REFLECT_SUBTYPE_MIRROR] = "SUBTYPE_MIRROR",
       };
 
       static tree_t reflect_cache[STD_19 + 1] = {};
@@ -1802,19 +1843,18 @@ static bool is_static(tree_t expr)
          case ATTR_LOW:
          case ATTR_HIGH:
             {
-               type_t type = tree_type(tree_name(expr));
-               if (type_is_unconstrained(type))
+               tree_t ref = name_to_ref(tree_name(expr));
+               if (ref == NULL)
                   return false;
 
-               int64_t dim = 1;
-               if (tree_params(expr) == 1)
-                  dim = assume_int(tree_value(tree_param(expr, 0)));
-
-               tree_t r = range_of(type, dim - 1);
-               if (tree_subkind(r) == RANGE_EXPR)
+               switch (tree_kind(tree_ref(ref))) {
+               case T_GENERIC_DECL:
+               case T_PORT_DECL:
+               case T_SIGNAL_DECL:
+                  return true;
+               default:
                   return false;
-
-               return is_static(tree_left(r)) && is_static(tree_right(r));
+               }
             }
          default:
             return true;
@@ -2680,23 +2720,6 @@ bool same_tree(tree_t a, tree_t b)
    }
 }
 
-void instance_name_to_path(text_buf_t *tb, const char *str)
-{
-   bool delete = false;
-   for (const char *p = str; *p; p++) {
-      if (*p == '@' || (*p == '(' && !isdigit_iso88591(*(p + 1))))
-         delete = true;
-      else if (*p == ')' && delete)
-         delete = false;
-      else if (*p == ':') {
-         delete = false;
-         tb_append(tb, ':');
-      }
-      else if (!delete)
-         tb_append(tb, *p);
-   }
-}
-
 bool calculate_aggregate_bounds(tree_t expr, range_kind_t *kind,
                                 int64_t *left, int64_t *right)
 {
