@@ -993,14 +993,13 @@ cover_scope_t *cover_create_instance(cover_data_t *data, cover_scope_t *parent,
       parent->name = parent->hier = lib_name(lib_work());
    }
 
-   // TODO: do not emit scopes for components
-   assert(tree_kind(unit) == T_ARCH);
+   assert(is_concurrent_block(unit));
    assert(tree_kind(block) == T_BLOCK);
 
    cover_scope_t *s = xcalloc(sizeof(cover_scope_t));
    s->name       = tree_ident(block);
    s->parent     = parent;
-   s->loc        = *tree_loc(block);
+   s->loc        = *tree_loc(unit);
    s->hier       = ident_prefix(s->parent->hier, s->name, '.');
    s->emit       = cover_should_emit_scope(data, s, block);
    s->block_name = ident_rfrom(tree_ident(unit), '.');
@@ -1132,7 +1131,8 @@ cover_data_t *cover_read_items(fbuf_t *f, uint32_t pre_mask)
    return data;
 }
 
-static void cover_merge_scope(cover_scope_t *old_s, cover_scope_t *new_s)
+static void cover_merge_scope(cover_scope_t *old_s, cover_scope_t *new_s,
+                              merge_mode_t mode)
 {
    // Most merged cover scopes have equal items.
    // Start from equal item index in old and new scope instead of iterating
@@ -1169,7 +1169,7 @@ static void cover_merge_scope(cover_scope_t *old_s, cover_scope_t *new_s)
          n_old_visits++;
       } while (true);
 
-      if (!found) {
+      if (!found && mode == MERGE_UNION) {
          APUSH(old_s->items, *new);
          n_added++;
       }
@@ -1181,18 +1181,18 @@ static void cover_merge_scope(cover_scope_t *old_s, cover_scope_t *new_s)
       for (int j = 0; j < old_s->children.count; j++) {
          cover_scope_t *old_c = old_s->children.items[j];
          if (new_c->name == old_c->name) {
-            cover_merge_scope(old_c, new_c);
+            cover_merge_scope(old_c, new_c, mode);
             found = true;
             break;
          }
       }
 
-      if (!found)
+      if (!found && mode == MERGE_UNION)
          APUSH(old_s->children, new_c);
    }
 }
 
-void cover_merge_items(fbuf_t *f, cover_data_t *data)
+void cover_merge_items(fbuf_t *f, cover_data_t *data, merge_mode_t mode)
 {
    assert (data != NULL);
 
@@ -1208,7 +1208,7 @@ void cover_merge_items(fbuf_t *f, cover_data_t *data)
       case CTRL_PUSH_SCOPE:
          {
             cover_scope_t *new = cover_read_scope(f, ident_ctx, loc_rd);
-            cover_merge_scope(data->root_scope, new);
+            cover_merge_scope(data->root_scope, new, mode);
          }
          break;
       case CTRL_END_OF_FILE:
