@@ -50,6 +50,7 @@ typedef A(tree_t) tree_list_t;
 
 typedef struct _elab_ctx elab_ctx_t;
 typedef struct _generic_list generic_list_t;
+typedef struct _elab_instance elab_instance_t;
 
 typedef struct _elab_ctx {
    const elab_ctx_t *parent;
@@ -57,7 +58,6 @@ typedef struct _elab_ctx {
    object_t         *root;
    tree_t            inst;
    tree_t            config;
-   ident_t           inst_name;     // Current 'INSTANCE_NAME
    ident_t           dotted;
    ident_t           prefix[2];
    lib_t             library;
@@ -93,51 +93,29 @@ typedef struct _generic_list {
 } generic_list_t;
 
 typedef struct {
-   vcode_unit_t shape;
-   tree_t       block;
-   tree_t       wrap;
-   vlog_node_t  module;
+   tree_t           wrap;
+   tree_t           entity;
+   vlog_node_t      module;
+   elab_instance_t *instances;
 } mod_cache_t;
 
+typedef struct _elab_instance {
+   elab_instance_t *next;
+   mod_cache_t     *module;
+   vlog_node_t      body;
+   tree_t           block;
+   tree_t           wrap;
+} elab_instance_t;
+
 static void elab_block(tree_t t, const elab_ctx_t *ctx);
 static void elab_stmts(tree_t t, const elab_ctx_t *ctx);
 static void elab_decls(tree_t t, const elab_ctx_t *ctx);
 static void elab_push_scope(tree_t t, elab_ctx_t *ctx);
 static void elab_pop_scope(elab_ctx_t *ctx);
+static void elab_verilog_stmts(vlog_node_t v, const elab_ctx_t *ctx);
 
 static generic_list_t *generic_override = NULL;
 
-static ident_t hpathf(ident_t path, char sep, const char *fmt, ...)
-{
-   va_list ap;
-   va_start(ap, fmt);
-   char *buf = xvasprintf(fmt, ap);
-   va_end(ap);
-
-   // LRM specifies instance path is lowercase
-   char *p = buf;
-   while (*p != '\0') {
-      *p = tolower_iso88591(*p);
-      ++p;
-   }
-
-   ident_t id = ident_new(buf);
-   free(buf);
-   return ident_prefix(path, id, sep);
-}
-
-static const char *simple_name(const char *full)
-{
-   // Strip off any library or entity prefix from the parameter
-   const char *start = full;
-   for (const char *p = full; *p != '\0'; p++) {
-      if (*p == '.' || *p == '-')
-         start = p + 1;
-   }
-
-   return start;
-}
-
 static lib_t elab_find_lib(ident_t name, const elab_ctx_t *ctx)
 {
    ident_t lib_name = ident_until(name, '.');
@@ -262,28 +240,49 @@ static mod_cache_t *elab_cached_module(vlog_node_t mod, const elab_ctx_t *ctx)
    assert(is_top_level(mod));
 
    mod_cache_t *mc = hash_get(ctx->modcache, mod);
-   if (mc == NULL) {
-      mc = xcalloc(sizeof(mod_cache_t));
-      mc->module = mod;
-      mc->shape  = vlog_lower(ctx->registry, ctx->mir, mod);
+   if (mc != NULL)
+      return mc;
 
-      mc->block = tree_new(T_BLOCK);
-      tree_set_loc(mc->block, vlog_loc(mod));
-      tree_set_ident(mc->block, vlog_ident(mod));
+   mc = xcalloc(sizeof(mod_cache_t));
+   mc->module = mod;
 
-      vlog_trans(mod, mc->block);
-
-      mc->wrap = tree_new(T_VERILOG);
-      tree_set_loc(mc->wrap, vlog_loc(mod));
-      tree_set_ident(mc->wrap, vlog_ident(mod));
-      tree_set_vlog(mc->wrap, mod);
-
-      hash_put(ctx->modcache, mod, mc);
-   }
+   mc->wrap = tree_new(T_VERILOG);
+   tree_set_loc(mc->wrap, vlog_loc(mod));
+   tree_set_ident(mc->wrap, vlog_ident(mod));
+   tree_set_vlog(mc->wrap, mod);
 
+   hash_put(ctx->modcache, mod, mc);
    return mc;
 }
 
+static elab_instance_t *elab_new_instance(vlog_node_t mod, vlog_node_t inst,
+                                          const elab_ctx_t *ctx)
+{
+   assert(is_top_level(mod));
+
+   mod_cache_t *mc = elab_cached_module(mod, ctx);
+
+   elab_instance_t *ei = xcalloc(sizeof(elab_instance_t));
+   ei->next   = mc->instances;
+   ei->module = mc;
+   ei->body   = vlog_new_instance(mod, inst, ctx->dotted);
+
+   ei->wrap = tree_new(T_VERILOG);
+   tree_set_loc(ei->wrap, vlog_loc(mod));
+   tree_set_ident(ei->wrap, vlog_ident(mod));
+   tree_set_vlog(ei->wrap, ei->body);
+
+   ei->block = tree_new(T_BLOCK);
+   tree_set_loc(ei->block, vlog_loc(inst));
+   tree_set_ident(ei->block, vlog_ident(inst));
+
+   vlog_trans(ei->body, ei->block);
+
+   mc->instances = ei;
+
+   return ei;
+}
+
 static bool elab_synth_binding_cb(lib_t lib, void *__ctx)
 {
    synth_binding_params_t *params = __ctx;
@@ -307,6 +306,10 @@ static tree_t elab_to_vhdl(type_t from, type_t to)
       { VERILOG_LOGIC, IEEE_STD_LOGIC, "NVC.VERILOG.TO_VHDL(" T_LOGIC ")U" },
       { VERILOG_NET_VALUE, IEEE_STD_LOGIC,
         "NVC.VERILOG.TO_VHDL(" T_NET_VALUE ")U" },
+      { VERILOG_WIRE_ARRAY, IEEE_STD_ULOGIC_VECTOR,
+        "NVC.VERILOG.TO_VHDL(" T_WIRE_ARRAY ")Y" },
+      { VERILOG_WIRE_ARRAY, IEEE_STD_LOGIC_VECTOR,
+        "NVC.VERILOG.TO_VHDL(" T_WIRE_ARRAY ")Y" },
    };
 
    INIT_ONCE({
@@ -337,7 +340,11 @@ static tree_t elab_to_verilog(type_t from, type_t to)
    } table[] = {
       { IEEE_STD_ULOGIC, VERILOG_LOGIC, "NVC.VERILOG.TO_VERILOG(U)" T_LOGIC },
       { IEEE_STD_ULOGIC, VERILOG_NET_VALUE,
-        "NVC.VERILOG.TO_VERILOG(U)" T_NET_VALUE }
+        "NVC.VERILOG.TO_VERILOG(U)" T_NET_VALUE },
+      { IEEE_STD_ULOGIC_VECTOR, VERILOG_WIRE_ARRAY,
+        "NVC.VERILOG.TO_VERILOG(Y)" T_NET_ARRAY },
+      { IEEE_STD_LOGIC_VECTOR, VERILOG_WIRE_ARRAY,
+        "NVC.VERILOG.TO_VERILOG(Y)" T_NET_ARRAY },
    };
 
    INIT_ONCE({
@@ -360,33 +367,43 @@ static tree_t elab_mixed_binding(tree_t comp, mod_cache_t *mc)
 {
    assert(tree_kind(comp) == T_COMPONENT);
 
+   if (mc->entity == NULL) {
+      mc->entity = tree_new(T_ENTITY);
+      tree_set_ident(mc->entity, vlog_ident(mc->module));
+      tree_set_loc(mc->entity, vlog_loc(mc->module));
+
+      vlog_trans(mc->module, mc->entity);
+   }
+
    tree_t bind = tree_new(T_BINDING);
    tree_set_ident(bind, vlog_ident(mc->module));
    tree_set_loc(bind, tree_loc(comp));
    tree_set_ref(bind, mc->wrap);
    tree_set_class(bind, C_ENTITY);
 
-   const int nports = tree_ports(comp);
-   const int ndecls = vlog_decls(mc->module);
+   const int vhdl_nports = tree_ports(comp);
+   const int vlog_nports = vlog_ports(mc->module);
 
    bit_mask_t have;
-   mask_init(&have, nports);
+   mask_init(&have, vhdl_nports);
 
    bool have_named = false;
-   for (int i = 0; i < ndecls; i++) {
-      vlog_node_t mport = vlog_decl(mc->module, i);
-      if (vlog_kind(mport) != V_PORT_DECL)
-         continue;
+   for (int i = 0; i < vlog_nports; i++) {
+      vlog_node_t ref = vlog_port(mc->module, i);
+      assert(vlog_kind(ref) == V_REF);
+
+      vlog_node_t mport = vlog_ref(ref);
+      assert(vlog_kind(mport) == V_PORT_DECL);
 
       ident_t name = vlog_ident2(mport);
 
-      tree_t vport = tree_port(mc->block, i);
+      tree_t vport = tree_port(mc->entity, i);
       assert(tree_ident(vport) == vlog_ident(mport));
 
       tree_t cport = NULL;
-      for (int j = 0; j < nports; j++) {
+      for (int j = 0; j < vhdl_nports; j++) {
          tree_t pj = tree_port(comp, j);
-         if (tree_ident(pj) == name) {
+         if (ident_casecmp(tree_ident(pj), name)) {
             cport = pj;
             mask_set(&have, j);
             break;
@@ -395,15 +412,7 @@ static tree_t elab_mixed_binding(tree_t comp, mod_cache_t *mc)
 
       if (cport == NULL) {
          error_at(tree_loc(comp), "missing matching VHDL port declaration for "
-                  "Verilog port %s in component %s", istr(vlog_ident(mport)),
-                  istr(tree_ident(comp)));
-         return NULL;
-      }
-
-      if (name != tree_ident(cport)) {
-         error_at(tree_loc(cport), "expected VHDL port name %s to match "
-                  "Verilog port name %s in component %s",
-                  istr(tree_ident(cport)), istr(vlog_ident(mport)),
+                  "Verilog port '%s' in component %s", istr(vlog_ident(mport)),
                   istr(tree_ident(comp)));
          return NULL;
       }
@@ -415,7 +424,7 @@ static tree_t elab_mixed_binding(tree_t comp, mod_cache_t *mc)
          tree_t func = elab_to_verilog(btype, vtype);
          if (func == NULL) {
             error_at(tree_loc(cport), "cannot connect VHDL signal with type "
-                     "%s to Verilog input port %s", type_pp(btype),
+                     "%s to Verilog input port '%s'", type_pp(btype),
                      istr(vlog_ident(mport)));
             return NULL;
          }
@@ -436,7 +445,7 @@ static tree_t elab_mixed_binding(tree_t comp, mod_cache_t *mc)
          tree_t func = elab_to_vhdl(vtype, btype);
          if (func == NULL) {
             error_at(tree_loc(cport), "cannot connect VHDL signal with type "
-                     "%s to Verilog output port %s", type_pp(btype),
+                     "%s to Verilog output port '%s'", type_pp(btype),
                      istr(vlog_ident(mport)));
             return NULL;
          }
@@ -453,7 +462,7 @@ static tree_t elab_mixed_binding(tree_t comp, mod_cache_t *mc)
       }
    }
 
-   for (int i = 0; i < nports; i++) {
+   for (int i = 0; i < vhdl_nports; i++) {
       if (!mask_test(&have, i)) {
          tree_t p = tree_port(comp, i);
          diag_t *d = diag_new(DIAG_ERROR, tree_loc(p));
@@ -465,6 +474,68 @@ static tree_t elab_mixed_binding(tree_t comp, mod_cache_t *mc)
 
    mask_free(&have);
 
+   const int vhdl_ngenerics = tree_generics(comp);
+   const int vlog_ndecls = vlog_decls(mc->module);
+
+   mask_init(&have, vhdl_ngenerics);
+
+   for (int i = 0, nth = 0; i < vlog_ndecls; i++) {
+      vlog_node_t d = vlog_decl(mc->module, i);
+      if (vlog_kind(d) != V_PARAM_DECL)
+         continue;
+
+      ident_t name = vlog_ident(d);
+
+      tree_t eg = tree_generic(mc->entity, nth++);
+      assert(tree_ident(eg) == name);
+
+      tree_t cg = NULL;
+      for (int j = 0; j < vhdl_ngenerics; j++) {
+         tree_t gj = tree_generic(comp, j);
+         if (ident_casecmp(tree_ident(gj), name)) {
+            cg = gj;
+            mask_set(&have, j);
+            break;
+         }
+      }
+
+      if (cg == NULL) {
+         error_at(tree_loc(comp), "missing matching VHDL generic declaration "
+                  "for Verilog parameter '%s' in component %s", istr(name),
+                  istr(tree_ident(comp)));
+         return NULL;
+      }
+
+      type_t ctype = tree_type(cg);
+      type_t etype = tree_type(eg);
+      assert(type_eq(etype, std_type(NULL, STD_INTEGER)));
+
+      if (!type_eq(ctype, etype)) {
+         error_at(tree_loc(cg), "generic %s should have type %s to match "
+                  "corresponding Verilog parameter", istr(tree_ident(cg)),
+                  type_pp(etype));
+         return NULL;
+      }
+
+      tree_t map = tree_new(T_PARAM);
+      tree_set_subkind(map, P_POS);
+      tree_set_pos(map, tree_genmaps(bind));
+      tree_set_value(map, make_ref(cg));
+
+      tree_add_genmap(bind, map);
+   }
+
+   for (int i = 0; i < vhdl_ngenerics; i++) {
+      if (!mask_test(&have, i)) {
+         tree_t g = tree_generic(comp, i);
+         diag_t *d = diag_new(DIAG_ERROR, tree_loc(g));
+         diag_printf(d, "generic %s not found in Verilog module %s",
+                     istr(tree_ident(g)), istr(vlog_ident2(mc->module)));
+         diag_emit(d);
+      }
+   }
+
+   mask_free(&have);
    return bind;
 }
 
@@ -508,34 +579,36 @@ static tree_t elab_verilog_conversion(type_t from, type_t to)
    return NULL;
 }
 
-static tree_t elab_verilog_binding(vlog_node_t inst, mod_cache_t *mc,
+static tree_t elab_verilog_binding(vlog_node_t inst, elab_instance_t *ei,
                                    const elab_ctx_t *ctx)
 {
    assert(vlog_kind(inst) == V_MOD_INST);
 
    tree_t bind = tree_new(T_BINDING);
-   tree_set_ident(bind, vlog_ident(mc->module));
+   tree_set_ident(bind, vlog_ident(ei->body));
    tree_set_loc(bind, vlog_loc(inst));
-   tree_set_ref(bind, mc->wrap);
+   tree_set_ref(bind, ei->module->wrap);
    tree_set_class(bind, C_ENTITY);
 
-   const int nports = vlog_ports(mc->module);
+   const int nports = vlog_ports(ei->body);
    const int nparams = vlog_params(inst);
    const int outports = tree_ports(ctx->out);
    const int outdecls = tree_decls(ctx->out);
 
    if (nports != nparams) {
       error_at(vlog_loc(inst), "expected %d port connections for module %s "
-               "but found %d", nports, istr(vlog_ident(mc->module)), nparams);
+               "but found %d", nports, istr(vlog_ident2(ei->body)), nparams);
       return NULL;
    }
 
    bool have_named = false;
    for (int i = 0; i < nports; i++) {
       vlog_node_t conn = vlog_param(inst, i);
-      assert(vlog_kind(conn) == V_REF);
+      assert(vlog_kind(conn) == V_PORT_CONN);
 
-      ident_t id = vlog_ident(conn);
+      vlog_node_t expr = vlog_value(conn);
+
+      ident_t id = vlog_ident(expr);
       tree_t decl = NULL;
 
       for (int j = 0; j < outports; j++) {
@@ -558,7 +631,7 @@ static tree_t elab_verilog_binding(vlog_node_t inst, mod_cache_t *mc,
 
       assert(decl != NULL);
 
-      tree_t port = tree_port(mc->block, i);
+      tree_t port = tree_port(ei->block, i);
 
       type_t dtype = tree_type(decl);
       type_t ptype = tree_type(port);
@@ -1319,7 +1392,6 @@ static void elab_inherit_context(elab_ctx_t *ctx, const elab_ctx_t *parent)
    ctx->mir       = parent->mir;
    ctx->root      = parent->root;
    ctx->dotted    = ctx->dotted ?: parent->dotted;
-   ctx->inst_name = ctx->inst_name ?: parent->inst_name;
    ctx->library   = ctx->library ?: parent->library;
    ctx->out       = ctx->out ?: parent->out;
    ctx->cover     = parent->cover;
@@ -1340,10 +1412,13 @@ static driver_set_t *elab_driver_set(const elab_ctx_t *ctx)
       return NULL;
 }
 
-static void elab_lower(tree_t b, vcode_unit_t shape, elab_ctx_t *ctx)
+static void elab_lower(tree_t b, const elab_instance_t *ei, elab_ctx_t *ctx)
 {
-   ctx->lowered = lower_instance(ctx->registry, ctx->parent->lowered, shape,
-                                 elab_driver_set(ctx), ctx->cover, b);
+   if (ei != NULL)
+      vlog_lower_block(ctx->mir, ctx->parent->dotted, b);
+   else
+      ctx->lowered = lower_instance(ctx->registry, ctx->parent->lowered,
+                                    elab_driver_set(ctx), ctx->cover, b);
 
    if (ctx->inst != NULL)
       diag_add_hint_fn(elab_hint_fn, ctx->inst);
@@ -1354,15 +1429,13 @@ static void elab_lower(tree_t b, vcode_unit_t shape, elab_ctx_t *ctx)
       diag_remove_hint_fn(elab_hint_fn);
 }
 
-static void elab_verilog_module(tree_t bind, ident_t label, mod_cache_t *mc,
+static void elab_verilog_module(tree_t bind, ident_t label,
+                                const elab_instance_t *ei,
                                 const elab_ctx_t *ctx)
 {
-   const char *label_str = istr(label);
-   ident_t ninst = hpathf(ctx->inst_name, ':', "%s", label_str);
    ident_t ndotted = ident_prefix(ctx->dotted, label, '.');
 
    elab_ctx_t new_ctx = {
-      .inst_name = ninst,
       .dotted    = ndotted,
    };
    elab_inherit_context(&new_ctx, ctx);
@@ -1374,40 +1447,137 @@ static void elab_verilog_module(tree_t bind, ident_t label, mod_cache_t *mc,
    tree_add_stmt(ctx->out, b);
    new_ctx.out = b;
 
-   elab_push_scope(mc->wrap, &new_ctx);
+   elab_push_scope(ei->wrap, &new_ctx);
 
    if (bind != NULL)
-      elab_ports(mc->block, bind, &new_ctx);
+      elab_ports(ei->block, bind, &new_ctx);
 
    if (error_count() == 0)
-      elab_decls(mc->block, &new_ctx);
+      elab_decls(ei->block, &new_ctx);
 
    if (error_count() == 0) {
-      new_ctx.drivers = find_drivers(mc->block);
-      elab_lower(b, mc->shape, &new_ctx);
-      elab_stmts(mc->block, &new_ctx);
+      new_ctx.drivers = find_drivers(ei->block);
+      elab_lower(b, ei, &new_ctx);
    }
 
+   if (error_count() == 0)
+      elab_verilog_stmts(ei->body, &new_ctx);
+
    elab_pop_scope(&new_ctx);
 }
 
+static void elab_verilog_instance_list(vlog_node_t v, const elab_ctx_t *ctx)
+{
+   ident_t modname = vlog_ident(v);
+   ident_t libname = lib_name(ctx->library);
+
+   text_buf_t *tb = tb_new();
+   tb_istr(tb, libname);
+   tb_append(tb, '.');
+   tb_istr(tb, modname);
+   tb_upcase(tb);
+
+   ident_t qual = ident_new(tb_get(tb));
+
+   object_t *obj = lib_get_generic(ctx->library, qual, NULL);
+   if (obj == NULL) {
+      error_at(vlog_loc(v), "module %s not found in library %s",
+               istr(modname), istr(libname));
+      return;
+   }
+
+   vlog_node_t mod = vlog_from_object(obj);
+   if (mod == NULL) {
+      error_at(&obj->loc, "unit %s is not a Verilog module", istr(qual));
+      return;
+   }
+   else if (vlog_ident2(mod) != modname) {
+      diag_t *d = diag_new(DIAG_ERROR, vlog_loc(v));
+      diag_printf(d, "name of Verilog module %s in library unit %s "
+                  "does not match name %s in module instance %s",
+                  istr(vlog_ident2(mod)), istr(qual), istr(modname),
+                  istr(vlog_ident(vlog_stmt(v, 0))));
+      diag_hint(d, NULL, "this tool does not preserve case sensitivity "
+                "in module names");
+      diag_emit(d);
+      return;
+   }
+
+   elab_instance_t *ei = elab_new_instance(mod, v, ctx);
+
+   const int nstmts = vlog_stmts(v);
+   for (int i = 0; i < nstmts; i++) {
+      vlog_node_t inst = vlog_stmt(v, i);
+      assert(vlog_kind(inst) == V_MOD_INST);
+
+      tree_t bind = elab_verilog_binding(inst, ei, ctx);
+      if (bind != NULL)
+         elab_verilog_module(bind, vlog_ident(inst), ei, ctx);
+   }
+}
+
+static void elab_verilog_stmts(vlog_node_t v, const elab_ctx_t *ctx)
+{
+   const int nstmts = vlog_stmts(v);
+   for (int i = 0; i < nstmts; i++) {
+      vlog_node_t s = vlog_stmt(v, i);
+
+      const vlog_kind_t kind = vlog_kind(s);
+      switch (kind) {
+      case V_INST_LIST:
+         elab_verilog_instance_list(s, ctx);
+         break;
+      case V_INITIAL:
+      case V_ALWAYS:
+      case V_ASSIGN:
+      case V_UDP_TABLE:
+      case V_GATE_INST:
+         {
+            ident_t name = vlog_ident(s);
+            ident_t sym = ident_prefix(ctx->dotted, name, '.');
+
+            tree_t w = tree_new(T_VERILOG);
+            tree_set_ident(w, name);
+            tree_set_loc(w, vlog_loc(s));
+            tree_set_vlog(w, s);
+
+            tree_add_stmt(ctx->out, w);
+
+            if (kind == V_UDP_TABLE)
+               mir_defer(ctx->mir, sym, ctx->dotted, MIR_UNIT_PROCESS,
+                         vlog_lower_udp, vlog_to_object(v));
+            else
+               mir_defer(ctx->mir, sym, ctx->dotted, MIR_UNIT_PROCESS,
+                         vlog_lower_deferred, vlog_to_object(s));
+         }
+         break;
+      case V_BLOCK:
+         // TODO: should create a block in hierarchy
+         elab_verilog_stmts(s, ctx);
+         break;
+      case V_IF_GENERATE:
+         error_at(vlog_loc(s), "if-generate construct could not be "
+                  "evaluated at elaboration time");
+         break;
+      default:
+         fatal_at(vlog_loc(s), "sorry, this Verilog statement is not "
+                  "currently supported");
+      }
+   }
+}
+
 static void elab_architecture(tree_t bind, tree_t arch, tree_t config,
                               const elab_ctx_t *ctx)
 {
    tree_t inst = NULL;
-   ident_t label, ninst = NULL;
+   ident_t label;
    switch (tree_kind(bind)) {
    case T_BINDING:
       label = ident_rfrom(tree_ident(tree_primary(arch)), '.');
       break;
    case T_INSTANCE:
-      {
-         label = tree_ident(bind);
-         inst = bind;
-         ninst = hpathf(ctx->inst_name, ':', "%s@%s(%s)", istr(label),
-                        simple_name(istr(tree_ident2(arch))),
-                        simple_name(istr(tree_ident(arch))));
-      }
+      label = tree_ident(bind);
+      inst = bind;
       break;
    default:
       fatal_trace("unexpected binding kind %s in elab_architecture",
@@ -1417,9 +1587,8 @@ static void elab_architecture(tree_t bind, tree_t arch, tree_t config,
    ident_t ndotted = ident_prefix(ctx->dotted, label, '.');
 
    elab_ctx_t new_ctx = {
-      .inst_name = ninst,
-      .dotted    = ndotted,
-      .inst      = inst,
+      .dotted = ndotted,
+      .inst   = inst,
    };
    elab_inherit_context(&new_ctx, ctx);
 
@@ -1554,17 +1723,9 @@ static void elab_component(tree_t inst, tree_t comp, const elab_ctx_t *ctx)
       }
    }
 
-   ident_t ninst = hpathf(ctx->inst_name, ':', "%s", istr(tree_ident(inst)));
-
-   if (arch != NULL && tree_kind(arch) != T_VERILOG)
-      ninst = hpathf(ninst, '@', "%s(%s)",
-                     simple_name(istr(tree_ident2(arch))),
-                     simple_name(istr(tree_ident(arch))));
-
    ident_t ndotted = ident_prefix(ctx->dotted, tree_ident(inst), '.');
 
    elab_ctx_t new_ctx = {
-      .inst_name = ninst,
       .dotted    = ndotted,
       .inst      = inst,
    };
@@ -1589,13 +1750,51 @@ static void elab_component(tree_t inst, tree_t comp, const elab_ctx_t *ctx)
    if (error_count() == 0)
       elab_lower(b, NULL, &new_ctx);
 
-   if (arch == NULL)
+   if (error_count() > 0 || arch == NULL)
       ;   // Unbound architecture
    else if (tree_kind(arch) == T_VERILOG) {
-      mod_cache_t *mc = elab_cached_module(tree_vlog(arch), ctx);
-      elab_verilog_module(bind, vlog_ident2(mc->module), mc, &new_ctx);
+      vlog_node_t mod = tree_vlog(arch);
+
+      vlog_node_t stmt = vlog_new(V_MOD_INST);
+      vlog_set_ident(stmt, vlog_ident2(mod));
+
+      vlog_node_t list = vlog_new(V_INST_LIST);
+      vlog_set_ident(list, vlog_ident2(mod));
+      vlog_add_stmt(list, stmt);
+
+      const int ngenmaps = tree_genmaps(bind);
+      for (int i = 0; i < ngenmaps; i++) {
+         tree_t decl = tree_ref(tree_value(tree_genmap(bind, i)));
+         assert(tree_kind(decl) == T_GENERIC_DECL);
+
+         tree_t map = hash_get(new_ctx.generics, decl);
+         if (map == NULL)
+            fatal_at(tree_loc(inst), "cannot get value of generic %s",
+                     istr(tree_ident(decl)));
+
+         const int64_t ival = assume_int(map);
+
+         // TODO: have some better way of constructing these
+         char buf[64];
+         checked_sprintf(buf, sizeof(buf), "%lld", llabs(ival));
+
+         number_t n = number_new(buf, NULL);
+         if (ival < 0)
+            n = number_negate(n);
+
+         vlog_node_t num = vlog_new(V_NUMBER);
+         vlog_set_number(num, n);
+
+         vlog_node_t pa = vlog_new(V_PARAM_ASSIGN);
+         vlog_set_value(pa, num);
+
+         vlog_add_param(list, pa);
+      }
+
+      elab_instance_t *ei = elab_new_instance(mod, list, &new_ctx);
+      elab_verilog_module(bind, vlog_ident2(mod), ei, &new_ctx);
    }
-   else if (error_count() == 0)
+   else
       elab_architecture(bind, arch, config, &new_ctx);
 
    elab_pop_scope(&new_ctx);
@@ -1689,6 +1888,7 @@ static void elab_decls(tree_t t, const elab_ctx_t *ctx)
       case T_PACK_INST:
       case T_PSL_DECL:
       case T_ATTR_SPEC:
+      case T_VERILOG:
          tree_add_decl(ctx->out, d);
          break;
       case T_FUNC_DECL:
@@ -1709,7 +1909,7 @@ static void elab_push_scope(tree_t t, elab_ctx_t *ctx)
    tree_set_subkind(h, tree_kind(t));
    tree_set_ref(h, t);
 
-   tree_set_ident(h, ctx->inst_name);
+   tree_set_ident(h, ctx->dotted);
    tree_set_ident2(h, ctx->dotted);
 
    tree_add_decl(ctx->out, h);
@@ -1812,13 +2012,10 @@ static void elab_for_generate(tree_t t, const elab_ctx_t *ctx)
       tree_add_generic(b, g);
       tree_add_genmap(b, map);
 
-      const char *label = istr(base);
-      ident_t ninst = hpathf(ctx->inst_name, ':', "%s(%"PRIi64")", label, i);
       ident_t ndotted = ident_prefix(ctx->dotted, id, '.');
 
       elab_ctx_t new_ctx = {
          .out       = b,
-         .inst_name = ninst,
          .dotted    = ndotted,
          .generics  = hash_new(16),
          .config    = elab_block_config(t, ctx),
@@ -1881,14 +2078,12 @@ static void elab_if_generate(tree_t t, const elab_ctx_t *ctx)
          tree_add_stmt(ctx->out, b);
 
          ident_t name = tree_ident(cond);
-         ident_t ninst = hpathf(ctx->inst_name, ':', "%s", name);
          ident_t ndotted = ident_prefix(ctx->dotted, name, '.');
 
          elab_ctx_t new_ctx = {
-            .out       = b,
-            .inst_name = ninst,
-            .dotted    = ndotted,
-            .config    = elab_block_config(cond, ctx),
+            .out    = b,
+            .dotted = ndotted,
+            .config = elab_block_config(cond, ctx),
          };
          elab_inherit_context(&new_ctx, ctx);
 
@@ -1923,13 +2118,11 @@ static void elab_case_generate(tree_t t, const elab_ctx_t *ctx)
 
    tree_add_stmt(ctx->out, b);
 
-   ident_t ninst = hpathf(ctx->inst_name, ':', "%s", istr(id));
    ident_t ndotted = ident_prefix(ctx->dotted, id, '.');
 
    elab_ctx_t new_ctx = {
-      .out       = b,
-      .inst_name = ninst,
-      .dotted    = ndotted,
+      .out    = b,
+      .dotted = ndotted,
    };
    elab_inherit_context(&new_ctx, ctx);
 
@@ -1962,60 +2155,6 @@ static void elab_psl(tree_t t, const elab_ctx_t *ctx)
    tree_add_stmt(ctx->out, t);
 }
 
-static void elab_verilog_stmt(tree_t wrap, const elab_ctx_t *ctx)
-{
-   vlog_node_t v = tree_vlog(wrap);
-   switch (vlog_kind(v)) {
-   case V_MOD_INST:
-      {
-         ident_t modname = vlog_ident2(v);
-         ident_t libname = lib_name(ctx->library);
-
-         text_buf_t *tb = tb_new();
-         tb_istr(tb, libname);
-         tb_append(tb, '.');
-         tb_istr(tb, modname);
-         tb_upcase(tb);
-
-         ident_t qual = ident_new(tb_get(tb));
-
-         object_t *obj = lib_get_generic(ctx->library, qual, NULL);
-         if (obj == NULL) {
-            error_at(vlog_loc(v), "module %s not found in library %s",
-                     istr(modname), istr(libname));
-            return;
-         }
-
-         vlog_node_t mod = vlog_from_object(obj);
-         if (mod == NULL) {
-            error_at(&obj->loc, "unit %s is not a Verilog module", istr(qual));
-            return;
-         }
-         else if (vlog_ident2(mod) != modname) {
-            diag_t *d = diag_new(DIAG_ERROR, vlog_loc(v));
-            diag_printf(d, "name of Verilog module %s in library unit %s "
-                        "does not match name %s in module instance %s",
-                        istr(vlog_ident2(mod)), istr(qual), istr(modname),
-                        istr(vlog_ident(v)));
-            diag_hint(d, NULL, "this tool does not preserve case sensitivity "
-                      "in module names");
-            diag_emit(d);
-            return;
-         }
-
-         mod_cache_t *mc = elab_cached_module(mod, ctx);
-
-         tree_t bind = elab_verilog_binding(v, mc, ctx);
-         if (bind != NULL)
-            elab_verilog_module(bind, vlog_ident(v), mc, ctx);
-      }
-      break;
-   default:
-      tree_add_stmt(ctx->out, wrap);
-      break;
-   }
-}
-
 static void elab_stmts(tree_t t, const elab_ctx_t *ctx)
 {
    const int nstmts = tree_stmts(t);
@@ -2044,9 +2183,6 @@ static void elab_stmts(tree_t t, const elab_ctx_t *ctx)
       case T_PSL_DIRECT:
          elab_psl(s, ctx);
          break;
-      case T_VERILOG:
-         elab_verilog_stmt(s, ctx);
-         break;
       default:
          fatal_trace("unexpected statement %s", tree_kind_str(tree_kind(s)));
       }
@@ -2063,14 +2199,12 @@ static void elab_block(tree_t t, const elab_ctx_t *ctx)
 
    tree_add_stmt(ctx->out, b);
 
-   ident_t ninst = hpathf(ctx->inst_name, ':', "%s", istr(id));
    ident_t ndotted = ident_prefix(ctx->dotted, id, '.');
 
    elab_ctx_t new_ctx = {
-      .out       = b,
-      .inst_name = ninst,
-      .dotted    = ndotted,
-      .config    = elab_block_config(t, ctx),
+      .out    = b,
+      .dotted = ndotted,
+      .config = elab_block_config(t, ctx),
    };
    elab_inherit_context(&new_ctx, ctx);
 
@@ -2201,10 +2335,6 @@ static void elab_vhdl_root_cb(void *arg)
       fatal("%s is not a suitable top-level unit", istr(tree_ident(vhdl)));
    }
 
-   const char *name = simple_name(istr(tree_ident2(arch)));
-   ctx->inst_name = hpathf(NULL, ':', ":%s(%s)", name,
-                           simple_name(istr(tree_ident(arch))));
-
    tree_t bind = elab_top_level_binding(arch, ctx);
 
    if (error_count() == 0)
@@ -2218,8 +2348,17 @@ static void elab_verilog_root_cb(void *arg)
    vlog_node_t vlog = vlog_from_object(ctx->root);
    assert(vlog != NULL);
 
-   mod_cache_t *mc = elab_cached_module(vlog, ctx);
-   elab_verilog_module(NULL, vlog_ident2(mc->module), mc, ctx);
+   ident_t label = vlog_ident2(vlog);
+
+   vlog_node_t stmt = vlog_new(V_MOD_INST);
+   vlog_set_ident(stmt, label);
+
+   vlog_node_t list = vlog_new(V_INST_LIST);
+   vlog_set_ident(list, label);
+   vlog_add_stmt(list, stmt);
+
+   elab_instance_t *ei = elab_new_instance(vlog, list, ctx);
+   elab_verilog_module(NULL, label, ei, ctx);
 }
 
 tree_t elab(object_t *top, jit_t *jit, unit_registry_t *ur, mir_context_t *mc,
@@ -2249,7 +2388,6 @@ tree_t elab(object_t *top, jit_t *jit, unit_registry_t *ur, mir_context_t *mc,
    elab_ctx_t ctx = {
       .out       = e,
       .root      = top,
-      .inst_name = NULL,
       .cover     = cover,
       .library   = work,
       .jit       = jit,
