@@ -25,6 +25,7 @@
 #include "thread.h"
 
 #include <assert.h>
+#include <math.h>
 #include <stdlib.h>
 #include <string.h>
 #include <stdio.h>
@@ -146,6 +147,7 @@ typedef struct _code_cache {
    code_span_t *spans;
    code_span_t *freelist[MAX_THREADS];
    code_span_t *globalfree;
+   shash_t     *symbols;
    FILE        *perfmap;
 #ifdef HAVE_CAPSTONE
    csh          capstone;
@@ -274,6 +276,41 @@ code_cache_t *code_cache_new(void)
       fatal_trace("failed to set capstone detailed mode");
 #endif
 
+   shash_t *s = shash_new(32);
+
+   extern void __nvc_putpriv(jit_handle_t, void *);
+   extern void __nvc_sched_waveform(jit_anchor_t *, jit_scalar_t *, tlab_t *);
+   extern void __nvc_sched_process(jit_anchor_t *, jit_scalar_t *, tlab_t *);
+   extern void __nvc_test_event(jit_anchor_t *, jit_scalar_t *, tlab_t *);
+   extern void __nvc_last_event(jit_anchor_t *, jit_scalar_t *, tlab_t *);
+   extern void __nvc_pack(const uint8_t *, int32_t, jit_scalar_t *);
+   extern void __nvc_unpack(uint64_t, uint64_t, jit_scalar_t *);
+
+   shash_put(s, "__nvc_sched_waveform", &__nvc_sched_waveform);
+   shash_put(s, "__nvc_sched_process", &__nvc_sched_process);
+   shash_put(s, "__nvc_test_event", &__nvc_test_event);
+   shash_put(s, "__nvc_last_event", &__nvc_last_event);
+   shash_put(s, "__nvc_mspace_alloc", &__nvc_mspace_alloc);
+   shash_put(s, "__nvc_putpriv", &__nvc_putpriv);
+   shash_put(s, "__nvc_do_exit", &__nvc_do_exit);
+   shash_put(s, "__nvc_pack", &__nvc_pack);
+   shash_put(s, "__nvc_unpack", &__nvc_unpack);
+   shash_put(s, "memmove", &memmove);
+   shash_put(s, "memcpy", &memcpy);
+   shash_put(s, "memset", &memset);
+   shash_put(s, "pow", &pow);
+
+#if defined __APPLE__ && defined ARCH_ARM64
+   shash_put(s, "bzero", &bzero);
+#elif defined __APPLE__ && defined ARCH_X86_64
+   shash_put(s, "__bzero", &bzero);
+#elif defined __MINGW32__ && defined ARCH_X86_64
+   extern void ___chkstk_ms(void);
+   shash_put(s, "___chkstk_ms", &___chkstk_ms);
+#endif
+
+   store_release(&code->symbols, s);
+
    return code;
 }
 
@@ -291,6 +328,7 @@ void code_cache_free(code_cache_t *code)
 
    for (code_span_t *it = code->spans, *tmp; it; it = tmp) {
       tmp = it->next;
+      DEBUG_ONLY(free(it->debug.comments));
       free(it);
    }
 
@@ -303,6 +341,7 @@ void code_cache_free(code_cache_t *code)
       debugf("JIT code footprint: %zu bytes", code->used);
 #endif
 
+   shash_free(code->symbols);
    free(code);
 }
 
@@ -315,6 +354,21 @@ static int code_print_spaces(int col, int tab)
 }
 #endif
 
+#if defined DEBUG && HAVE_CAPSTONE
+static int code_comment_compare(const void *a, const void *b)
+{
+   const code_comment_t *ca = a;
+   const code_comment_t *cb = b;
+
+   if (ca->addr < cb->addr)
+      return -1;
+   else if (ca->addr > cb->addr)
+      return 1;
+   else
+      return 0;
+}
+#endif
+
 static void code_disassemble(code_span_t *span, uintptr_t mark,
                              struct cpu_state *cpu)
 {
@@ -332,6 +386,8 @@ static void code_disassemble(code_span_t *span, uintptr_t mark,
    cs_insn *insn = cs_malloc(span->owner->capstone);
 
 #ifdef DEBUG
+   qsort(span->debug.comments, span->debug.count, sizeof(code_comment_t),
+         code_comment_compare);
    code_comment_t *comment = span->debug.comments;
 #endif
 
@@ -682,7 +738,7 @@ static void code_blob_print_value(text_buf_t *tb, jit_value_t value)
    }
 }
 
-static void code_blob_add_comment(code_blob_t *blob, char *text)
+static void code_blob_add_comment(code_blob_t *blob, uintptr_t addr, char *text)
 {
    code_debug_t *dbg = &(blob->span->debug);
 
@@ -692,7 +748,7 @@ static void code_blob_add_comment(code_blob_t *blob, char *text)
                                      sizeof(code_comment_t));
    }
 
-   dbg->comments[dbg->count].addr = (uintptr_t)blob->wptr;
+   dbg->comments[dbg->count].addr = addr;
    dbg->comments[dbg->count].text = text;
    dbg->count++;
 }
@@ -721,49 +777,36 @@ void code_blob_print_ir(code_blob_t *blob, jit_ir_t *ir)
       code_blob_print_value(tb, ir->arg2);
    }
 
-   code_blob_add_comment(blob, tb_claim(tb));
+   code_blob_add_comment(blob, (uintptr_t)blob->wptr, tb_claim(tb));
 }
 
 void code_blob_printf(code_blob_t *blob, const char *fmt, ...)
 {
-   code_debug_t *dbg = &(blob->span->debug);
-
-   if (dbg->count == dbg->max) {
-      dbg->max = MAX(128, dbg->max * 2);
-      dbg->comments = xrealloc_array(dbg->comments, dbg->max,
-                                     sizeof(code_comment_t));
-   }
-
    va_list ap;
    va_start(ap, fmt);
 
    char *text = xvasprintf(fmt, ap);
-   code_blob_add_comment(blob, text);
+   code_blob_add_comment(blob, (uintptr_t)blob->wptr, text);
 
    va_end(ap);
 }
+
+__attribute__((format(printf, 3, 4)))
+static void debug_reloc(code_blob_t *blob, void *patch, const char *fmt, ...)
+{
+   va_list ap;
+   va_start(ap, fmt);
+
+   char *text = xvasprintf(fmt, ap);
+   code_blob_add_comment(blob, (uintptr_t)patch, text);
+
+   va_end(ap);
+}
+#else
+#define debug_reloc(...)
 #endif   // DEBUG
 
 #ifdef ARCH_ARM64
-static void *arm64_emit_trampoline(code_blob_t *blob, uintptr_t dest)
-{
-   const uint8_t veneer[] = {
-      0x50, 0x00, 0x00, 0x58,   // LDR X16, [PC+8]
-      0x00, 0x02, 0x1f, 0xd6,   // BR X16
-      __IMM64(dest)
-   };
-
-   void *prev = memmem(blob->span->base, blob->span->size,
-                       veneer, ARRAY_LEN(veneer));
-   if (prev != NULL)
-      return prev;
-   else {
-      void *addr = blob->wptr;
-      code_blob_emit(blob, veneer, ARRAY_LEN(veneer));
-      return addr;
-   }
-}
-
 static void arm64_patch_page_offset21(code_blob_t *blob, uint32_t *patch,
                                       void *ptr)
 {
@@ -789,11 +832,59 @@ static void arm64_patch_page_base_rel21(uint32_t *patch, void *ptr)
    const intptr_t dst_page = (intptr_t)ptr & ~UINT64_C(0xfff);
    const intptr_t src_page = (intptr_t)patch & ~UINT64_C(0xfff);
    const intptr_t upper21 = (dst_page - src_page) >> 12;
+   assert((upper21 & ~UINT64_C(0x1fffff)) == 0);
    *(uint32_t *)patch |= (upper21 & 3) << 29;
    *(uint32_t *)patch |= ((upper21 >> 2) & 0x7ffff) << 5;
 }
+#endif
+
+static void *code_emit_trampoline(code_blob_t *blob, void *dest)
+{
+#if defined ARCH_X86_64
+   const uint8_t veneer[] = {
+      0x48, 0xb8, __IMM64((uintptr_t)dest),  // MOVABS RAX, dest
+      0xff, 0xe0                             // CALL RAX
+   };
+#elif defined ARCH_ARM64
+   const uint8_t veneer[] = {
+      0x50, 0x00, 0x00, 0x58,   // LDR X16, [PC+8]
+      0x00, 0x02, 0x1f, 0xd6,   // BR X16
+      __IMM64((uintptr_t)dest)
+   };
 #else
-#define arm64_emit_trampoline(blob, dest) NULL
+   should_not_reach_here();
+#endif
+
+   void *prev = memmem(blob->veneers, blob->wptr - blob->veneers,
+                       veneer, ARRAY_LEN(veneer));
+   if (prev != NULL)
+      return prev;
+   else {
+      DEBUG_ONLY(code_blob_printf(blob, "Trampoline for %p", dest));
+
+      void *addr = blob->wptr;
+      code_blob_emit(blob, veneer, ARRAY_LEN(veneer));
+      return addr;
+   }
+}
+
+#if !defined __MINGW32__ && !defined __APPLE__
+static void *code_emit_got(code_blob_t *blob, void *dest)
+{
+   const uint8_t data[] = { __IMM64((uintptr_t)dest) };
+
+   void *prev = memmem(blob->veneers, blob->veneers - blob->wptr,
+                       data, ARRAY_LEN(data));
+   if (prev != NULL)
+      return prev;
+   else {
+      DEBUG_ONLY(code_blob_printf(blob, "GOT entry for %p", dest));
+
+      void *addr = blob->wptr;
+      code_blob_emit(blob, data, ARRAY_LEN(data));
+      return addr;
+   }
+}
 #endif
 
 #if defined __MINGW32__
@@ -836,6 +927,10 @@ static void code_load_pe(code_blob_t *blob, const void *data, size_t size)
    if (blob->overflow)
       return;   // Relocations might point outside of code span
 
+   blob->veneers = blob->wptr;
+
+   shash_t *external = load_acquire(&blob->span->owner->symbols);
+
    for (int i = 0; i < imghdr->NumberOfSections; i++) {
       const IMAGE_RELOCATION *relocs = data + sections[i].PointerToRelocations;
       for (int j = 0; j < sections[i].NumberOfRelocations; j++) {
@@ -858,14 +953,8 @@ static void code_load_pe(code_blob_t *blob, const void *data, size_t size)
             assert(sym->SectionNumber - 1 < imghdr->NumberOfSections);
             ptr = load_addr[sym->SectionNumber - 1] + sym->Value;
          }
-#ifdef ARCH_X86_64
-         else if (strcmp(name, "___chkstk_ms") == 0) {
-            extern void ___chkstk_ms(void);
-            ptr = &___chkstk_ms;
-         }
-#endif
          else
-            ptr = ffi_find_symbol(NULL, name);
+            ptr = shash_get(external, name);
 
          if (ptr == NULL && icmp(blob->span->name, name))
             ptr = blob->span->base;
@@ -888,7 +977,7 @@ static void code_load_pe(code_blob_t *blob, const void *data, size_t size)
 #elif defined ARCH_ARM64
          case IMAGE_REL_ARM64_BRANCH26:
             {
-               void *veneer = arm64_emit_trampoline(blob, (uintptr_t)ptr);
+               void *veneer = code_emit_trampoline(blob, ptr);
                const ptrdiff_t pcrel = (veneer - patch) >> 2;
                *(uint32_t *)patch &= ~0x3ffffff;
                *(uint32_t *)patch |= pcrel & 0x3ffffff;
@@ -968,6 +1057,7 @@ static void code_load_macho(code_blob_t *blob, const void *data, size_t size)
                   + j * sizeof(struct section_64);
                code_blob_align(blob, 1 << sec->align);
                load_addr[j] = blob->wptr;
+               DEBUG_ONLY(code_blob_printf(blob, "%s", sec->sectname));
                code_blob_emit(blob, data + sec->offset, sec->size);
             }
          }
@@ -992,9 +1082,13 @@ static void code_load_macho(code_blob_t *blob, const void *data, size_t size)
    if (blob->overflow)
       return;   // Relocations might point outside of code span
 
+   blob->veneers = blob->wptr;
+
    assert(seg != NULL);
    assert(symtab != NULL);
 
+   shash_t *external = load_acquire(&blob->span->owner->symbols);
+
    for (int i = 0; i < seg->nsects; i++) {
       const struct section_64 *sec =
          (void *)seg + sizeof(struct segment_command_64)
@@ -1015,7 +1109,7 @@ static void code_load_macho(code_blob_t *blob, const void *data, size_t size)
             if (nl->n_type & N_EXT) {
                if (icmp(blob->span->name, name + 1))
                   ptr = blob->span->base;
-               else if ((ptr = ffi_find_symbol(NULL, name + 1)) == NULL)
+               else if ((ptr = shash_get(external, name + 1)) == NULL)
                   fatal_trace("failed to resolve symbol %s", name + 1);
             }
             else if (nl->n_sect != NO_SECT)
@@ -1049,8 +1143,10 @@ static void code_load_macho(code_blob_t *blob, const void *data, size_t size)
             break;
          case ARM64_RELOC_BRANCH26:
             {
-               void *veneer = arm64_emit_trampoline(blob, (uintptr_t)ptr);
+               void *veneer = code_emit_trampoline(blob, ptr);
                const ptrdiff_t pcrel = (veneer - patch) >> 2;
+               debug_reloc(blob, patch, "ARM64_RELOC_BRANCH26 %s PC%+"PRIiPTR,
+                           name, pcrel);
                *(uint32_t *)patch &= ~0x3ffffff;
                *(uint32_t *)patch |= pcrel & 0x3ffffff;
             }
@@ -1117,6 +1213,7 @@ static void code_load_elf(code_blob_t *blob, const void *data, size_t size)
          if (shdr->sh_flags & SHF_ALLOC) {
             code_blob_align(blob, shdr->sh_addralign);
             load_addr[i] = blob->wptr;
+            DEBUG_ONLY(code_blob_printf(blob, "%s", strtab + shdr->sh_name));
             code_blob_emit(blob, data + shdr->sh_offset, shdr->sh_size);
          }
          break;
@@ -1158,6 +1255,10 @@ static void code_load_elf(code_blob_t *blob, const void *data, size_t size)
    if (blob->overflow)
       return;   // Relocations might point outside of code span
 
+   blob->veneers = blob->wptr;
+
+   shash_t *external = load_acquire(&blob->span->owner->symbols);
+
    for (int i = 0; i < ehdr->e_shnum; i++) {
       const Elf64_Shdr *shdr = data + ehdr->e_shoff + i * ehdr->e_shentsize;
       if (shdr->sh_type != SHT_RELA)
@@ -1181,55 +1282,90 @@ static void code_load_elf(code_blob_t *blob, const void *data, size_t size)
          const Elf64_Sym *sym = data + symtab->sh_offset
             + ELF64_R_SYM(r->r_info) * symtab->sh_entsize;
 
-         char *ptr = NULL;
+         void *ptr = NULL;
          switch (ELF64_ST_TYPE(sym->st_info)) {
          case STT_NOTYPE:
          case STT_FUNC:
-            ptr = ffi_find_symbol(NULL, strtab + sym->st_name);
+            if (sym->st_shndx == 0)
+               ptr = shash_get(external, strtab + sym->st_name);
+            else
+               ptr = load_addr[sym->st_shndx] + sym->st_value;
             break;
          case STT_SECTION:
             ptr = load_addr[sym->st_shndx];
             break;
+         default:
+            fatal_trace("cannot handle ELF symbol type %d",
+                        ELF64_ST_TYPE(sym->st_info));
          }
 
-         if (ptr == NULL && icmp(blob->span->name, strtab + sym->st_name))
-            ptr = (char *)blob->span->base;
-
          if (ptr == NULL)
             fatal_trace("cannot resolve symbol %s type %d",
                         strtab + sym->st_name, ELF64_ST_TYPE(sym->st_info));
 
-         ptr += r->r_addend;
-
          void *patch = load_addr[shdr->sh_info] + r->r_offset;
          assert(r->r_offset < mod->sh_size);
 
          switch (ELF64_R_TYPE(r->r_info)) {
          case R_X86_64_64:
-            *(uint64_t *)patch = (uint64_t)ptr;
+            debug_reloc(blob, patch, "R_X86_64_64 %s", strtab + sym->st_name);
+            *(uint64_t *)patch = (uint64_t)ptr + r->r_addend;
+            break;
+         case R_X86_64_PC32:
+            {
+               const ptrdiff_t pcrel = ptr + r->r_addend - patch;
+               debug_reloc(blob, patch, "R_X86_64_PC32 %s PC%+"PRIiPTR,
+                           strtab + sym->st_name, pcrel);
+               assert(pcrel >= INT32_MIN && pcrel <= INT32_MAX);
+               *(uint32_t *)patch = pcrel;
+            }
+            break;
+         case R_X86_64_GOTPCREL:
+            {
+               void *got = code_emit_got(blob, ptr);
+               const ptrdiff_t pcrel = got + r->r_addend - patch;
+               debug_reloc(blob, patch, "R_X86_64_GOTPCREL %s PC%+"PRIiPTR,
+                           strtab + sym->st_name, pcrel);
+               assert(pcrel >= INT32_MIN && pcrel <= INT32_MAX);
+               *(uint32_t *)patch = pcrel;
+            }
+            break;
+         case R_X86_64_PLT32:
+            {
+               void *veneer = code_emit_trampoline(blob, ptr);
+               const ptrdiff_t pcrel = veneer + r->r_addend - patch;
+               debug_reloc(blob, patch, "R_X86_64_PLT32 %s PC%+"PRIiPTR,
+                           strtab + sym->st_name, pcrel);
+               assert(pcrel >= INT32_MIN && pcrel <= INT32_MAX);
+               *(uint32_t *)patch = pcrel;
+            }
             break;
          case R_AARCH64_CALL26:
             {
-               void *veneer = arm64_emit_trampoline(blob, (uintptr_t)ptr);
-               const ptrdiff_t pcrel = (veneer - patch) >> 2;
+               void *veneer = code_emit_trampoline(blob, ptr);
+               const ptrdiff_t pcrel = (veneer + r->r_addend - patch) >> 2;
                *(uint32_t *)patch &= ~0x3ffffff;
                *(uint32_t *)patch |= pcrel & 0x3ffffff;
             }
             break;
          case R_AARCH64_PREL64:
-            *(uint64_t *)patch = ptr - (char *)patch;
+            *(uint64_t *)patch = ptr + r->r_addend - patch;
             break;
          case R_AARCH64_MOVW_UABS_G0_NC:
-            *(uint32_t *)patch |= ((uintptr_t)ptr & 0xffff) << 5;
+            *(uint32_t *)patch |=
+               (((uintptr_t)ptr + r->r_addend) & 0xffff) << 5;
             break;
          case R_AARCH64_MOVW_UABS_G1_NC:
-            *(uint32_t *)patch |= (((uintptr_t)ptr >> 16) & 0xffff) << 5;
+            *(uint32_t *)patch |=
+               ((((uintptr_t)ptr + r->r_addend) >> 16) & 0xffff) << 5;
             break;
          case R_AARCH64_MOVW_UABS_G2_NC:
-            *(uint32_t *)patch |= (((uintptr_t)ptr >> 32) & 0xffff) << 5;
+            *(uint32_t *)patch |=
+               ((((uintptr_t)ptr + r->r_addend) >> 32) & 0xffff) << 5;
             break;
          case R_AARCH64_MOVW_UABS_G3:
-            *(uint32_t *)patch |= (((uintptr_t)ptr >> 48) & 0xffff) << 5;
+            *(uint32_t *)patch |=
+               ((((uintptr_t)ptr + r->r_addend) >> 48) & 0xffff) << 5;
             break;
          default:
             blob->span->size = blob->wptr - blob->span->base;
