@@ -99,7 +99,6 @@ typedef struct _jit {
    int               exit_status;
    jit_tier_t       *tiers;
    aot_dll_t        *aotlib;
-   aot_dll_t        *preloadlib;
    jit_pack_t       *pack;
    func_array_t     *funcs;
    unsigned          next_handle;
@@ -187,7 +186,7 @@ void jit_free(jit_t *j)
    store_release(&j->shutdown, true);
    async_barrier();
 
-   aot_dll_t *libs[] = { j->aotlib, j->preloadlib };
+   aot_dll_t *libs[] = { j->aotlib };
    for (int i = 0; i < ARRAY_LEN(libs); i++) {
       if (libs[i] != NULL) {
          ffi_unload_dll(libs[i]->dll);
@@ -219,6 +218,11 @@ void jit_free(jit_t *j)
    free(j);
 }
 
+bool jit_is_shutdown(jit_t *j)
+{
+   return load_acquire(&j->shutdown);
+}
+
 mspace_t *jit_get_mspace(jit_t *j)
 {
    return j->mspace;
@@ -266,11 +270,11 @@ static jit_handle_t jit_lazy_compile_locked(jit_t *j, ident_t name)
       return f->handle;
 
    aot_descr_t *descr = NULL;
-   if (j->aotlib != NULL || j->preloadlib != NULL) {
+   if (j->aotlib != NULL) {
       LOCAL_TEXT_BUF tb = safe_symbol(name);
       tb_cat(tb, ".descr");
 
-      aot_dll_t *try[] = { j->aotlib, j->preloadlib };
+      aot_dll_t *try[] = { j->aotlib };
       for (int i = 0; i < ARRAY_LEN(try); i++) {
          if (try[i] == NULL)
             continue;
@@ -425,9 +429,6 @@ void jit_fill_irbuf(jit_func_t *f)
    if (jit_fill_from_aot(f, f->jit->aotlib))
       goto done;
 
-   if (jit_fill_from_aot(f, f->jit->preloadlib))
-      goto done;
-
    if (f->jit->pack != NULL && jit_pack_fill(f->jit->pack, f->jit, f))
       goto done;
 
@@ -436,11 +437,8 @@ void jit_fill_irbuf(jit_func_t *f)
    if (mu == NULL && f->jit->registry != NULL) {
       // Unit registry and MIR import is not thread-safe
       SCOPED_LOCK(f->jit->lock);
-      vcode_unit_t unit = unit_registry_get(f->jit->registry, f->name);
-      if (unit != NULL) {
-         mu = mir_import(f->jit->mir, unit);
-         mir_put_unit(f->jit->mir, mu);
-      }
+      (void)unit_registry_get(f->jit->registry, f->name);
+      mu = mir_get_unit(f->jit->mir, f->name);
    }
 
    if (mu == NULL) {
@@ -571,6 +569,9 @@ static void jit_emit_trace(diag_t *d, const loc_t *loc, object_t *enclosing,
       case T_PARAM_DECL:
          diag_trace(d, loc, "Parameter %s", istr(tree_ident(tree)));
          break;
+      case T_PARAM:
+         diag_trace(d, loc, "Port conversion");
+         break;
       default:
          {
             const char *class = class_str(class_of(tree));
@@ -882,32 +883,6 @@ static aot_dll_t *load_dll_internal(jit_t *j, const char *path)
    return lib;
 }
 
-void jit_preload(jit_t *j)
-{
-#ifdef HAVE_LLVM
-   if (j->preloadlib != NULL)
-      fatal_trace("Preload library already loaded");
-
-   lib_t std = lib_find(well_known(W_STD));
-   if (std == NULL)
-      fatal("cannot find STD library");
-
-   LOCAL_TEXT_BUF tb = tb_new();
-   tb_cat(tb, lib_path(std));
-   tb_cat(tb, DIR_SEP "..");
-
-   const char *preload_vers[] = { "93", "93", "93", "93", "08", "19" };
-   tb_printf(tb, DIR_SEP "preload%s." DLL_EXT, preload_vers[standard()]);
-
-   const char *path = tb_get(tb);
-   file_info_t info;
-   if (!get_file_info(path, &info) || info.type != FILE_REGULAR)
-      fatal("missing preload library at %s", path);
-
-   j->preloadlib = load_dll_internal(j, path);
-#endif  // HAVE_LLVM
-}
-
 void jit_load_dll(jit_t *j, ident_t name)
 {
    lib_t lib = lib_require(ident_until(name, '.'));
@@ -1004,7 +979,7 @@ static void jit_async_cgen(void *context, void *arg)
    jit_func_t *f = context;
    jit_tier_t *tier = arg;
 
-   if (!load_acquire(&f->jit->shutdown))
+   if (!jit_is_shutdown(f->jit))
       (*tier->plugin.cgen)(f->jit, f->handle, tier->context);
 }
 
