@@ -150,6 +150,17 @@ void x_length_fail(int64_t left, int64_t right, int32_t dim, tree_t where)
    jit_msg(tree_loc(where), DIAG_FATAL, "%s", tb_get(tb));
 }
 
+void x_dir_fail(range_kind_t value, range_kind_t dir, tree_t where)
+{
+   assert(tree_kind(where) == T_RANGE);
+   assert(value == RANGE_TO || value == RANGE_DOWNTO);
+   assert(dir == RANGE_TO || dir == RANGE_DOWNTO);
+
+   const char *map[] = { "TO", "DOWNTO" };
+   jit_msg(tree_loc(where), DIAG_FATAL, "range direction of slice %s does "
+           "not match prefix %s", map[value], map[dir]);
+}
+
 void x_range_fail(int64_t value, int64_t left, int64_t right, int8_t dir,
                   tree_t where, tree_t hint)
 {
@@ -260,50 +271,6 @@ void x_func_wait(void)
    jit_msg(NULL, DIAG_FATAL, "cannot wait inside %s", what);
 }
 
-void x_instance_name(attr_kind_t kind, text_buf_t *tb)
-{
-   assert(kind == ATTR_INSTANCE_NAME || kind == ATTR_PATH_NAME);
-
-   jit_stack_trace_t *trace LOCAL = jit_stack_trace();
-   for (int i = 0; i < trace->count; i++) {
-      tree_t where = tree_from_object(trace->frames[i].object);
-      if (where == NULL)
-         continue;
-
-      switch (tree_kind(where)) {
-      case T_BLOCK:
-         {
-            tree_t hier = tree_decl(where, 0);
-            assert(tree_kind(hier) == T_HIER);
-
-            ident_t inst = tree_ident(hier);
-
-            if (kind == ATTR_PATH_NAME)
-               instance_name_to_path(tb, istr(inst));
-            else
-               tb_istr(tb, inst);
-         }
-         return;
-
-      case T_PACKAGE:
-      case T_PACK_BODY:
-      case T_PACK_INST:
-         {
-            tb_append(tb, ':');
-            tb_istr(tb, tree_ident(primary_unit_of(where)));
-            tb_replace(tb, '.', ':');
-            tb_downcase(tb);
-         }
-         return;
-
-      default:
-         break;
-      }
-   }
-
-   fatal_trace("cannot get instance name");
-}
-
 ////////////////////////////////////////////////////////////////////////////////
 // Entry point from interpreter or JIT compiled code
 
@@ -436,11 +403,12 @@ void __nvc_do_exit(jit_exit_t which, jit_anchor_t *anchor, jit_scalar_t *args,
 
    case JIT_EXIT_RESOLVE_SIGNAL:
       {
-         sig_shared_t *shared  = args[0].pointer;
-         jit_handle_t  handle  = args[1].integer;
-         void         *context = args[2].pointer;
-         int32_t       nlits   = args[3].integer;
-         int32_t       flags   = args[4].integer;
+         sig_shared_t *shared = args[0].pointer;
+
+         jit_handle_t  handle  = *(jit_handle_t *)(args[1].pointer + 0);
+         void         *context = *(void **)(args[1].pointer + 8);
+         int32_t       nlits   = *(int64_t *)(args[1].pointer + 16);
+         int32_t       flags   = *(int32_t *)(args[1].pointer + 24);
 
          x_resolve_signal(shared, handle, context, nlits, flags);
       }
@@ -683,6 +651,16 @@ void __nvc_do_exit(jit_exit_t which, jit_anchor_t *anchor, jit_scalar_t *args,
       }
       break;
 
+   case JIT_EXIT_DIR_FAIL:
+      {
+         range_kind_t value = args[0].integer;
+         range_kind_t dir   = args[1].integer;
+         tree_t       where = args[2].pointer;
+
+         x_dir_fail(value, dir, where);
+      }
+      break;
+
    case JIT_EXIT_NULL_DEREF:
       {
          tree_t where = args[0].pointer;
@@ -1034,3 +1012,30 @@ object_t *__nvc_get_object(const char *unit, ptrdiff_t offset)
 {
    return object_from_locus(ident_new(unit), offset, lib_load_handler);
 }
+
+DLLEXPORT
+void __nvc_pack(const uint8_t *src, int32_t size, jit_scalar_t *args)
+{
+   assert(size <= 64);  // TODO
+
+   uint64_t abits = 0, bbits = 0;
+
+   for (size_t i = 0; i < size; i++) {
+      abits = (abits << 1) | (src[i] & 1);
+      bbits = (bbits << 1) | ((src[i] >> 1) & 1);
+   }
+
+   args[0].integer = abits;
+   args[1].integer = bbits;
+}
+
+DLLEXPORT
+void __nvc_unpack(uint64_t abits, uint64_t bbits, jit_scalar_t *args)
+{
+   uint8_t *dest = args[0].pointer;
+   size_t size = args[1].integer;
+   uint8_t strength = args[2].integer;
+
+   for (size_t i = 0; i < size; i++, abits >>= 1, bbits >>= 1)
+      dest[size - i - 1] = (abits & 1) | ((bbits & 1) << 1) | strength;
+}
