@@ -86,6 +86,8 @@ typedef enum {
    MACRO_MEMSET,
    MACRO_REEXEC,
    MACRO_SADD,
+   MACRO_PACK,
+   MACRO_UNPACK,
 } jit_op_t;
 
 typedef enum {
@@ -172,6 +174,7 @@ typedef enum {
    JIT_EXIT_BIND_EXTERNAL,
    JIT_EXIT_SYSCALL,
    JIT_EXIT_PUT_CONVERSION,
+   JIT_EXIT_DIR_FAIL,
 } jit_exit_t;
 
 typedef uint16_t jit_reg_t;
@@ -351,6 +354,7 @@ typedef struct {
    uint8_t      *wptr;
    ihash_t      *labels;
    patch_list_t *patches;
+   uint8_t      *veneers;
    bool          overflow;
 } code_blob_t;
 
