@@ -63,7 +63,6 @@ void jit_set_silent(jit_t *j, bool silent);
 mspace_t *jit_get_mspace(jit_t *j);
 void jit_load_dll(jit_t *j, ident_t name);
 void jit_load_pack(jit_t *j, FILE *f);
-void jit_preload(jit_t *j);
 bool jit_exit_status(jit_t *j, int *status);
 void jit_reset_exit_status(jit_t *j);
 void jit_add_tier(jit_t *j, int threshold, const jit_plugin_t *plugin);
@@ -75,6 +74,7 @@ void jit_check_interrupt(jit_t *j);
 void jit_reset(jit_t *j);
 int32_t *jit_get_cover_mem(jit_t *j, int mintags);
 int32_t *jit_sync_cover_mem(jit_t *j, int mintags);
+bool jit_is_shutdown(jit_t *j);
 
 void *jit_mspace_alloc(size_t size) RETURNS_NONNULL;
 jit_stack_trace_t *jit_stack_trace(void);
