@@ -24,6 +24,7 @@
 #include "mir/mir-unit.h"
 #include "thread.h"
 #include "tree.h"
+#include "vlog/vlog-node.h"
 
 #include <assert.h>
 #include <ctype.h>
@@ -48,11 +49,8 @@ const char *mir_op_string(mir_op_t op)
       [MIR_OP_CMP] = "cmp",
       [MIR_OP_PHI] = "phi",
       [MIR_OP_AND] = "and",
-      [MIR_OP_NAND] = "nand",
       [MIR_OP_OR] = "or",
-      [MIR_OP_NOR] = "nor",
       [MIR_OP_XOR] = "xor",
-      [MIR_OP_XNOR] = "xnor",
       [MIR_OP_CONSUME] = "consume",
       [MIR_OP_SUB] = "sub",
       [MIR_OP_MUL] = "mul",
@@ -89,6 +87,7 @@ const char *mir_op_string(mir_op_t op)
       [MIR_OP_CONST_ARRAY] = "const array",
       [MIR_OP_CONST_REP] = "const rep",
       [MIR_OP_CONST_RECORD] = "const record",
+      [MIR_OP_CONST_VEC] = "const vec",
       [MIR_OP_ARRAY_REF] = "array ref",
       [MIR_OP_RECORD_REF] = "record ref",
       [MIR_OP_ADDRESS_OF] = "address of",
@@ -166,6 +165,11 @@ const char *mir_op_string(mir_op_t op)
       [MIR_OP_REFLECT_VALUE] = "reflect value",
       [MIR_OP_REFLECT_SUBTYPE] = "reflect subtype",
       [MIR_OP_DEBUG_OUT] = "debug out",
+      [MIR_OP_PACK] = "pack",
+      [MIR_OP_UNPACK] = "unpack",
+      [MIR_OP_BINARY] = "vector binary",
+      [MIR_OP_UNARY] = "vector unary",
+      [MIR_OP_DIR_CHECK] = "dir check",
    };
 
    return map[op];
@@ -322,8 +326,6 @@ static void mir_dump_one_type(mir_unit_t *mu, mir_type_t type)
 
    case MIR_TYPE_CLOSURE:
       printf("C<");
-      mir_dump_one_type(mu, td->u.closure.atype);
-      printf(" -> ");
       mir_dump_one_type(mu, td->u.closure.rtype);
       printf(">");
       break;
@@ -351,6 +353,14 @@ static void mir_dump_one_type(mir_unit_t *mu, mir_type_t type)
    case MIR_TYPE_OPAQUE:
       printf("?");
       break;
+
+   case MIR_TYPE_VEC2:
+      printf("Vec2%s<%u>", td->u.vec.issigned ? "S" : "", td->u.vec.size);
+      break;
+
+   case MIR_TYPE_VEC4:
+      printf("Vec4%s<%u>", td->u.vec.issigned ? "S" : "", td->u.vec.size);
+      break;
    }
 }
 
@@ -465,6 +475,46 @@ static int mir_dump_arg(mir_unit_t *mu, mir_value_t node, int nth,
    return mir_dump_value(mu, arg, cb, ctx);
 }
 
+static int mir_dump_strength(mir_unit_t *mu, mir_value_t value)
+{
+   assert(value.tag == MIR_TAG_ENUM);
+
+   static const char *names[] = {
+      "highz", "small", "medium", "weak", "large", "pull", "strong", "supply"
+   };
+   return color_printf("$green$(%s1,%s0)$$", names[STRENGTH1(value.id)],
+                       names[STRENGTH0(value.id)]);
+}
+
+static int mir_dump_vector_op(mir_unit_t *mu, mir_value_t value)
+{
+   assert(value.tag == MIR_TAG_ENUM);
+
+   switch (value.id) {
+   case MIR_VEC_BIT_AND:  return printf(" & ");
+   case MIR_VEC_BIT_OR:   return printf(" | ");
+   case MIR_VEC_BIT_XOR:  return printf(" ^ ");
+   case MIR_VEC_LOG_AND:  return printf(" && ");
+   case MIR_VEC_LOG_OR:   return printf(" || ");
+   case MIR_VEC_BIT_NOT:  return printf(" ~ ");
+   case MIR_VEC_LOG_NOT:  return printf(" ! ");
+   case MIR_VEC_LT:       return printf(" < ");
+   case MIR_VEC_LEQ:      return printf(" <= ");
+   case MIR_VEC_GT:       return printf(" > ");
+   case MIR_VEC_GEQ:      return printf(" >= ");
+   case MIR_VEC_LOG_EQ:   return printf(" == ");
+   case MIR_VEC_LOG_NEQ:  return printf(" != ");
+   case MIR_VEC_CASE_EQ:  return printf(" === ");
+   case MIR_VEC_CASE_NEQ: return printf(" !== ");
+   case MIR_VEC_ADD:      return printf(" + ");
+   case MIR_VEC_SUB:      return printf(" - ");
+   case MIR_VEC_MUL:      return printf(" * ");
+   case MIR_VEC_SLL:      return printf(" << ");
+   case MIR_VEC_SRL:      return printf(" >> ");
+   default: return printf(" <%d> ", value.id);
+   }
+}
+
 static int mir_dump_locus(object_t *obj)
 {
    int col = color_printf("$magenta$");
@@ -473,6 +523,10 @@ static int mir_dump_locus(object_t *obj)
    if (t != NULL)
       col += printf("%s@", tree_kind_str(tree_kind(t)));
 
+   vlog_node_t v = vlog_from_object(obj);
+   if (v != NULL)
+      col += printf("%s@", vlog_kind_str(vlog_kind(v)));
+
    col += color_printf("%p$$", obj);
    return col;
 }
@@ -577,7 +631,6 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
    case MIR_UNIT_THUNK:       printf("thunk"); break;
    case MIR_UNIT_PROPERTY:    printf("property"); break;
    case MIR_UNIT_PROTECTED:   printf("protected"); break;
-   case MIR_UNIT_PLACEHOLDER: printf("placeholder"); break;
    }
    color_printf("$$\n");
    if (mu->parent != NULL)
@@ -653,11 +706,8 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
          case MIR_OP_MUL:
          case MIR_OP_DIV:
          case MIR_OP_AND:
-         case MIR_OP_NAND:
          case MIR_OP_OR:
-         case MIR_OP_NOR:
          case MIR_OP_XOR:
-         case MIR_OP_XNOR:
          case MIR_OP_MOD:
          case MIR_OP_REM:
          case MIR_OP_EXP:
@@ -671,11 +721,8 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
                case MIR_OP_MUL:  col += printf(" * "); break;
                case MIR_OP_DIV:  col += printf(" / "); break;
                case MIR_OP_AND:  col += printf(" && "); break;
-               case MIR_OP_NAND: col += printf(" !& "); break;
                case MIR_OP_OR:   col += printf(" || "); break;
-               case MIR_OP_NOR:  col += printf(" !| "); break;
                case MIR_OP_XOR:  col += printf(" ^ "); break;
-               case MIR_OP_XNOR: col += printf(" !^ "); break;
                case MIR_OP_MOD:  col += printf(" %% "); break;
                case MIR_OP_REM:  col += printf(" %% "); break;
                case MIR_OP_EXP:  col += printf(" ** "); break;
@@ -750,6 +797,16 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
             }
             break;
 
+         case MIR_OP_CONST_VEC:
+            {
+               col += mir_dump_value(mu, result, cb, ctx);
+               col += printf(" := const 0x%"PRIx64", 0x%"PRIx64 , n->bits[0],
+                             n->bits[1]);
+               mir_dump_type(mu, col, n->type);
+               mir_dump_stamp(mu, n->type, n->stamp);
+            }
+            break;
+
          case MIR_OP_STORE:
          case MIR_OP_COPY:
          case MIR_OP_SET:
@@ -933,12 +990,11 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
                }
                col += printf("%s ", mir_op_string(n->op));
                col += mir_dump_arg(mu, result, 0, cb, ctx);
-               col += printf(" ");
                for (int i = 2; i < n->nargs; i++) {
-                  if (i > 2) col += printf(", ");
+                  col += printf("%s ", i > 2 ? "," : "");
                   col += mir_dump_arg(mu, result, i, cb, ctx);
                }
-               col += color_printf("locus ");
+               col += color_printf(" locus ");
                col += mir_dump_arg(mu, result, 1, cb, ctx);
                mir_dump_type(mu, col, n->type);
                mir_dump_stamp(mu, n->type, n->stamp);
@@ -1174,9 +1230,13 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
                col += printf(" := %s ", mir_op_string(n->op));
                col += mir_dump_value(mu, n->args[0], cb, ctx);
                ident_t link = mir_get_name(mu, n->args[1]);
-               mir_shape_t *s = mir_get_shape(mu->context, link);
-               ident_t name = s->slots[n->args[2].id].name;
-               col += color_printf(", $magenta$%s$$", istr(name));
+               if (link == NULL)
+                  col += color_printf(", $red$invalid$$");
+               else {
+                  mir_shape_t *s = mir_get_shape(mu->context, link);
+                  ident_t name = s->slots[n->args[2].id].name;
+                  col += color_printf(", $magenta$%s$$", istr(name));
+               }
                mir_dump_type(mu, col, n->type);
                mir_dump_stamp(mu, n->type, n->stamp);
             }
@@ -1264,6 +1324,17 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
             }
             break;
 
+         case MIR_OP_DIR_CHECK:
+            {
+               printf("%s ", mir_op_string(n->op));
+               mir_dump_arg(mu, result, 0, cb, ctx);
+               printf(" == ");
+               mir_dump_arg(mu, result, 1, cb, ctx);
+               printf(" locus ");
+               mir_dump_arg(mu, result, 2, cb, ctx);
+            }
+            break;
+
          case MIR_OP_LENGTH_CHECK:
             {
                printf("%s left ", mir_op_string(n->op));
@@ -1644,6 +1715,58 @@ void mir_annotate(mir_unit_t *mu, const mir_annotate_t *cb, void *ctx)
             }
             break;
 
+         case MIR_OP_UNPACK:
+            {
+               if (!mir_is_null(n->type)) {
+                  col += mir_dump_value(mu, result, cb, ctx);
+                  col += printf(" := ");
+               }
+               col += printf("%s ", mir_op_string(n->op));
+               col += mir_dump_arg(mu, result, 0, cb, ctx);
+               col += printf(" strength ");
+               col += mir_dump_strength(mu, mir_get_arg(mu, result, 1));
+               if (n->nargs > 2) {
+                  col += printf(" into ");
+                  col += mir_dump_arg(mu, result, 2, cb, ctx);
+               }
+               mir_dump_type(mu, col, n->type);
+               mir_dump_stamp(mu, n->type, n->stamp);
+            }
+            break;
+
+         case MIR_OP_PACK:
+            {
+               col += mir_dump_value(mu, result, cb, ctx);
+               col += printf(" := %s ", mir_op_string(n->op));
+               col += mir_dump_arg(mu, result, 0, cb, ctx);
+               mir_dump_type(mu, col, n->type);
+               mir_dump_stamp(mu, n->type, n->stamp);
+            }
+            break;
+
+         case MIR_OP_BINARY:
+            {
+               col += mir_dump_value(mu, result, cb, ctx);
+               col += printf(" := vector ");
+               col += mir_dump_arg(mu, result, 1, cb, ctx);
+               col += mir_dump_vector_op(mu, n->args[0]);
+               col += mir_dump_arg(mu, result, 2, cb, ctx);
+               mir_dump_type(mu, col, n->type);
+               mir_dump_stamp(mu, n->type, n->stamp);
+            }
+            break;
+
+         case MIR_OP_UNARY:
+            {
+               col += mir_dump_value(mu, result, cb, ctx);
+               col += printf(" := vector");
+               col += mir_dump_vector_op(mu, n->args[0]);
+               col += mir_dump_arg(mu, result, 1, cb, ctx);
+               mir_dump_type(mu, col, n->type);
+               mir_dump_stamp(mu, n->type, n->stamp);
+            }
+            break;
+
 #ifdef DEBUG
          case MIR_OP_COMMENT:
             {
