@@ -19,6 +19,7 @@
 #define _MIR_NODE_H
 
 #include "prim.h"
+#include "diag.h"
 #include "util.h"
 
 #include <limits.h>
@@ -129,7 +130,6 @@ typedef enum {
    MIR_OP_CMP,
    MIR_OP_PHI,
    MIR_OP_AND,
-   MIR_OP_NAND,
    MIR_OP_SUB,
    MIR_OP_MUL,
    MIR_OP_DIV,
@@ -137,9 +137,7 @@ typedef enum {
    MIR_OP_MOD,
    MIR_OP_EXP,
    MIR_OP_OR,
-   MIR_OP_NOR,
    MIR_OP_XOR,
-   MIR_OP_XNOR,
    MIR_OP_CONSUME,
    MIR_OP_FCALL,
    MIR_OP_PCALL,
@@ -170,6 +168,7 @@ typedef enum {
    MIR_OP_CONST_ARRAY,
    MIR_OP_CONST_REP,
    MIR_OP_CONST_RECORD,
+   MIR_OP_CONST_VEC,
    MIR_OP_ADDRESS_OF,
    MIR_OP_ARRAY_REF,
    MIR_OP_VAR_UPREF,
@@ -247,6 +246,11 @@ typedef enum {
    MIR_OP_REFLECT_VALUE,
    MIR_OP_REFLECT_SUBTYPE,
    MIR_OP_DEBUG_OUT,
+   MIR_OP_PACK,
+   MIR_OP_UNPACK,
+   MIR_OP_BINARY,
+   MIR_OP_UNARY,
+   MIR_OP_DIR_CHECK,
 } mir_op_t;
 
 typedef enum {
@@ -268,6 +272,8 @@ typedef enum {
    MIR_TYPE_TRIGGER,
    MIR_TYPE_CONVERSION,
    MIR_TYPE_OPAQUE,
+   MIR_TYPE_VEC2,
+   MIR_TYPE_VEC4,
 } mir_class_t;
 
 typedef struct {
@@ -304,6 +310,29 @@ typedef enum {
    MIR_MEM_TOP,
 } mir_mem_t;
 
+typedef enum {
+   MIR_VEC_BIT_AND,
+   MIR_VEC_BIT_OR,
+   MIR_VEC_BIT_XOR,
+   MIR_VEC_BIT_NOT,
+   MIR_VEC_LOG_AND,
+   MIR_VEC_LOG_OR,
+   MIR_VEC_LOG_NOT,
+   MIR_VEC_LT,
+   MIR_VEC_LEQ,
+   MIR_VEC_GT,
+   MIR_VEC_GEQ,
+   MIR_VEC_LOG_EQ,
+   MIR_VEC_LOG_NEQ,
+   MIR_VEC_CASE_EQ,
+   MIR_VEC_CASE_NEQ,
+   MIR_VEC_ADD,
+   MIR_VEC_SUB,
+   MIR_VEC_MUL,
+   MIR_VEC_SLL,
+   MIR_VEC_SRL,
+} mir_vec_op_t;
+
 #define MIR_APPEND UINT_MAX
 
 mir_type_t mir_int_type(mir_unit_t *mu, int64_t low, int64_t high);
@@ -315,6 +344,7 @@ mir_type_t mir_trigger_type(mir_unit_t *mu);
 mir_type_t mir_self_type(mir_unit_t *mu);
 mir_type_t mir_bool_type(mir_unit_t *mu);
 mir_type_t mir_time_type(mir_unit_t *mu);
+mir_type_t mir_logic_type(mir_unit_t *mu);
 mir_type_t mir_double_type(mir_unit_t *mu);
 mir_type_t mir_char_type(mir_unit_t *mu);
 mir_type_t mir_string_type(mir_unit_t *mu);
@@ -330,6 +360,8 @@ mir_type_t mir_record_type(mir_unit_t *mu, ident_t name,
 mir_type_t mir_closure_type(mir_unit_t *mu, mir_type_t atype, mir_type_t rtype);
 mir_type_t mir_resolution_type(mir_unit_t *mu, mir_type_t base);
 mir_type_t mir_file_type(mir_unit_t *mu, mir_type_t base);
+mir_type_t mir_vec2_type(mir_unit_t *mu, int size, bool issigned);
+mir_type_t mir_vec4_type(mir_unit_t *mu, int size, bool issigned);
 
 mir_type_t mir_get_base(mir_unit_t *mu, mir_type_t type);
 mir_type_t mir_get_elem(mir_unit_t *mu, mir_type_t type);
@@ -376,6 +408,8 @@ mir_stamp_t mir_get_stamp(mir_unit_t *mu, mir_value_t value);
 mir_mem_t mir_get_mem(mir_unit_t *mu, mir_value_t value);
 ident_t mir_get_name(mir_unit_t *mu, mir_value_t value);
 object_t *mir_get_locus(mir_unit_t *mu, mir_value_t value);
+void mir_get_bits(mir_unit_t *mu, mir_value_t value, uint64_t *abits,
+                  uint64_t *bbits);
 mir_var_flags_t mir_get_var_flags(mir_unit_t *mu, mir_value_t value);
 mir_type_t mir_get_var_type(mir_unit_t *mu, mir_value_t value);
 void mir_set_input(mir_unit_t *mu, mir_value_t phi, unsigned nth,
@@ -386,7 +420,9 @@ void mir_set_result(mir_unit_t *mu, mir_type_t type);
 mir_type_t mir_get_result(mir_unit_t *mu);
 
 bool mir_is_integral(mir_unit_t *mu, mir_value_t value);
+bool mir_is_numeric(mir_unit_t *mu, mir_value_t value);
 bool mir_is_scalar(mir_unit_t *mu, mir_value_t value);
+bool mir_is_vector(mir_unit_t *mu, mir_value_t value);
 bool mir_is_bool(mir_unit_t *mu, mir_value_t value);
 bool mir_is_time(mir_unit_t *mu, mir_value_t value);
 bool mir_is_signal(mir_unit_t *mu, mir_value_t value);
@@ -404,10 +440,25 @@ void mir_comment(mir_unit_t *mu, const char *fmt, ...);
 #define mir_comment(mu, fmt, ...)
 #endif
 
+typedef struct {
+   mir_unit_t *mu;
+   loc_t       loc;
+} mir_saved_loc_t;
+
+mir_saved_loc_t _mir_push_debug_info(mir_unit_t *mu, const loc_t *loc);
+
+__attribute__((always_inline))
+static inline void _mir_pop_debug_info(const mir_saved_loc_t *sl)
+{
+   mir_set_loc(sl->mu, &sl->loc);
+}
+
 // Constants
 mir_value_t mir_enum(unsigned value);
 mir_value_t mir_const(mir_unit_t *mu, mir_type_t type, int64_t value);
 mir_value_t mir_const_real(mir_unit_t *mu, mir_type_t type, double value);
+mir_value_t mir_const_vec(mir_unit_t *mu, mir_type_t type, uint64_t abits,
+                          uint64_t bbits);
 mir_value_t mir_const_array(mir_unit_t *mu, mir_type_t type,
                             const mir_value_t *values, size_t count);
 mir_value_t mir_const_string(mir_unit_t *mu, const char *str);
@@ -452,15 +503,21 @@ mir_value_t mir_build_trap_neg(mir_unit_t *mu, mir_type_t type,
 
 // Logical
 mir_value_t mir_build_and(mir_unit_t *mu, mir_value_t left, mir_value_t right);
-mir_value_t mir_build_nand(mir_unit_t *mu, mir_value_t left, mir_value_t right);
 mir_value_t mir_build_or(mir_unit_t *mu, mir_value_t left, mir_value_t right);
-mir_value_t mir_build_nor(mir_unit_t *mu, mir_value_t left, mir_value_t right);
 mir_value_t mir_build_xor(mir_unit_t *mu, mir_value_t left, mir_value_t right);
-mir_value_t mir_build_xnor(mir_unit_t *mu, mir_value_t left, mir_value_t right);
 mir_value_t mir_build_not(mir_unit_t *mu, mir_value_t value);
 mir_value_t mir_build_cmp(mir_unit_t *mu, mir_cmp_t cmp, mir_value_t left,
                           mir_value_t right);
 
+// Vectors
+mir_value_t mir_build_pack(mir_unit_t *mu, mir_type_t type, mir_value_t arg);
+mir_value_t mir_build_unpack(mir_unit_t *mu, mir_value_t vec, uint8_t strength,
+                             mir_value_t dest);
+mir_value_t mir_build_binary(mir_unit_t *mu, mir_vec_op_t op, mir_type_t type,
+                             mir_value_t left, mir_value_t right);
+mir_value_t mir_build_unary(mir_unit_t *mu, mir_vec_op_t op, mir_type_t type,
+                            mir_value_t arg);
+
 // Memory
 void mir_build_store(mir_unit_t *mu, mir_value_t dest, mir_value_t src);
 mir_value_t mir_build_load(mir_unit_t *mu, mir_value_t value);
@@ -525,6 +582,8 @@ void mir_build_range_check(mir_unit_t *mu, mir_value_t value, mir_value_t left,
 void mir_build_index_check(mir_unit_t *mu, mir_value_t value, mir_value_t left,
                            mir_value_t right, mir_value_t dir,
                            mir_value_t locus, mir_value_t hint);
+void mir_build_dir_check(mir_unit_t *mu, mir_value_t value, mir_value_t dir,
+                         mir_value_t locus);
 void mir_build_null_check(mir_unit_t *mu, mir_value_t ptr, mir_value_t locus);
 void mir_build_zero_check(mir_unit_t *mu, mir_value_t value, mir_value_t locus);
 void mir_build_length_check(mir_unit_t *mu, mir_value_t llen, mir_value_t rlen,
