@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2024  Nick Gasson
+//  Copyright (C) 2024-2025  Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -53,8 +53,9 @@ typedef struct {
 #endif
 } cfg_block_t;
 
-typedef unsigned valnum_t;
+typedef uint32_t valnum_t;
 #define VN_INVALID UINT_MAX
+#define VN_CONST   0x80000000
 
 typedef struct {
    mir_value_t value;
@@ -132,6 +133,9 @@ static cfg_block_t *mir_get_cfg(mir_unit_t *mu)
    cfg_block_t *cfg = xcalloc_array(mu->blocks.count, sizeof(cfg_block_t));
    cfg[0].entry = true;
 
+   if (mir_get_kind(mu) == MIR_UNIT_PROCESS && mu->blocks.count > 1)
+      cfg[1].entry = true;
+
    for (int i = 0; i < mu->blocks.count; i++) {
       const block_data_t *bd = &(mu->blocks.items[i]);
       mir_block_t this = { .tag = MIR_TAG_BLOCK, .id = i };
@@ -313,11 +317,11 @@ static valnum_t gvn_get_value(mir_value_t value, gvn_state_t *gvn)
 {
    switch (value.tag) {
    case MIR_TAG_NODE:
-      assert(gvn->nodevn[value.id] != VN_INVALID);
       return gvn->nodevn[value.id];
    case MIR_TAG_PARAM:
-      assert(gvn->paramvn[value.id] != VN_INVALID);
       return gvn->paramvn[value.id];
+   case MIR_TAG_CONST:
+      return VN_CONST + value.id;
    default:
       should_not_reach_here();
    }
@@ -326,16 +330,18 @@ static valnum_t gvn_get_value(mir_value_t value, gvn_state_t *gvn)
 static uint32_t gvn_hash_node(mir_unit_t *mu, const node_data_t *n,
                               gvn_state_t *gvn)
 {
-   if (n->op == MIR_OP_CONST)
-      return mix_bits_64(n->iconst) & 0xffffffff;
-   else if (n->op == MIR_OP_CONST_REAL)
-      return mix_bits_64(FLOAT_BITS(n->dconst)) & 0xffffffff;
-   else if (n->op == MIR_OP_LOCUS)
-      return mix_bits_64(n->locus) & 0xffffffff;
-   else {
-      uint32_t hash = knuth_hash(n->op) + knuth_hash(n->nargs);
-      hash ^= mir_type_data(mu, n->type)->hash;
+   uint32_t hash = knuth_hash(n->op) + knuth_hash(n->nargs);
+   hash ^= mir_type_data(mu, n->type)->hash;
 
+   if (n->op == MIR_OP_CONST)
+      return hash ^ mix_bits_64(n->iconst);
+   else if (n->op == MIR_OP_CONST_REAL)
+      return hash ^ mix_bits_64(FLOAT_BITS(n->dconst));
+   else if (n->op == MIR_OP_CONST_VEC)
+      return hash ^ mix_bits_64(n->bits[0]) ^ mix_bits_64(n->bits[1]);
+   else if (n->op == MIR_OP_LOCUS)
+      return hash ^ mix_bits_64(n->locus);
+   else {
       const mir_value_t *args = mir_get_args(mu, n);
       for (int i = 0; i < n->nargs; i++) {
          hash <<= 1;   // Argument order should affect hash
@@ -352,12 +358,14 @@ static uint32_t gvn_hash_node(mir_unit_t *mu, const node_data_t *n,
 static bool gvn_compare(mir_unit_t *mu, const node_data_t *a,
                         const node_data_t *b, gvn_state_t *gvn)
 {
-   if (a->op != b->op || a->nargs != b->nargs)
+   if (a->op != b->op || a->nargs != b->nargs || !mir_equals(a->type, b->type))
       return false;
    else if (a->op == MIR_OP_CONST)
       return a->iconst == b->iconst;
    else if (a->op == MIR_OP_CONST_REAL)
       return a->dconst == b->dconst;
+   else if (a->op == MIR_OP_CONST_VEC)
+      return a->bits[0] == b->bits[0] && a->bits[1] == b->bits[1];
    else if (a->op == MIR_OP_LOCUS)
       return a->locus == b->locus;
 
@@ -409,6 +417,31 @@ static void gvn_generic(mir_unit_t *mu, mir_value_t node, mir_block_t block,
    opt->gvn->nodevn[node.id] = gvn_new_value(node, opt->gvn);
 }
 
+static void gvn_sub(mir_unit_t *mu, mir_value_t node, mir_block_t block,
+                    mir_optim_t *opt)
+{
+   node_data_t *n = mir_node_data(mu, node);
+   assert(n->nargs == 2);
+
+   // (X + Y) - Y ==> X
+
+   mir_value_t left = n->args[0], right = n->args[1];
+   if (left.tag == MIR_TAG_NODE) {
+      node_data_t *nl = mir_node_data(mu, left);
+      if (nl->op == MIR_OP_ADD) {
+         valnum_t left0_vn = gvn_get_value(nl->args[0], opt->gvn);
+         valnum_t left1_vn = gvn_get_value(nl->args[1], opt->gvn);
+         valnum_t right_vn = gvn_get_value(right, opt->gvn);
+         if (left1_vn == right_vn) {
+            opt->gvn->nodevn[node.id] = left0_vn;
+            return;
+         }
+      }
+   }
+
+   gvn_generic(mu, node, block, opt);
+}
+
 static void gvn_commutative(mir_unit_t *mu, mir_value_t node, mir_block_t block,
                             mir_optim_t *opt, int first)
 {
@@ -494,13 +527,31 @@ static void gvn_phi(mir_unit_t *mu, mir_value_t node, mir_block_t block,
    valnum_t vn = gvn_get_value(args[1], opt->gvn);
 
    for (int i = 3; i < n->nargs; i += 2) {
-      if (!gvn_tracked(args[i]) || gvn_get_value(args[i], opt->gvn) != vn)
+      if (gvn_get_value(args[i], opt->gvn) != vn)
          return;
    }
 
    opt->gvn->nodevn[node.id] = vn;
 }
 
+static void gvn_unpack(mir_unit_t *mu, mir_value_t node, mir_block_t block,
+                       mir_optim_t *opt)
+{
+   const node_data_t *n = mir_node_data(mu, node);
+   if (mir_is_null(n->type))
+      return;   // Unpack into memory
+
+   gvn_generic(mu, node, block, opt);
+}
+
+static void gvn_load(mir_unit_t *mu, mir_value_t node, mir_block_t block,
+                     mir_optim_t *opt)
+{
+   // TODO: check if stamp is const or pointer
+
+   opt->gvn->nodevn[node.id] = gvn_new_value(node, opt->gvn);
+}
+
 static void gvn_visit_block(mir_unit_t *mu, mir_block_t block,
                             mir_optim_t *opt)
 {
@@ -526,12 +577,27 @@ static void gvn_visit_block(mir_unit_t *mu, mir_block_t block,
          gvn_commutative(mu, node, block, opt, 0);
          break;
       case MIR_OP_SUB:
+         gvn_sub(mu, node, block, opt);
+         break;
       case MIR_OP_NOT:
       case MIR_OP_RESOLVED:
       case MIR_OP_SELECT:
       case MIR_OP_REM:
       case MIR_OP_LINK_PACKAGE:
+      case MIR_OP_LINK_VAR:
+      case MIR_OP_PACKAGE_INIT:
       case MIR_OP_LOCUS:
+      case MIR_OP_UARRAY_LEN:
+      case MIR_OP_RANGE_LENGTH:
+      case MIR_OP_CAST:
+      case MIR_OP_PACK:
+      case MIR_OP_VAR_UPREF:
+      case MIR_OP_CONST_VEC:
+      case MIR_OP_BINARY:
+      case MIR_OP_UNARY:
+      case MIR_OP_CLOSURE:
+      case MIR_OP_CONTEXT_UPREF:
+      case MIR_OP_UNWRAP:
          gvn_generic(mu, node, block, opt);
          break;
       case MIR_OP_AND:
@@ -541,6 +607,16 @@ static void gvn_visit_block(mir_unit_t *mu, mir_block_t block,
       case MIR_OP_PHI:
          gvn_phi(mu, node, block, opt);
          break;
+      case MIR_OP_UNPACK:
+         gvn_unpack(mu, node, block, opt);
+         break;
+      case MIR_OP_LOAD:
+         gvn_load(mu, node, block, opt);
+         break;
+      case MIR_OP_INIT_SIGNAL:
+      case MIR_OP_PORT_CONVERSION:
+         opt->gvn->nodevn[node.id] = gvn_new_value(node, opt->gvn);
+         break;
       default:
          break;
       }
@@ -582,7 +658,7 @@ static void mir_do_gvn(mir_unit_t *mu, mir_optim_t *opt)
          if (args[j].tag == MIR_TAG_NODE) {
             const valnum_t vn = opt->gvn->nodevn[args[j].id];
             if (unlikely(vn == VN_INVALID)) {
-               mir_dump(mu);
+               mir_dump_optim(mu, opt);
                fatal_trace("node %%%d has no value number", args[j].id);
             }
             else if (!mir_equals(opt->gvn->canon[vn], args[j]))
@@ -623,7 +699,7 @@ static void dce_visit_block(mir_unit_t *mu, mir_block_t block, mir_optim_t *opt)
             opt->dce->uses[arg.id] = saturate_add(opt->dce->uses[arg.id], 1);
       }
 
-      if (mir_is_null(n->type))
+      if (mir_is_null(n->type) || n->op == MIR_OP_STORE)
          opt->dce->uses[node.id] = 1;   // Does not produce value
    }
 
@@ -775,8 +851,12 @@ void mir_optimise(mir_unit_t *mu, mir_pass_t passes)
 {
    mir_optim_t opt = {};
 
-   opt.cfg = mir_get_cfg(mu);
-   mir_dominator_tree(mu, opt.cfg);
+   const mir_pass_t need_cfg = MIR_PASS_GVN | MIR_PASS_DCE;
+
+   if (passes & need_cfg) {
+      opt.cfg = mir_get_cfg(mu);
+      mir_dominator_tree(mu, opt.cfg);
+   }
 
    if (passes & MIR_PASS_GVN)
       mir_do_gvn(mu, &opt);
@@ -784,5 +864,6 @@ void mir_optimise(mir_unit_t *mu, mir_pass_t passes)
    if (passes & MIR_PASS_DCE)
       mir_do_dce(mu, &opt);
 
-   mir_free_cfg(mu, opt.cfg);
+   if (passes & need_cfg)
+      mir_free_cfg(mu, opt.cfg);
 }
