@@ -625,7 +625,8 @@ static void import_nor(mir_unit_t *mu, mir_import_t *imp, int op)
    mir_value_t left  = imp->map[vcode_get_arg(op, 0)];
    mir_value_t right = imp->map[vcode_get_arg(op, 1)];
 
-   imp->map[vcode_get_result(op)] = mir_build_nor(mu, left, right);
+   mir_value_t or = mir_build_or(mu, left, right);
+   imp->map[vcode_get_result(op)] = mir_build_not(mu, or);
 }
 
 static void import_xor(mir_unit_t *mu, mir_import_t *imp, int op)
@@ -641,7 +642,8 @@ static void import_xnor(mir_unit_t *mu, mir_import_t *imp, int op)
    mir_value_t left  = imp->map[vcode_get_arg(op, 0)];
    mir_value_t right = imp->map[vcode_get_arg(op, 1)];
 
-   imp->map[vcode_get_result(op)] = mir_build_xnor(mu, left, right);
+   mir_value_t xor = mir_build_xor(mu, left, right);
+   imp->map[vcode_get_result(op)] = mir_build_not(mu, xor);
 }
 
 static void import_and(mir_unit_t *mu, mir_import_t *imp, int op)
@@ -657,7 +659,8 @@ static void import_nand(mir_unit_t *mu, mir_import_t *imp, int op)
    mir_value_t left  = imp->map[vcode_get_arg(op, 0)];
    mir_value_t right = imp->map[vcode_get_arg(op, 1)];
 
-   imp->map[vcode_get_result(op)] = mir_build_nand(mu, left, right);
+   mir_value_t and = mir_build_and(mu, left, right);
+   imp->map[vcode_get_result(op)] = mir_build_not(mu, and);
 }
 
 static void import_event(mir_unit_t *mu, mir_import_t *imp, int op)
@@ -879,11 +882,20 @@ static void import_index_check(mir_unit_t *mu, mir_import_t *imp, int op)
    mir_value_t right = imp->map[vcode_get_arg(op, 2)];
    mir_value_t dir   = imp->map[vcode_get_arg(op, 3)];
    mir_value_t locus = imp->map[vcode_get_arg(op, 4)];
-   mir_value_t hint  = imp->map[vcode_get_arg(op, 5)];;
+   mir_value_t hint  = imp->map[vcode_get_arg(op, 5)];
 
    mir_build_index_check(mu, value, left, right, dir, locus, hint);
 }
 
+static void import_dir_check(mir_unit_t *mu, mir_import_t *imp, int op)
+{
+   mir_value_t value = imp->map[vcode_get_arg(op, 0)];
+   mir_value_t dir   = imp->map[vcode_get_arg(op, 1)];
+   mir_value_t locus = imp->map[vcode_get_arg(op, 2)];
+
+   mir_build_dir_check(mu, value, dir, locus);
+}
+
 static void import_alias_signal(mir_unit_t *mu, mir_import_t *imp, int op)
 {
    mir_value_t signal = imp->map[vcode_get_arg(op, 0)];
@@ -971,10 +983,6 @@ static void import_closure(mir_unit_t *mu, mir_import_t *imp, int op)
    ident_t func = vcode_get_func(op);
 
    mir_type_t atype = MIR_NULL_TYPE;
-   vcode_type_t vatype = vcode_get_type(op);
-   if (vatype != VCODE_INVALID_TYPE)
-      atype = import_type(mu, imp, vatype);
-
    vcode_reg_t result = vcode_get_result(op);
    mir_type_t rtype = import_type(mu, imp, vtype_base(vcode_reg_type(result)));
 
@@ -1606,6 +1614,9 @@ static void import_block(mir_unit_t *mu, mir_import_t *imp)
       case VCODE_OP_INDEX_CHECK:
          import_index_check(mu, imp, i);
          break;
+      case VCODE_OP_DIR_CHECK:
+         import_dir_check(mu, imp, i);
+         break;
       case VCODE_OP_ALIAS_SIGNAL:
          import_alias_signal(mu, imp, i);
          break;
@@ -1774,7 +1785,6 @@ mir_unit_t *mir_import(mir_context_t *mc, vcode_unit_t vu)
       [VCODE_UNIT_THUNK] = MIR_UNIT_THUNK,
       [VCODE_UNIT_PROTECTED] = MIR_UNIT_PROTECTED,
       [VCODE_UNIT_PROPERTY] = MIR_UNIT_PROPERTY,
-      [VCODE_UNIT_SHAPE] = MIR_UNIT_PLACEHOLDER,
    };
 
    const mir_unit_kind_t kind = kind_map[vcode_unit_kind(vu)];
