@@ -61,7 +61,7 @@
 #if !defined HAVE_LLVM || !defined SYSTEM_CC
 #define DEFAULT_JIT true
 #else
-#define DEFAULT_JIT false
+#define DEFAULT_JIT true
 #endif
 
 typedef struct {
@@ -109,7 +109,7 @@ static int scan_cmd(int start, int argc, char **argv)
 {
    const char *commands[] = {
       "-a", "-e", "-r", "-c", "--dump", "--make", "--syntax", "--list",
-      "--init", "--install", "--print-deps", "--aotgen", "--do", "-i",
+      "--init", "--install", "--print-deps", "--do", "-i",
       "--cover-export", "--preprocess", "--gui", "--cover-merge",
       "--cover-report",
    };
@@ -416,6 +416,7 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
       { "verbose",         no_argument,       0, 'V' },
       { "no-save",         no_argument,       0, 'N' },
       { "jit",             no_argument,       0, 'j' },
+      { "precompile",      no_argument,       0, 'p' },
       { "no-collapse",     no_argument,       0, 'C' },
       { "trace",           no_argument,       0, 't' },
       { 0, 0, 0, 0 }
@@ -461,6 +462,9 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
       case 'j':
          use_jit = true;
          break;
+      case 'p':
+         use_jit = false;
+         break;
       case 'g':
          parse_generic(optarg);
          break;
@@ -590,10 +594,8 @@ static int elaborate(int argc, char **argv, cmd_state_t *state)
       // Must discard current JIT state to load AOT library later
       model_free(state->model);
       jit_free(state->jit);
-      mir_context_free(state->mir);
       state->jit = NULL;
       state->model = NULL;
-      state->mir = NULL;
    }
 
    argc -= next_cmd - 1;
@@ -632,14 +634,16 @@ static int parse_int(const char *str)
    return n;
 }
 
-static bool parse_on_off(const char *str)
+static ieee_warnings_t parse_ieee_warnings(const char *str)
 {
-   if (strcasecmp(str, "on") == 0)
-      return true;
-   else if (strcasecmp(str, "off") == 0)
-      return false;
-
-   fatal("specifiy 'on' or 'off' instead of '%s'", str);
+   if (strcasecmp(str, "off") == 0)
+      return IEEE_WARNINGS_OFF;
+   else if (strcasecmp(str, "on") == 0)
+      return IEEE_WARNINGS_ON;
+   else if (strcasecmp(str, "off-at-0") == 0)
+      return IEEE_WARNINGS_OFF_AT_0;
+   else
+      fatal("specify 'on', 'off' or 'off-at-0' instead of '%s'", str);
 }
 
 static vhdl_severity_t parse_severity(const char *str)
@@ -687,10 +691,6 @@ static jit_t *get_jit(unit_registry_t *ur, mir_context_t *mc)
 {
    jit_t *jit = jit_new(ur, mc);
 
-#ifdef HAVE_LLVM
-   jit_preload(jit);
-#endif
-
 #if defined HAVE_LLVM && 1
    jit_register_llvm_plugin(jit);
 #elif defined ARCH_X86_64 && 0
@@ -753,6 +753,26 @@ static void emit_coverage(const unit_meta_t *meta, jit_t *j, cover_data_t *db)
    fbuf_close(f, NULL);
 }
 
+static void enable_ieee_warnings_cb(rt_model_t *m, void *ctx)
+{
+   cmd_state_t *state = ctx;
+
+   assert(opt_get_int(OPT_IEEE_WARNINGS) == IEEE_WARNINGS_OFF_AT_0);
+   opt_set_int(OPT_IEEE_WARNINGS, IEEE_WARNINGS_ON);
+
+   ident_t pkg_name = ident_new("NVC.SIM_PKG");
+   ident_t var_name = ident_new("IEEE_NO_WARNING");
+
+   jit_handle_t handle = jit_lazy_compile(state->jit, pkg_name);
+   jit_link(state->jit, handle);
+
+   uint8_t *ptr = jit_get_frame_var(state->jit, handle, var_name);
+   assert(ptr != NULL);
+   assert(*ptr == 1);
+
+   *ptr = 0;
+}
+
 static int run_cmd(int argc, char **argv, cmd_state_t *state)
 {
    static struct option long_options[] = {
@@ -857,7 +877,7 @@ static int run_cmd(int argc, char **argv, cmd_state_t *state)
          break;
       case 'I':
          {
-            const bool on = parse_on_off(optarg);
+            const ieee_warnings_t on = parse_ieee_warnings(optarg);
 
             // TODO: add an unconditional warning after 1.16
             if (state->jit != NULL && opt_get_int(OPT_IEEE_WARNINGS) != on)
@@ -981,6 +1001,10 @@ static int run_cmd(int argc, char **argv, cmd_state_t *state)
    if (dumper != NULL)
       wave_dumper_restart(dumper, state->model, state->jit);
 
+   if (opt_get_int(OPT_IEEE_WARNINGS) == IEEE_WARNINGS_OFF_AT_0)
+      model_set_global_cb(state->model, RT_END_TIME_STEP,
+                          enable_ieee_warnings_cb, state);
+
    model_run(state->model, stop_time);
 
    set_ctrl_c_handler(NULL, NULL);
@@ -1383,49 +1407,6 @@ static int dump_cmd(int argc, char **argv, cmd_state_t *state)
    return argc > 1 ? process_command(argc, argv, state) : EXIT_SUCCESS;
 }
 
-#if ENABLE_LLVM
-static int aotgen_cmd(int argc, char **argv, cmd_state_t *state)
-{
-   static struct option long_options[] = {
-      { 0, 0, 0, 0 }
-   };
-
-   const char *outfile = "preload." DLL_EXT;
-
-   const int next_cmd = scan_cmd(2, argc, argv);
-   int c, index = 0;
-   const char *spec = ":o:VO:";
-   while ((c = getopt_long(next_cmd, argv, spec, long_options, &index)) != -1) {
-      switch (c) {
-      case 0: break;  // Set a flag
-      case 'V':
-         opt_set_int(OPT_VERBOSE, 1);
-         break;
-      case 'O':
-         opt_set_int(OPT_OPTIMISE, parse_optimise_level(optarg));
-         break;
-      case 'o': outfile = optarg; break;
-      case '?': bad_option("aotgen", argv);
-      case ':': missing_argument("aotgen", argv);
-      default: abort();
-      }
-   }
-
-   const int count = next_cmd - optind;
-   aotgen(outfile, argv + optind, count);
-
-   argc -= next_cmd - 1;
-   argv += next_cmd - 1;
-
-   return argc > 1 ? process_command(argc, argv, state) : EXIT_SUCCESS;
-}
-#else
-static int aotgen_cmd(int argc, char **argv, cmd_state_t *state)
-{
-   fatal("$bold$--aotgen$$ not supported without LLVM");
-}
-#endif
-
 static int do_cmd(int argc, char **argv, cmd_state_t *state)
 {
    static struct option long_options[] = {
@@ -1594,6 +1575,23 @@ static uint32_t parse_cover_print_spec(char *str)
    return mask;
 }
 
+static merge_mode_t parse_cover_merge_mode(char *str)
+{
+   if (strcmp(str, "intersect") == 0)
+      return MERGE_INTERSECT;
+   else if (strcmp(str, "union") == 0)
+      return MERGE_UNION;
+   else {
+      diag_t *d = diag_new(DIAG_FATAL, NULL);
+      diag_printf(d, "invalid option: '%s' for $bold$--merge-mode$$", str);
+      diag_hint(d, NULL, "valid options are: 'union', 'intersect'");
+      diag_emit(d);
+      fatal_exit(EXIT_FAILURE);
+   }
+
+   return 0;
+}
+
 static int coverage_cmd(int argc, char **argv, cmd_state_t *state)
 {
    warnf("the $bold$-c$$ sub-command is deprecated, use $bold$--cover-report$$ "
@@ -1665,7 +1663,7 @@ static int coverage_cmd(int argc, char **argv, cmd_state_t *state)
          if (i == optind)
             cover = cover_read_items(f, rpt_mask);
          else
-            cover_merge_items(f, cover);
+            cover_merge_items(f, cover, MERGE_UNION);
       }
       else
          fatal("Could not open coverage database: %s", argv[i]);
@@ -1774,7 +1772,8 @@ static int gui_cmd(int argc, char **argv, cmd_state_t *state)
 #endif
 
 static cover_data_t *merge_coverage_files(int argc, int next_cmd, char **argv,
-                                          cover_mask_t rpt_mask)
+                                          cover_mask_t rpt_mask,
+                                          merge_mode_t mode)
 {
    // Merge all input coverage databases given on command line
 
@@ -1815,7 +1814,7 @@ static cover_data_t *merge_coverage_files(int argc, int next_cmd, char **argv,
       if (i == optind)
          cover = cover_read_items(f, rpt_mask);
       else
-         cover_merge_items(f, cover);
+         cover_merge_items(f, cover, mode);
 
       fbuf_close(f, NULL);
    }
@@ -1885,7 +1884,7 @@ static int cover_export_cmd(int argc, char **argv, cmd_state_t *state)
 
    cover_data_t *cover;
    if (looks_like_file)
-      cover = merge_coverage_files(argc, next_cmd, argv, 0);
+      cover = merge_coverage_files(argc, next_cmd, argv, 0, MERGE_UNION);
    else {
       set_top_level(argv, next_cmd, state);
 
@@ -1985,7 +1984,8 @@ static int cover_report_cmd(int argc, char **argv, cmd_state_t *state)
 
    progress("initialising");
 
-   cover_data_t *cover = merge_coverage_files(argc, next_cmd, argv, rpt_mask);
+   cover_data_t *cover =
+      merge_coverage_files(argc, next_cmd, argv, rpt_mask, MERGE_UNION);
 
    if (exclude_file && cover) {
       progress("loading exclude file %s", exclude_file);
@@ -2005,6 +2005,7 @@ static int cover_merge_cmd(int argc, char **argv, cmd_state_t *state)
 {
    static struct option long_options[] = {
       { "output",       required_argument, 0, 'o' },
+      { "merge-mode",   required_argument, 0, 'm' },
       { "verbose",      no_argument,       0, 'V' },
       { 0, 0, 0, 0 }
    };
@@ -2014,12 +2015,16 @@ static int cover_merge_cmd(int argc, char **argv, cmd_state_t *state)
    const char *out_db = NULL;
    int c, index;
    const char *spec = ":Vo:";
+   merge_mode_t mode = MERGE_UNION;
 
    while ((c = getopt_long(next_cmd, argv, spec, long_options, &index)) != -1) {
       switch (c) {
       case 'o':
          out_db = optarg;
          break;
+      case 'm':
+         mode = parse_cover_merge_mode(optarg);
+         break;
       case 'V':
          opt_set_int(OPT_VERBOSE, 1);
          break;
@@ -2037,7 +2042,7 @@ static int cover_merge_cmd(int argc, char **argv, cmd_state_t *state)
 
    progress("initialising");
 
-   cover_data_t *cover = merge_coverage_files(argc, next_cmd, argv, 0);
+   cover_data_t *cover = merge_coverage_files(argc, next_cmd, argv, 0, mode);
 
    progress("saving merged coverage database to %s", out_db);
 
@@ -2129,7 +2134,7 @@ static void usage(void)
              "Generate HTML report from coverage database" },
            { "--cover-merge FILE...", "Merge multiple coverage databases" },
 #ifdef ENABLE_TCL
-           { "--do [TOP] SCRIPT...", "Evaluate TCL script" },
+           { "--do [TOP] FILE...", "Evaluate TCL script" },
 #endif
 #ifdef ENABLE_GUI
            { "--gui", "Launch browser-based GUI" },
@@ -2147,7 +2152,7 @@ static void usage(void)
         {
            { "-h, --help", "Display this message and exit" },
            { "-H SIZE", "Set the maximum heap size to SIZE bytes" },
-           { "--ieee-warnings={on,off}",
+           { "--ieee-warnings={on,off,off-at-0}",
              "Enable or disable warnings from IEEE packages" },
            { "--ignore-time", "Skip source file timestamp check" },
            { "--load=PLUGIN", "Load VHPI plugin at startup" },
@@ -2229,15 +2234,24 @@ static void usage(void)
 #endif
       { "Coverage report options",
         {
+           { "-o, --output=dir", "Output directory for HTML report" },
            { "--exclude-file=FILE",
              "Apply exclude file when generating report, see manual for syntax"
            },
            { "--dont-print={covered,uncovered,excluded}",
              "Exclude specified items from coverage report" },
+           { "--item-limit=LIMIT",
+             "Display at most LIMIT items in a single report file "
+             "(default 5000)" },
+           { "--per-file",
+             "Create source file code coverage report." },
         }
       },
       { "Coverage merge options",
         {
+           { "--merge-mode={union,intersect}",
+             "Merge hierarchies by union or intersection, default is union"
+           },
            { "-o, --output=FILE", "Output database file name" },
         }
       },
@@ -2406,7 +2420,6 @@ static int process_command(int argc, char **argv, cmd_state_t *state)
       { "init",         no_argument, 0, 'n' },
       { "install",      no_argument, 0, 'I' },
       { "print-deps",   no_argument, 0, 'P' },
-      { "aotgen",       no_argument, 0, 'A' },
       { "do",           no_argument, 0, 'D' },
       { "cover-export", no_argument, 0, 'E' },
       { "cover-merge",  no_argument, 0, 'M' },
@@ -2446,8 +2459,6 @@ static int process_command(int argc, char **argv, cmd_state_t *state)
       return install_cmd(argc, argv, state);
    case 'P':
       return print_deps_cmd(argc, argv, state);
-   case 'A':
-      return aotgen_cmd(argc, argv, state);
    case 'D':
       return do_cmd(argc, argv, state);
    case 'i':
@@ -2488,7 +2499,6 @@ int main(int argc, char **argv)
       { "work",          required_argument, 0, 'w' },
       { "std",           required_argument, 0, 's' },
       { "messages",      required_argument, 0, 'I' },
-      { "native",        no_argument,       0, 'n' }, // DEPRECATED 1.4
       { "map",           required_argument, 0, 'p' },
       { "ieee-warnings", required_argument, 0, 'W' },
       { "ignore-time",   no_argument,       0, 'i' },
@@ -2542,9 +2552,6 @@ int main(int argc, char **argv)
       case 'f':
          warnf("the --force-init option is deprecated and has no effect");
          break;
-      case 'n':
-         warnf("the --native option is deprecated and has no effect");
-         break;
       case 'M':
          opt_set_size(OPT_ARENA_SIZE, parse_size(optarg));
          break;
@@ -2570,7 +2577,7 @@ int main(int argc, char **argv)
          opt_set_int(OPT_PLI_DEBUG, 1);
          break;
       case 'W':
-         opt_set_int(OPT_IEEE_WARNINGS, parse_on_off(optarg));
+         opt_set_int(OPT_IEEE_WARNINGS, parse_ieee_warnings(optarg));
          break;
       case 'S':
          opt_set_int(OPT_RANDOM_SEED, parse_int(optarg));
