@@ -265,7 +265,12 @@ const char *token_str(token_t tok)
          "shortreal", "realtime", "`__nvc_push", "`__nvc_pop", "++", "--",
          "var", "`default_nettype", "tri", "tri0", "tri1", "wand", "triand",
          "wor", "trior", "trireg", "uwire", "none", "localparam", "always_comb",
-         "always_ff", "always_latch"
+         "always_ff", "always_latch", "(*)", "endcase", "casex", "casez",
+         "ifnone", "edge", "*>", "$setup", "$hold", "$recovery", "$removal",
+         "$setuphold", "$recrem", "$width", "+:", "-:", "endgenerate",
+         "`resetall", "event", "&&&", "specparam", "fork", "join", "automatic",
+         "genvar", "highz0", "highz1", "strong0", "strong1", "pull0", "pull1",
+         "weak0", "weak1", "small", "medium", "large", "vectored", "scalared",
       };
 
       if (tok >= 200 && tok - 200 < ARRAY_LEN(token_strs))
