@@ -527,9 +527,9 @@ static tree_t simp_attr_ref(tree_t t, simp_ctx_t *ctx)
                      tree_t p0 = tree_param(name, 0);
                      int64_t dim;
                      if (!folded_int(tree_value(p0), &dim))
-                        break;
+                        return t;
                      else if (dim < 1 || dim > dimension_of(tree_type(prefix)))
-                        break;
+                        return t;
 
                      tree_add_param(new, p0);
                   }
@@ -539,8 +539,6 @@ static tree_t simp_attr_ref(tree_t t, simp_ctx_t *ctx)
                   return new;
                }
             }
-
-            break;  // Prefix may contain errors
          }
 
          type_t type = tree_type(name);
@@ -1322,7 +1320,9 @@ static tree_t simp_if_generate(tree_t t)
          else {
             tree_t b = tree_new(T_BLOCK);
             tree_set_loc(b, tree_loc(t));
-            if (tree_has_ident(t))
+            if (tree_has_ident(c))
+               tree_set_ident(b, tree_ident(c));
+            else if (tree_has_ident(t))
                tree_set_ident(b, tree_ident(t));
 
             const int ndecls = tree_decls(c);
