@@ -500,7 +500,7 @@ static tree_kind_t decl_kinds[] = {
    T_GROUP_TEMPLATE, T_GROUP,          T_SUBTYPE_DECL,    T_PACKAGE,
    T_PACK_BODY,      T_PACK_INST,      T_GENERIC_DECL,    T_PARAM_DECL,
    T_PROC_INST,      T_FUNC_INST,      T_PSL_DECL,        T_VIEW_DECL,
-   T_PROT_DECL,
+   T_PROT_DECL,      T_VERILOG,
 };
 
 static void tree_assert_kind(tree_t t, const tree_kind_t *list, size_t len,
@@ -1375,39 +1375,11 @@ tree_t tree_container(tree_t t)
    return container_of(o, struct _tree, object);
 }
 
-void tree_locus(tree_t t, ident_t *unit, ptrdiff_t *offset)
-{
-   assert(t != NULL);
-   object_locus(&(t->object), unit, offset);
-}
-
-tree_t tree_from_locus(ident_t unit, ptrdiff_t offset,
-                       tree_load_fn_t find_deps_fn)
-{
-   object_t *o = object_from_locus(unit, offset,
-                                   (object_load_fn_t)find_deps_fn);
-   assert(o->tag == OBJECT_TAG_TREE);
-   return container_of(o, struct _tree, object);
-}
-
 void tree_walk_deps(tree_t t, tree_deps_fn_t fn, void *ctx)
 {
    arena_walk_deps(object_arena(&(t->object)), fn, ctx);
 }
 
-int tree_stable_compar(const void *pa, const void *pb)
-{
-   tree_t a = *(tree_t *)pa;
-   tree_t b = *(tree_t *)pb;
-
-   if (a == NULL)
-      return -1;
-   else if (b == NULL)
-      return 1;
-   else
-      return tree_loc(a)->first_line - tree_loc(b)->first_line;
-}
-
 object_t *tree_to_object(tree_t t)
 {
    return t ? &(t->object) : NULL;
