@@ -172,6 +172,7 @@ typedef enum {
    VCODE_OP_RECORD_SCOPE,
    VCODE_OP_SYSCALL,
    VCODE_OP_PUT_CONVERSION,
+   VCODE_OP_DIR_CHECK,
 } vcode_op_t;
 
 typedef enum {
@@ -215,7 +216,6 @@ typedef enum {
    VCODE_UNIT_PACKAGE,
    VCODE_UNIT_PROTECTED,
    VCODE_UNIT_PROPERTY,
-   VCODE_UNIT_SHAPE,
 } vunit_kind_t;
 
 typedef struct {
@@ -313,7 +313,6 @@ vcode_unit_t vcode_active_unit(void);
 vcode_unit_t vcode_unit_context(vcode_unit_t vu);
 object_t *vcode_unit_object(vcode_unit_t vu);
 void vcode_set_result(vcode_type_t type);
-void vcode_check_shape(vcode_unit_t vu, vcode_unit_t shape);
 
 void vcode_state_save(vcode_state_t *state);
 void vcode_state_restore(const vcode_state_t *state);
@@ -361,7 +360,6 @@ vcode_unit_t emit_function(ident_t name, object_t *obj, vcode_unit_t context);
 vcode_unit_t emit_procedure(ident_t name, object_t *obj, vcode_unit_t context);
 vcode_unit_t emit_process(ident_t name, object_t *obj, vcode_unit_t context);
 vcode_unit_t emit_instance(ident_t name, object_t *obj, vcode_unit_t context);
-vcode_unit_t emit_shape(ident_t name, object_t *obj, vcode_unit_t context);
 vcode_unit_t emit_package(ident_t name, object_t *obj, vcode_unit_t context);
 vcode_unit_t emit_protected(ident_t name, object_t *obj, vcode_unit_t context);
 vcode_unit_t emit_property(ident_t name, object_t *obj, vcode_unit_t context);
@@ -411,6 +409,7 @@ void emit_range_check(vcode_reg_t reg, vcode_reg_t left, vcode_reg_t right,
                       vcode_reg_t dir, vcode_reg_t locus, vcode_reg_t hint);
 void emit_index_check(vcode_reg_t reg, vcode_reg_t left, vcode_reg_t right,
                       vcode_reg_t dir, vcode_reg_t locus, vcode_reg_t hint);
+void emit_dir_check(vcode_reg_t reg, vcode_reg_t dir, vcode_reg_t locus);
 vcode_reg_t emit_index(vcode_var_t var, vcode_reg_t offset);
 vcode_reg_t emit_cast(vcode_type_t type, vcode_reg_t bounds, vcode_reg_t reg);
 void emit_return(vcode_reg_t reg);
@@ -505,8 +504,7 @@ void emit_transfer_signal(vcode_reg_t target, vcode_reg_t source,
                           vcode_reg_t after);
 vcode_reg_t emit_resolution_wrapper(vcode_type_t type, vcode_reg_t closure,
                                     vcode_reg_t nlits);
-vcode_reg_t emit_closure(ident_t func, vcode_reg_t context, vcode_type_t atype,
-                         vcode_type_t rtype);
+vcode_reg_t emit_closure(ident_t func, vcode_reg_t context, vcode_type_t rtype);
 vcode_reg_t emit_protected_init(vcode_type_t type, vcode_reg_t context,
                                 vcode_reg_t path_name, vcode_reg_t inst_name);
 void emit_process_init(ident_t name, vcode_reg_t locus);
