@@ -17,6 +17,7 @@
 
 #include "util.h"
 #include "common.h"
+#include "object.h"
 #include "ident.h"
 #include "vlog/vlog-node.h"
 #include "vlog/vlog-number.h"
@@ -24,11 +25,35 @@
 
 #include <ctype.h>
 
+#define DUMP_ADDRESS 0
+
 static inline void tab(int indent)
 {
    print_syntax("%*s", indent, "");
 }
 
+static void vlog_dump_address(vlog_node_t v)
+{
+#if DUMP_ADDRESS
+   uint32_t a = (uint32_t)((uintptr_t)object_arena(vlog_to_object(v)) >> 2);
+   a = (a ^ 61) ^ (a >> 16);
+   a = a + (a << 3);
+   a = a ^ (a >> 4);
+   a = a * UINT32_C(0x27d4eb2d);
+   a = a ^ (a >> 15);
+
+   const int r = 1 + a % 5;
+   const int g = 1 + (a >> 8) % 5;
+   const int b = 1 + (a >> 16) % 5;
+
+   const int color = 16 + 36*r + 6*g + b;
+
+   char *LOCAL fmt = xasprintf("$!#%d${%%p:%s}$$", color,
+                               vlog_kind_str(vlog_kind(v)));
+   color_printf(fmt, v);
+#endif
+}
+
 static void vlog_dump_paren(vlog_node_t v, int indent)
 {
    switch (vlog_kind(v)) {
@@ -46,6 +71,7 @@ static void vlog_dump_paren(vlog_node_t v, int indent)
 
 static void vlog_dump_module(vlog_node_t v, int indent)
 {
+   vlog_dump_address(v);
    print_syntax("#module %s", istr(vlog_ident2(v)));
 
    const int nports = vlog_ports(v);
@@ -73,6 +99,7 @@ static void vlog_dump_module(vlog_node_t v, int indent)
 
 static void vlog_dump_primitive(vlog_node_t v, int indent)
 {
+   vlog_dump_address(v);
    print_syntax("#primitive %s", istr(vlog_ident2(v)));
 
    const int nports = vlog_ports(v);
@@ -101,7 +128,22 @@ static void vlog_dump_primitive(vlog_node_t v, int indent)
 static void vlog_dump_udp_entry(vlog_node_t v, int indent)
 {
    tab(indent);
-   print_syntax("%s;\n", vlog_text(v));
+
+   vlog_udp_symbol_t last = V_UDP_SYMBOL_INPUT;
+   const int nsymbols = vlog_params(v);
+   for (int i = 0; i < nsymbols; i++) {
+      vlog_node_t sym = vlog_param(v, i);
+
+      const vlog_udp_symbol_t kind = vlog_subkind(sym);
+      if (kind != last) {
+         print_syntax(":");
+         last = kind;
+      }
+
+      print_syntax("%c", (char)vlog_ival(sym));
+   }
+
+   print_syntax(";\n");
 }
 
 static void vlog_dump_udp_table(vlog_node_t v, int indent)
@@ -120,6 +162,7 @@ static void vlog_dump_udp_table(vlog_node_t v, int indent)
 static void vlog_dump_port_decl(vlog_node_t v, int indent)
 {
    tab(indent);
+   vlog_dump_address(v);
 
    switch (vlog_subkind(v)) {
    case V_PORT_INPUT: print_syntax("#input"); break;
@@ -147,6 +190,7 @@ static void vlog_dump_dimensions(vlog_node_t v, int indent)
 static void vlog_dump_net_decl(vlog_node_t v, int indent)
 {
    tab(indent);
+   vlog_dump_address(v);
 
    switch (vlog_subkind(v)) {
    case V_NET_WIRE: print_syntax("#wire "); break;
@@ -161,6 +205,7 @@ static void vlog_dump_net_decl(vlog_node_t v, int indent)
 static void vlog_dump_var_decl(vlog_node_t v, int indent)
 {
    tab(indent);
+   vlog_dump_address(v);
 
    vlog_dump(vlog_type(v), indent);
    print_syntax(" %s", istr(vlog_ident(v)));
@@ -180,6 +225,7 @@ static void vlog_dump_always(vlog_node_t v, int indent)
 
    print_syntax("#always ");
    vlog_dump(vlog_stmt(v, 0), indent);
+   print_syntax("\n");
 }
 
 static void vlog_dump_initial(vlog_node_t v, int indent)
@@ -188,18 +234,26 @@ static void vlog_dump_initial(vlog_node_t v, int indent)
 
    print_syntax("#initial ");
    vlog_dump(vlog_stmt(v, 0), indent);
+   print_syntax("\n");
 }
 
-static void vlog_dump_seq_block(vlog_node_t v, int indent)
+static void vlog_dump_block(vlog_node_t v, int indent)
 {
    print_syntax("#begin\n");
 
+   const int ndecls = vlog_decls(v);
+   for (int i = 0; i < ndecls; i++)
+      vlog_dump(vlog_decl(v, i), indent + 2);
+
    const int nstmts = vlog_stmts(v);
-   for (int i = 0; i < nstmts; i++)
+   for (int i = 0; i < nstmts; i++) {
+      tab(indent + 2);
       vlog_dump(vlog_stmt(v, i), indent + 2);
+      print_syntax("\n");
+   }
 
    tab(indent);
-   print_syntax("#end\n");
+   print_syntax("#end");
 }
 
 static void vlog_dump_timing(vlog_node_t v, int indent)
@@ -207,7 +261,7 @@ static void vlog_dump_timing(vlog_node_t v, int indent)
    vlog_dump(vlog_value(v), indent);
 
    if (vlog_stmts(v) > 0)
-      vlog_dump(vlog_stmt(v, 0), 0);
+      vlog_dump(vlog_stmt(v, 0), indent);
    else
       print_syntax(";\n");
 }
@@ -227,26 +281,26 @@ static void vlog_dump_event_control(vlog_node_t v)
    print_syntax("@(");
 
    const int nparams = vlog_params(v);
-   for (int i = 0; i < nparams; i++)
+   for (int i = 0; i < nparams; i++) {
+      if (i > 0) print_syntax(", ");
       vlog_dump(vlog_param(v, i), 0);
+   }
 
    print_syntax(") ");
 }
 
 static void vlog_dump_nbassign(vlog_node_t v, int indent)
 {
-   tab(indent);
    vlog_dump(vlog_target(v), 0);
    print_syntax(" <= ");
    if (vlog_has_delay(v))
       vlog_dump(vlog_delay(v), 0);
    vlog_dump(vlog_value(v), 0);
-   print_syntax(";\n");
+   print_syntax(";");
 }
 
 static void vlog_dump_bassign(vlog_node_t v, int indent)
 {
-   tab(indent);
    vlog_dump(vlog_target(v), 0);
 
    switch (vlog_subkind(v)) {
@@ -259,7 +313,7 @@ static void vlog_dump_bassign(vlog_node_t v, int indent)
       vlog_dump(vlog_delay(v), 0);
 
    vlog_dump(vlog_value(v), 0);
-   print_syntax(";\n");
+   print_syntax(";");
 }
 
 static void vlog_dump_assign(vlog_node_t v, int indent)
@@ -274,24 +328,21 @@ static void vlog_dump_assign(vlog_node_t v, int indent)
 
 static void vlog_dump_stmt_or_null(vlog_node_t v, int indent)
 {
-   const int nstmts = vlog_stmts(v);
-   if (nstmts == 0)
-      print_syntax(";\n");
-   else {
-      print_syntax("\n");
-      for (int i = 0; i < nstmts; i++)
-         vlog_dump(vlog_stmt(v, i), indent + 2);
+   switch (vlog_stmts(v)) {
+   case 0:  print_syntax(";"); break;
+   case 1:  vlog_dump(vlog_stmt(v, 0), indent); break;
+   default: should_not_reach_here();
    }
 }
 
 static void vlog_dump_if(vlog_node_t v, int indent)
 {
-   tab(indent);
    print_syntax("#if ");
 
    const int nconds = vlog_conds(v);
    for (int i = 0; i < nconds; i++) {
       if (i > 0) {
+         print_syntax("\n");
          tab(indent);
          print_syntax("#else ");
       }
@@ -300,7 +351,7 @@ static void vlog_dump_if(vlog_node_t v, int indent)
       if (vlog_has_value(c)) {
          print_syntax("(");
          vlog_dump(vlog_value(c), 0);
-         print_syntax(")");
+         print_syntax(") ");
       }
 
       vlog_dump_stmt_or_null(c, indent);
@@ -309,7 +360,6 @@ static void vlog_dump_if(vlog_node_t v, int indent)
 
 static void vlog_dump_sys_tcall(vlog_node_t v, int indent)
 {
-   tab(indent);
    print_syntax("%s", istr(vlog_ident(v)));
 
    const int nparams = vlog_params(v);
@@ -322,7 +372,7 @@ static void vlog_dump_sys_tcall(vlog_node_t v, int indent)
       print_syntax(")");
    }
 
-   print_syntax(";\n");
+   print_syntax(";");
 }
 
 static void vlog_dump_sys_fcall(vlog_node_t v)
@@ -342,7 +392,13 @@ static void vlog_dump_sys_fcall(vlog_node_t v)
 
 static void vlog_dump_string(vlog_node_t v)
 {
-   print_syntax("\"%s\"", vlog_text(v));
+   print_syntax("\"");
+
+   number_t n = vlog_number(v);
+   for (int i = number_width(n)/8 - 1; i >= 0; i--)
+      print_syntax("%c", number_byte(n, i));
+
+   print_syntax("\"");
 }
 
 static void vlog_dump_number(vlog_node_t v)
@@ -363,15 +419,18 @@ static void vlog_dump_binary(vlog_node_t v)
    vlog_dump_paren(vlog_left(v), 0);
 
    switch (vlog_subkind(v)) {
-   case V_BINARY_OR: print_syntax(" | "); break;
-   case V_BINARY_AND: print_syntax(" & "); break;
-   case V_BINARY_CASE_EQ: print_syntax(" === "); break;
+   case V_BINARY_OR:       print_syntax(" | "); break;
+   case V_BINARY_AND:      print_syntax(" & "); break;
+   case V_BINARY_CASE_EQ:  print_syntax(" === "); break;
    case V_BINARY_CASE_NEQ: print_syntax(" !== "); break;
-   case V_BINARY_LOG_EQ: print_syntax(" == "); break;
-   case V_BINARY_LOG_NEQ: print_syntax(" != "); break;
-   case V_BINARY_LOG_OR: print_syntax(" || "); break;
-   case V_BINARY_LOG_AND: print_syntax(" && "); break;
-   case V_BINARY_PLUS: print_syntax(" + "); break;
+   case V_BINARY_LOG_EQ:   print_syntax(" == "); break;
+   case V_BINARY_LOG_NEQ:  print_syntax(" != "); break;
+   case V_BINARY_LOG_OR:   print_syntax(" || "); break;
+   case V_BINARY_LOG_AND:  print_syntax(" && "); break;
+   case V_BINARY_PLUS:     print_syntax(" + "); break;
+   case V_BINARY_MINUS:    print_syntax(" - "); break;
+   case V_BINARY_LT:       print_syntax(" < "); break;
+   case V_BINARY_GT:       print_syntax(" > "); break;
    }
 
    vlog_dump_paren(vlog_right(v), 0);
@@ -387,9 +446,28 @@ static void vlog_dump_unary(vlog_node_t v)
    vlog_dump(vlog_value(v), 0);
 }
 
+static void vlog_dump_postfix(vlog_node_t v)
+{
+   vlog_dump(vlog_target(v), 0);
+
+   switch (vlog_subkind(v)) {
+   case V_INCDEC_PLUS: print_syntax("++"); break;
+   case V_INCDEC_MINUS: print_syntax("--"); break;
+   }
+}
+
+static void vlog_dump_prefix(vlog_node_t v)
+{
+   switch (vlog_subkind(v)) {
+   case V_INCDEC_PLUS: print_syntax("++"); break;
+   case V_INCDEC_MINUS: print_syntax("--"); break;
+   }
+
+   vlog_dump(vlog_target(v), 0);
+}
+
 static void vlog_dump_delay_control(vlog_node_t v, int indent)
 {
-   tab(indent);
    print_syntax("##");
    vlog_dump(vlog_value(v), 0);
    print_syntax(" ");
@@ -424,25 +502,54 @@ static void vlog_dump_gate_inst(vlog_node_t v, int indent)
 
 static void vlog_dump_mod_inst(vlog_node_t v, int indent)
 {
-   tab(indent);
+   print_syntax("%s ", istr(vlog_ident(v)));
 
-   print_syntax("%s ", istr(vlog_ident2(v)));
+   const int nparams = vlog_params(v);
+   print_syntax("(");
+   for (int i = 0; i < nparams; i++) {
+      if (i > 0) print_syntax(",");
+      vlog_dump(vlog_param(v, i), 0);
+   }
+   print_syntax(")");
+}
+
+static void vlog_dump_inst_list(vlog_node_t v, int indent)
+{
+   tab(indent);
 
    print_syntax("%s ", istr(vlog_ident(v)));
 
    const int nparams = vlog_params(v);
    if (nparams > 0) {
-      print_syntax("(");
+      print_syntax("##(");
       for (int i = 0; i < nparams; i++) {
          if (i > 0) print_syntax(",");
          vlog_dump(vlog_param(v, i), 0);
       }
-      print_syntax(")");
+      print_syntax(") ");
+   }
+
+   const int nstmts = vlog_stmts(v);
+   for (int i = 0; i < nstmts; i++) {
+      if (i > 0) print_syntax(", ");
+      vlog_dump(vlog_stmt(v, i), indent + 2);
    }
 
    print_syntax(";\n");
 }
 
+static void vlog_dump_param_assign(vlog_node_t v, int indent)
+{
+   if (vlog_has_ident(v)) {
+      print_syntax(".%s(", istr(vlog_ident(v)));
+      if (vlog_has_value(v))
+         vlog_dump(vlog_value(v), indent);
+      print_syntax(")");
+   }
+   else
+      vlog_dump(vlog_value(v), indent);
+}
+
 static void vlog_dump_strength(vlog_node_t v, int indent)
 {
    static const char *map[] = {
@@ -455,7 +562,7 @@ static void vlog_dump_strength(vlog_node_t v, int indent)
 
 static void vlog_dump_bit_select(vlog_node_t v, int indent)
 {
-   print_syntax("%s", istr(vlog_ident(v)));
+   vlog_dump(vlog_value(v), 0);
 
    const int nparams = vlog_params(v);
    for (int i = 0; i < nparams; i++) {
@@ -465,6 +572,16 @@ static void vlog_dump_bit_select(vlog_node_t v, int indent)
    }
 }
 
+static void vlog_dump_part_select(vlog_node_t v, int indent)
+{
+   vlog_dump(vlog_value(v), 0);
+   print_syntax("[");
+   vlog_dump(vlog_left(v), 0);
+   print_syntax(":");
+   vlog_dump(vlog_right(v), 0);
+   print_syntax("]");
+}
+
 static void vlog_dump_data_type(vlog_node_t v, int indent)
 {
    switch (vlog_subkind(v)) {
@@ -474,6 +591,7 @@ static void vlog_dump_data_type(vlog_node_t v, int indent)
    case DT_SHORTREAL: print_syntax("#shortreal"); break;
    case DT_INTEGER: print_syntax("#integer"); break;
    case DT_BIT: print_syntax("#bit"); break;
+   case DT_INT: print_syntax("#int"); break;
    default: should_not_reach_here();
    }
 
@@ -482,20 +600,17 @@ static void vlog_dump_data_type(vlog_node_t v, int indent)
 
 static void vlog_dump_do_while(vlog_node_t v, int indent)
 {
-   tab(indent);
-   print_syntax("#do");
+   print_syntax("#do ");
 
    vlog_dump_stmt_or_null(v, indent);
 
-   tab(indent);
-   print_syntax("#while (");
+   print_syntax(" #while (");
    vlog_dump(vlog_value(v), indent);
-   print_syntax(");\n");
+   print_syntax(");");
 }
 
 static void vlog_dump_while(vlog_node_t v, int indent)
 {
-   tab(indent);
    print_syntax("#while (");
    vlog_dump(vlog_value(v), indent);
    print_syntax(")");
@@ -505,14 +620,96 @@ static void vlog_dump_while(vlog_node_t v, int indent)
 
 static void vlog_dump_repeat(vlog_node_t v, int indent)
 {
-   tab(indent);
    print_syntax("#repeat (");
    vlog_dump(vlog_value(v), indent);
-   print_syntax(")");
+   print_syntax(") ");
 
    vlog_dump_stmt_or_null(v, indent);
 }
 
+static void vlog_dump_forever(vlog_node_t v, int indent)
+{
+   print_syntax("#forever ");
+   vlog_dump_stmt_or_null(v, indent);
+}
+
+static void vlog_dump_for_init(vlog_node_t v, int indent)
+{
+   const int nstmts = vlog_stmts(v);
+   if (nstmts == 0)
+      print_syntax(";");
+   else {
+      for (int i = 0; i < nstmts; i++) {
+         vlog_dump(vlog_stmt(v, i), indent);
+         print_syntax(" ");
+      }
+   }
+}
+
+static void vlog_dump_for_step(vlog_node_t v, int indent)
+{
+   const int nstmts = vlog_stmts(v);
+   if (nstmts == 0)
+      print_syntax(";");
+   else {
+      for (int i = 0; i < nstmts; i++) {
+         print_syntax(" ");
+         vlog_dump(vlog_stmt(v, i), indent);
+      }
+   }
+}
+
+static void vlog_dump_for_loop(vlog_node_t v, int indent)
+{
+   print_syntax("#for (");
+   vlog_dump(vlog_left(v), 0);
+
+   if (vlog_has_value(v))
+      vlog_dump(vlog_value(v), 0);
+
+   print_syntax(";");
+   vlog_dump(vlog_right(v), 0);
+   print_syntax(") ");
+
+   vlog_dump_stmt_or_null(v, indent);
+}
+
+static void vlog_dump_case(vlog_node_t v, int indent)
+{
+   static const char *suffix[] = { "", "x", "z" };
+
+   print_syntax("#case%s (", suffix[vlog_subkind(v)]);
+   vlog_dump(vlog_value(v), indent);
+   print_syntax(")\n");
+
+   const int nstmts = vlog_stmts(v);
+   for (int i = 0; i < nstmts; i++)
+      vlog_dump(vlog_stmt(v, i), indent + 2);
+
+   tab(indent);
+   print_syntax("#endcase");
+}
+
+static void vlog_dump_case_item(vlog_node_t v, int indent)
+{
+   tab(indent);
+
+   const int nparams = vlog_params(v);
+   if (nparams == 0)
+      print_syntax("#default");
+   else {
+      vlog_dump(vlog_param(v, 0), 0);
+      for (int i = 1; i < nparams; i++) {
+         print_syntax(", ");
+         vlog_dump(vlog_param(v, i), 0);
+      }
+   }
+
+   print_syntax(": ");
+   vlog_dump_stmt_or_null(v, indent);
+   print_syntax("\n");
+}
+
 static void vlog_dump_cond_expr(vlog_node_t v, int indent)
 {
    vlog_dump(vlog_value(v), 0);
@@ -528,7 +725,7 @@ static void vlog_dump_concat(vlog_node_t v, int indent)
 
    const bool multiple = vlog_has_value(v);
    if (multiple) {
-      vlog_dump(vlog_value(v), 0);
+      vlog_dump_paren(vlog_value(v), 0);
       print_syntax("{");
    }
 
@@ -547,6 +744,7 @@ static void vlog_dump_concat(vlog_node_t v, int indent)
 static void vlog_dump_param_decl(vlog_node_t v, int indent)
 {
    tab(indent);
+   vlog_dump_address(v);
 
    switch (vlog_kind(v)) {
    case V_PARAM_DECL:
@@ -570,9 +768,64 @@ static void vlog_dump_param_decl(vlog_node_t v, int indent)
    print_syntax(";\n");
 }
 
+static void vlog_dump_ref(vlog_node_t v, int indent)
+{
+   if (vlog_has_ref(v))
+      vlog_dump_address(vlog_ref(v));
+
+   print_syntax("%s", istr(vlog_ident(v)));
+}
+
+static void vlog_dump_task_decl(vlog_node_t v, int indent)
+{
+   tab(indent);
+   print_syntax("#task %s(", istr(vlog_ident(v)));
+
+   print_syntax(");\n");
+
+   const int nstmts = vlog_stmts(v);
+   for (int i = 0; i < nstmts; i++) {
+      tab(indent + 2);
+      vlog_dump(vlog_stmt(v, i), indent + 2);
+      print_syntax("\n");
+   }
+
+   tab(indent);
+   print_syntax("#endtask\n");
+}
+
+static void vlog_dump_func_decl(vlog_node_t v, int indent)
+{
+   tab(indent);
+   print_syntax("#function %s(", istr(vlog_ident(v)));
+
+   print_syntax(");\n");
+
+   const int nstmts = vlog_stmts(v);
+   for (int i = 0; i < nstmts; i++) {
+      tab(indent + 2);
+      vlog_dump(vlog_stmt(v, i), indent + 2);
+      print_syntax("\n");
+   }
+
+   tab(indent);
+   print_syntax("#endfunction\n");
+}
+
+static void vlog_dump_wait(vlog_node_t v, int indent)
+{
+   print_syntax("#wait (");
+   vlog_dump(vlog_value(v), 0);
+   print_syntax(") ");
+   vlog_dump_stmt_or_null(v, indent);
+}
+
 void vlog_dump(vlog_node_t v, int indent)
 {
    switch (vlog_kind(v)) {
+   case V_INST_BODY:
+      print_syntax("// Instantiated module %s\n", istr(vlog_ident(v)));
+      // Fall-through
    case V_MODULE:
       vlog_dump_module(v, indent);
       break;
@@ -580,7 +833,7 @@ void vlog_dump(vlog_node_t v, int indent)
       vlog_dump_primitive(v, indent);
       break;
    case V_REF:
-      print_syntax("%s", istr(vlog_ident(v)));
+      vlog_dump_ref(v, indent);
       break;
    case V_PORT_DECL:
       vlog_dump_port_decl(v, indent);
@@ -615,8 +868,8 @@ void vlog_dump(vlog_node_t v, int indent)
    case V_ASSIGN:
       vlog_dump_assign(v, indent);
       break;
-   case V_SEQ_BLOCK:
-      vlog_dump_seq_block(v, indent);
+   case V_BLOCK:
+      vlog_dump_block(v, indent);
       break;
    case V_SYS_TCALL:
       vlog_dump_sys_tcall(v, indent);
@@ -634,6 +887,7 @@ void vlog_dump(vlog_node_t v, int indent)
       vlog_dump_real(v);
       break;
    case V_IF:
+   case V_IF_GENERATE:
       vlog_dump_if(v, indent);
       break;
    case V_BINARY:
@@ -642,6 +896,12 @@ void vlog_dump(vlog_node_t v, int indent)
    case V_UNARY:
       vlog_dump_unary(v);
       break;
+   case V_POSTFIX:
+      vlog_dump_postfix(v);
+      break;
+   case V_PREFIX:
+      vlog_dump_prefix(v);
+      break;
    case V_DELAY_CONTROL:
       vlog_dump_delay_control(v, indent);
       break;
@@ -651,12 +911,18 @@ void vlog_dump(vlog_node_t v, int indent)
    case V_MOD_INST:
       vlog_dump_mod_inst(v, indent);
       break;
+   case V_INST_LIST:
+      vlog_dump_inst_list(v, indent);
+      break;
    case V_STRENGTH:
       vlog_dump_strength(v, indent);
       break;
    case V_BIT_SELECT:
       vlog_dump_bit_select(v, indent);
       break;
+   case V_PART_SELECT:
+      vlog_dump_part_select(v, indent);
+      break;
    case V_UDP_ENTRY:
       vlog_dump_udp_entry(v, indent);
       break;
@@ -677,9 +943,27 @@ void vlog_dump(vlog_node_t v, int indent)
    case V_REPEAT:
       vlog_dump_repeat(v, indent);
       break;
+   case V_FOREVER:
+      vlog_dump_forever(v, indent);
+      break;
+   case V_FOR_LOOP:
+      vlog_dump_for_loop(v, indent);
+      break;
+   case V_FOR_INIT:
+      vlog_dump_for_init(v, indent);
+      break;
+   case V_FOR_STEP:
+      vlog_dump_for_step(v, indent);
+      break;
+   case V_CASE:
+      vlog_dump_case(v, indent);
+      break;
    case V_COND_EXPR:
       vlog_dump_cond_expr(v, indent);
       break;
+   case V_CASE_ITEM:
+      vlog_dump_case_item(v, indent);
+      break;
    case V_CONCAT:
       vlog_dump_concat(v, indent);
       break;
@@ -687,6 +971,19 @@ void vlog_dump(vlog_node_t v, int indent)
    case V_LOCALPARAM:
       vlog_dump_param_decl(v, indent);
       break;
+   case V_PARAM_ASSIGN:
+   case V_PORT_CONN:
+      vlog_dump_param_assign(v, indent);
+      break;
+   case V_TASK_DECL:
+      vlog_dump_task_decl(v, indent);
+      break;
+   case V_FUNC_DECL:
+      vlog_dump_func_decl(v, indent);
+      break;
+   case V_WAIT:
+      vlog_dump_wait(v, indent);
+      break;
    default:
       print_syntax("\n");
       fflush(stdout);
