@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2024 Nick Gasson
+//  Copyright (C) 2024-2025 Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -22,6 +22,17 @@
 
 bool vlog_is_net(vlog_node_t v);
 unsigned vlog_dimensions(vlog_node_t v);
+int64_t vlog_get_const(vlog_node_t v);
+bool vlog_is_const(vlog_node_t v);
+void vlog_bounds(vlog_node_t v, int64_t *left, int64_t *right);
+unsigned vlog_size(vlog_node_t v);
 bool is_top_level(vlog_node_t v);
+bool is_data_type(vlog_node_t v);
+vlog_node_t vlog_longest_static_prefix(vlog_node_t v);
+
+#define CANNOT_HANDLE(v) do {                                           \
+      fatal_at(vlog_loc(v), "cannot handle %s in %s" ,                  \
+               vlog_kind_str(vlog_kind(v)), __FUNCTION__);              \
+   } while (0)
 
 #endif  // _VLOG_UTIL_H
