@@ -38,10 +38,14 @@
 
 typedef struct {
    PLI_INT32 type;
-   vpiHandle handle;
    loc_t     loc;
 } c_vpiObject;
 
+typedef struct {
+   c_vpiObject object;
+   uint32_t    refcount;
+} c_refcounted;
+
 typedef struct {
    unsigned      count;
    unsigned      limit;
@@ -58,12 +62,12 @@ typedef struct {
 } vpiLazyList;
 
 typedef struct {
-   c_vpiObject      object;
+   c_refcounted     refcounted;
    ident_t          name;
    s_vpi_systf_data systf;
 } c_callback;
 
-DEF_CLASS(callback, vpiCallback, object);
+DEF_CLASS(callback, vpiCallback, refcounted.object);
 
 typedef struct {
    c_vpiObject  object;
@@ -128,12 +132,12 @@ typedef struct {
 DEF_CLASS(operation, vpiOperation, expr.object);
 
 typedef struct {
-   c_vpiObject    object;
+   c_refcounted   refcounted;
    vpiObjectList *list;
    uint32_t       pos;
 } c_iterator;
 
-DEF_CLASS(iterator, vpiIterator, object);
+DEF_CLASS(iterator, vpiIterator, refcounted.object);
 
 typedef struct {
    c_vpiObject      object;
@@ -157,12 +161,19 @@ DEF_CLASS(reg, vpiReg, decl.object);
 #define HANDLE_BITS      (sizeof(vpiHandle) * 8)
 #define HANDLE_MAX_INDEX ((UINT64_C(1) << (HANDLE_BITS / 2)) - 1)
 
+typedef enum {
+   HANDLE_USER,
+   HANDLE_INTERNAL,
+} handle_kind_t;
+
 typedef struct {
-   c_vpiObject *obj;
-   uint32_t     refcount;
-   uint32_t     generation;
+   c_vpiObject  *obj;
+   handle_kind_t kind;
+   uint32_t      generation;
 } handle_slot_t;
 
+STATIC_ASSERT(sizeof(handle_slot_t) <= 16);
+
 typedef struct _vpi_context {
    shash_t       *strtab;
    rt_model_t    *model;
@@ -184,6 +195,7 @@ typedef struct _vpi_context {
 
 static c_vpiObject *build_expr(vlog_node_t v, c_abstractScope *scope);
 static void vpi_lazy_decls(c_vpiObject *obj);
+static c_refcounted *is_refcounted(c_vpiObject *obj);
 
 static vpi_context_t *global_context = NULL;   // TODO: thread local
 
@@ -208,28 +220,21 @@ static handle_slot_t *decode_handle(vpi_context_t *c, vpiHandle handle)
    else if (slot->generation != generation)
       return NULL;   // Use-after-free
 
-   assert(slot->refcount > 0);
-   assert(slot->obj->handle == handle);
-
    return slot;
 }
 
-static vpiHandle handle_for(c_vpiObject *obj)
+static inline vpiHandle encode_handle(handle_slot_t *slot, uint32_t index)
+{
+   const uintptr_t bits = (uintptr_t)slot->generation << HANDLE_BITS/2 | index;
+   return (vpiHandle)bits;
+}
+
+static vpiHandle handle_for(c_vpiObject *obj, handle_kind_t kind)
 {
    assert(obj != NULL);
 
    vpi_context_t *c = vpi_context();
 
-   if (obj->handle != NULL) {
-      handle_slot_t *slot = decode_handle(c, obj->handle);
-      assert(slot != NULL);
-      assert(slot->refcount > 0);
-      assert(slot->obj == obj);
-
-      slot->refcount++;
-      return obj->handle;
-   }
-
    uint32_t index = c->free_hint;
    if (index >= c->num_handles || c->handles[index].obj != NULL) {
       for (index = 0; index < c->num_handles; index++) {
@@ -250,20 +255,31 @@ static vpiHandle handle_for(c_vpiObject *obj)
 
       for (int i = index; i < new_size; i++) {
          c->handles[i].obj = NULL;
-         c->handles[i].refcount = 0;
          c->handles[i].generation = 1;
       }
    }
 
    handle_slot_t *slot = &(c->handles[index]);
-   assert(slot->refcount == 0);
-   slot->refcount = 1;
-   slot->obj = obj;
+   slot->obj  = obj;
+   slot->kind = kind;
 
    c->free_hint = index + 1;
 
-   const uintptr_t bits = (uintptr_t)slot->generation << HANDLE_BITS/2 | index;
-   return (obj->handle = (vpiHandle)bits);
+   c_refcounted *rc = is_refcounted(obj);
+   if (rc != NULL)
+      rc->refcount++;
+
+   return encode_handle(slot, index);
+}
+
+static inline vpiHandle user_handle_for(c_vpiObject *obj)
+{
+   return handle_for(obj, HANDLE_USER);
+}
+
+static inline vpiHandle internal_handle_for(c_vpiObject *obj)
+{
+   return handle_for(obj, HANDLE_INTERNAL);
 }
 
 static void drop_handle(vpi_context_t *c, vpiHandle handle)
@@ -272,25 +288,17 @@ static void drop_handle(vpi_context_t *c, vpiHandle handle)
    if (slot == NULL)
       return;
 
-   assert(slot->refcount > 0);
-   if (--(slot->refcount) > 0)
-      return;
-
    c_vpiObject *obj = slot->obj;
    slot->obj = NULL;
    slot->generation++;
 
-   obj->handle = NULL;
-
    c->free_hint = slot - c->handles;
 
-   switch (obj->type) {
-   case vpiCallback:
-   case vpiIterator:
-      APUSH(c->recycle, obj);
-      break;
-   default:
-      break;
+   c_refcounted *rc = is_refcounted(obj);
+   if (rc != NULL) {
+      assert(rc->refcount > 0);
+      if (--(rc->refcount) == 0)
+         APUSH(c->recycle, obj);
    }
 }
 
@@ -304,6 +312,17 @@ static inline c_vpiObject *from_handle(vpiHandle handle)
    return NULL;
 }
 
+static c_refcounted *is_refcounted(c_vpiObject *obj)
+{
+   switch (obj->type) {
+   case vpiCallback:
+   case vpiIterator:
+      return container_of(obj, c_refcounted, object);
+   default:
+      return NULL;
+   }
+}
+
 static c_tfCall *is_tfCall(c_vpiObject *obj)
 {
    switch (obj->type) {
@@ -513,7 +532,8 @@ static c_constant *build_constant(vlog_node_t v)
    switch (vlog_kind(v)) {
    case V_STRING:
       con->subtype = vpiStringConst;
-      con->size = strlen(vlog_text(v));
+      con->size = number_width(vlog_number(v));
+      assert(con->size % 8 == 0);
       break;
    case V_NUMBER:
       con->subtype = vpiBinaryConst;
@@ -535,6 +555,14 @@ static c_operation *build_operation(vlog_node_t v)
    case V_EMPTY:
       op->subtype = vpiNullOp;
       break;
+   case V_POSTFIX:
+      op->subtype = vlog_subkind(v) == V_INCDEC_PLUS
+         ? vpiPostIncOp : vpiPostDecOp;
+      break;
+   case V_PREFIX:
+      op->subtype = vlog_subkind(v) == V_INCDEC_PLUS
+         ? vpiPreIncOp : vpiPreDecOp;
+      break;
    default:
       break;
    }
@@ -551,6 +579,8 @@ static c_vpiObject *build_expr(vlog_node_t v, c_abstractScope *scope)
    case V_REF:
       {
          vlog_node_t d = vlog_ref(v);
+         if (vlog_kind(d) == V_PORT_DECL)
+            d = vlog_ref(d);
 
          vpiObjectList *list =
             expand_lazy_list(&(scope->object), &(scope->decls));
@@ -568,6 +598,9 @@ static c_vpiObject *build_expr(vlog_node_t v, c_abstractScope *scope)
    case V_UNARY:
    case V_SYS_FCALL:
    case V_EMPTY:
+   case V_PREFIX:
+   case V_POSTFIX:
+   case V_BIT_SELECT:   // XXX: check this
       return &(build_operation(v)->expr.object);
    default:
       fatal_trace("cannot build VPI expr for node kind %s",
@@ -634,10 +667,10 @@ static c_module *cached_module(tree_t block, rt_scope_t *s)
       tree_t wrap = tree_ref(hier);
       assert(tree_kind(wrap) == T_VERILOG);
 
-      vlog_node_t mod = tree_vlog(wrap);
-      assert(vlog_kind(mod) == V_MODULE);
+      vlog_node_t body = tree_vlog(wrap);
+      assert(vlog_kind(body) == V_INST_BODY);
 
-      m = build_module(mod, block, s);
+      m = build_module(body, block, s);
       hash_put(cache, block, m);
    }
 
@@ -713,7 +746,7 @@ vpiHandle vpi_register_systf(p_vpi_systf_data systf_data_p)
    cb->systf = *systf_data_p;
    cb->name  = ident_new(systf_data_p->tfname);
 
-   vpiHandle handle = handle_for(&cb->object);
+   vpiHandle handle = internal_handle_for(&cb->refcounted.object);
    APUSH(vpi_context()->systasks, handle);
 
    return handle;
@@ -726,15 +759,20 @@ PLI_INT32 vpi_release_handle(vpiHandle object)
 
    VPI_TRACE("handle=%s", handle_pp(object));
 
-   c_vpiObject *obj = from_handle(object);
-   if (obj == NULL) {
+   vpi_context_t *c = vpi_context();
+   handle_slot_t *slot = decode_handle(c, object);
+   if (slot == NULL) {
       vpi_error(vpiError, NULL, "invalid handle %p", object);
       return 0;
    }
+   else if (slot->kind == HANDLE_INTERNAL) {
+      vpi_error(vpiError, &(slot->obj->loc), "cannot release this handle as "
+                "it is owned by the system");
+      return 0;
+   }
 
-   drop_handle(vpi_context(), object);
+   drop_handle(c, object);
    return 1;
-
 }
 
 DLLEXPORT
@@ -751,7 +789,7 @@ vpiHandle vpi_handle(PLI_INT32 type, vpiHandle refHandle)
       switch (type) {
       case vpiSysTfCall:
          if (c->call != NULL)
-            return handle_for(&c->call->tfcall.object);
+            return user_handle_for(&c->call->tfcall.object);
          else
             return NULL;
       }
@@ -779,7 +817,7 @@ vpiHandle vpi_iterate(PLI_INT32 type, vpiHandle refHandle)
       return NULL;
    }
 
-   return handle_for(&(it->object));
+   return user_handle_for(&(it->refcounted.object));
 }
 
 DLLEXPORT
@@ -798,7 +836,7 @@ vpiHandle vpi_scan(vpiHandle iterator)
       return NULL;
 
    if (it->pos < it->list->count)
-      return handle_for(it->list->items[it->pos++]);
+      return user_handle_for(it->list->items[it->pos++]);
 
    return NULL;
 }
@@ -837,7 +875,7 @@ PLI_INT32 vpi_get(PLI_INT32 property, vpiHandle object)
          {
             vpi_context_t *c = vpi_context();
             if (c->args != NULL)
-               return ffi_array_length(c->args[op->argslot + 2].integer);
+               return c->args[op->argslot].integer;
 
             goto missing_property;
          }
@@ -880,10 +918,14 @@ void vpi_get_value(vpiHandle handle, p_vpi_value value_p)
    if (con != NULL) {
       switch (con->subtype) {
       case vpiStringConst:
-         tb_cat(c->valuestr, vlog_text(con->expr.where));
+         {
+            number_t n = vlog_number(con->expr.where);
+            for (int i = con->size/8 - 1; i >= 0; i--)
+               tb_append(c->valuestr, number_byte(n, i));
 
-         value_p->format = vpiStringVal;
-         value_p->value.str = (PLI_BYTE8 *)tb_get(c->valuestr);
+            value_p->format = vpiStringVal;
+            value_p->value.str = (PLI_BYTE8 *)tb_get(c->valuestr);
+         }
          return;
 
       case vpiBinaryConst:
@@ -899,12 +941,13 @@ void vpi_get_value(vpiHandle handle, p_vpi_value value_p)
 
    c_operation *op = is_operation(obj);
    if (op != NULL && c->args != NULL) {
-      void *ptr = c->args[op->argslot].pointer;
-      const int nbits = ffi_array_length(c->args[op->argslot + 2].integer);
-      number_t num = number_pack(ptr, nbits);
+      int size = c->args[op->argslot].integer;
+      assert(size <= 64);
 
-      vpi_format_number(num, value_p->format, c->valuestr);
-      number_free(&num);
+      uint64_t abits = c->args[op->argslot + 1].integer;
+      uint64_t bbits = c->args[op->argslot + 2].integer;
+
+      vpi_format_number2(size, abits, bbits, value_p->format, c->valuestr);
 
       value_p->value.str = (PLI_BYTE8 *)tb_get(c->valuestr);
       return;
@@ -917,7 +960,6 @@ void vpi_get_value(vpiHandle handle, p_vpi_value value_p)
       number_t num = number_pack(signal_value(s), signal_width(s));
 
       vpi_format_number(num, value_p->format, c->valuestr);
-      number_free(&num);
 
       value_p->value.str = (PLI_BYTE8 *)tb_get(c->valuestr);
       return;
@@ -945,18 +987,9 @@ vpiHandle vpi_put_value(vpiHandle handle, p_vpi_value value_p,
       assert(fcall->systfcall.callback->systf.sysfunctype == vpiTimeFunc);
       assert(value_p->format == vpiTimeVal);
 
-      uint8_t *bits = tlab_alloc(c->tlab, 64);
-
-      c->args[0].pointer = bits;
-      c->args[1].integer = 1;
-      c->args[2].integer = 64;
-
       const p_vpi_time tm = value_p->value.time;
-      PLI_UINT32 mask = 1;
-      for (int i = 0; i < 32; i++, mask <<= 1) {
-         bits[63 - i] = tm->low & mask ? LOGIC_1 : LOGIC_0;
-         bits[31 - i] = tm->high & mask ? LOGIC_1 : LOGIC_0;
-      }
+      c->args[0].integer = (uint64_t)tm->high << 32 | tm->low;
+      c->args[1].integer = 0;
 
       return NULL;
    }
@@ -992,27 +1025,50 @@ void vpi_context_initialise(vpi_context_t *c, tree_t top, rt_model_t *model,
    c->jit   = jit;
 }
 
+static void vpi_handles_diag(vpi_context_t *c, diag_t *d, handle_kind_t kind)
+{
+   for (int i = 0; i < c->num_handles; i++) {
+      handle_slot_t *slot = &(c->handles[i]);
+      if (slot->obj == NULL || slot->kind != kind)
+         continue;
+
+      diag_printf(d, "\n%s", handle_pp(encode_handle(slot, i)));
+
+      c_refcounted *rc = is_refcounted(slot->obj);
+      if (rc != NULL)
+         diag_printf(d, " with %d reference%s", rc->refcount,
+                     rc->refcount != 1 ? "s" : "");
+   }
+}
+
 static void vpi_check_leaks(vpi_context_t *c)
 {
-   int nactive = 0;
+   int nuser = 0, ninternal UNUSED = 0;
    for (int i = 0; i < c->num_handles; i++) {
-      if (c->handles[i].obj != NULL)
-         nactive++;
+      if (c->handles[i].obj == NULL)
+         continue;
+      else if (c->handles[i].kind == HANDLE_INTERNAL)
+         ninternal++;
+      else
+         nuser++;
    }
 
-   if (nactive > 0) {
+   if (nuser > 0) {
       diag_t *d = diag_new(DIAG_WARN, NULL);
-      diag_printf(d, "VPI program exited with %d active handles", nactive);
-
-      for (int i = 0; i < c->num_handles; i++) {
-         handle_slot_t *slot = &(c->handles[i]);
-         if (slot->obj != NULL)
-            diag_printf(d, "\n%s with %d references",
-                        handle_pp(slot->obj->handle), slot->refcount);
-      }
-
+      diag_printf(d, "VPI program exited with %d active handles", nuser);
+      vpi_handles_diag(c, d, HANDLE_USER);
       diag_emit(d);
    }
+
+#ifdef DEBUG
+   if (ninternal > 0) {
+      diag_t *d = diag_new(DIAG_DEBUG, NULL);
+      diag_printf(d, "VPI program exited with %d active internal handles",
+                  ninternal);
+      vpi_handles_diag(c, d, HANDLE_INTERNAL);
+      diag_emit(d);
+   }
+#endif
 }
 
 void vpi_context_free(vpi_context_t *c)
@@ -1092,7 +1148,7 @@ vpiHandle vpi_bind_foreign(ident_t name, vlog_node_t where)
       else
          call = &(build_sysFuncCall(where, cb, &(mod->scope))->systfcall);
 
-      return (call->handle = handle_for(&(call->tfcall.object)));
+      return (call->handle = internal_handle_for(&(call->tfcall.object)));
    }
 
    return NULL;
