@@ -504,30 +504,33 @@ START_TEST(test_vlog1)
              "  wire logic [7:0] bus;\n"
              "  wire logic w;\n"
              "  logic r;\n"
+             "  integer i;\n"
              "  parameter logic [7:0] p1 = 5;\n"
              "  localparam bit p2 = 0;\n"
              "  logic [7:0] array1 [127:0];\n"
              "  initial begin\n"
              "    $display(\"hello\", $time);\n"
-             "    if (bus)\n"
-             "      r <= 1 | r;\n"
+             "    if (bus) r <= 1 | r;\n"
              "    $finish;\n"
              "    r = 1;\n"
              "    #1 r <= 0;\n"
              "    r = ~w;\n"
              "    r = #1 5;\n"
              "    r <= #5 1;\n"
-             "    do\n"
-             "      w = r && w;\n"
-             "    while (r);\n"
+             "    do w = r && w; while (r);\n"
              "    while (1);\n"
-             "    repeat (5)\n"
-             "      r = r + 1;\n"
+             "    repeat (5) r = r + 1;\n"
+             "    for (i = 0; i < 10; i++) r = !r;\n"
+             "    forever i = ++i;\n"
              "  end\n"
              "  assign bus = 3;\n"
-             "  pullup (supply0,supply1) p1 (w);\n"
-             "  mod u1 (w);\n"
+             "  pullup (supply0,supply1) g1 (w);\n"
+             "  mod #(42) u1 (w), u2 (bus[1]);\n"
              "  assign bus[2] = 4'd1;\n"
+             "  initial casex (w)\n"
+             "    1'b0, 1'b1: r = p1[3:0];\n"
+             "    default: r = 0;\n"
+             "  endcase\n"
              "endmodule // mod2\n\n");
    tb_rewind(tb);
 
@@ -561,11 +564,12 @@ START_TEST(test_vlog1)
              "  wire logic z;\n"
              "  real v1 = 1.5;\n"
              "  integer v2 = x ? 1 : 2;\n"
+             "  wire logic zz;\n"
              "  initial begin\n"
-             "    if (x || (y === z));\n"
-             "    if ((x & y) | (y & (z === (x + z))));\n"
-             "    if ((x == y) || (y == z));\n"
-             "    if ((x & y) == z);\n"
+             "    if (x || (y === z)) ;\n"
+             "    if ((x & y) | (y & (z === (x + z)))) ;\n"
+             "    if ((x == y) || (y == z)) ;\n"
+             "    if ((x & y) == z) ;\n"
              "  end\n"
              "  assign z = {x, y};\n"
              "  assign zz = {5{z}};\n"
