@@ -462,9 +462,9 @@ START_TEST(test_gvn1)
    static const mir_match_t bb0[] = {
       { MIR_OP_ADD, PARAM("p1"), PARAM("p2") },
       { MIR_OP_SUB, NODE(0), PARAM("p2") },
-      { MIR_OP_MUL, PARAM("p1"), NODE(1) },
+      { MIR_OP_MUL, PARAM("p1"), PARAM("p1") },
       { MIR_OP_CMP, ENUM(MIR_CMP_EQ), NODE(2), CONST(10) },
-      { MIR_OP_CMP, ENUM(MIR_CMP_EQ), NODE(1), CONST(10) },
+      { MIR_OP_CMP, ENUM(MIR_CMP_EQ), PARAM("p1"), CONST(10) },
       { MIR_OP_AND, NODE(3), NODE(4) },
       { MIR_OP_OR, NODE(4), NODE(5) },
       { MIR_OP_NOT, NODE(6) },
@@ -491,11 +491,16 @@ START_TEST(test_dce1)
                                  MIR_UNIT_FUNCTION, NULL);
 
    mir_type_t t_int32 = mir_int_type(mu, INT32_MIN, INT32_MAX);
+   mir_type_t t_ptr = mir_pointer_type(mu, t_int32);
+
    mir_value_t p1 = mir_add_param(mu, t_int32, MIR_NULL_STAMP, ident_new("p1"));
    mir_value_t p2 = mir_add_param(mu, t_int32, MIR_NULL_STAMP, ident_new("p2"));
+   mir_value_t p3 = mir_add_param(mu, t_ptr, MIR_NULL_STAMP, ident_new("p3"));
 
    mir_set_result(mu, t_int32);
 
+   mir_build_store(mu, p3, p2);
+
    mir_value_t add1 = mir_build_add(mu, t_int32, p1, p2);
    mir_build_add(mu, t_int32, add1, p2);
    mir_build_return(mu, p1);
@@ -503,6 +508,7 @@ START_TEST(test_dce1)
    mir_optimise(mu, MIR_PASS_DCE);
 
    static const mir_match_t bb0[] = {
+      { MIR_OP_STORE, PARAM("p3"), PARAM("p2") },
       { MIR_OP_RETURN, PARAM("p1") },
    };
    mir_match(mu, 0, bb0);
@@ -596,15 +602,6 @@ START_TEST(test_logical1)
    mir_value_t and4 = mir_build_and(mu, p1, t);
    ck_assert_int_eq(and4.bits, p1.bits);
 
-   mir_value_t nand1 = mir_build_nand(mu, t, f);
-   mir_assert_const_eq(mu, nand1, 1);
-
-   mir_value_t nor1 = mir_build_nor(mu, t, f);
-   mir_assert_const_eq(mu, nor1, 0);
-
-   mir_value_t xnor1 = mir_build_nor(mu, t, f);
-   mir_assert_const_eq(mu, xnor1, 0);
-
    mir_unit_free(mu);
    mir_context_free(mc);
 }
@@ -1085,6 +1082,76 @@ START_TEST(test_cast1)
 }
 END_TEST
 
+START_TEST(test_vec1)
+{
+   mir_unit_t *mu = mir_unit_new(get_mir(), ident_new("vec1"), NULL,
+                                 MIR_UNIT_FUNCTION, NULL);
+
+   mir_type_t t_vec2_8 = mir_vec2_type(mu, 8, false);
+   mir_type_t t_vec2_9 = mir_vec2_type(mu, 9, false);
+
+   mir_set_result(mu, t_vec2_9);
+
+   mir_value_t p1 = mir_add_param(mu, t_vec2_8, MIR_NULL_STAMP,
+                                  ident_new("p1"));
+
+   mir_value_t c1 = mir_const_vec(mu, t_vec2_8, 5, 0);
+   mir_value_t add1 = mir_build_add(mu, t_vec2_9, p1, c1);
+   mir_value_t add2 = mir_build_add(mu, t_vec2_9, add1, c1);
+
+   mir_type_t t_carray9 = mir_carray_type(mu, 9, mir_int_type(mu, 0, 3));
+   mir_value_t tmp1 = mir_add_var(mu, t_carray9, MIR_NULL_STAMP,
+                                  ident_new("tmp1"), MIR_VAR_TEMP);
+
+   mir_value_t unpack1 = mir_build_unpack(mu, add2, 0, tmp1);
+   mir_value_t pack1 = mir_build_pack(mu, t_vec2_9, unpack1);
+
+   mir_build_return(mu, pack1);
+
+   mir_unit_free(mu);
+}
+END_TEST
+
+START_TEST(test_vec2)
+{
+   mir_unit_t *mu = mir_unit_new(get_mir(), ident_new("vec2"), NULL,
+                                 MIR_UNIT_FUNCTION, NULL);
+
+   mir_type_t t_vec2_8 = mir_vec2_type(mu, 8, false);
+   mir_type_t t_vec2_1 = mir_vec2_type(mu, 1, false);
+   mir_type_t t_offset = mir_offset_type(mu);
+
+   mir_value_t const1 = mir_const_vec(mu, t_vec2_8, 1, 0);
+   mir_value_t cast1 = mir_build_cast(mu, t_vec2_1, const1);
+   ck_assert_int_eq(mir_get_op(mu, cast1), MIR_OP_CONST_VEC);
+
+   mir_value_t cast2 = mir_build_cast(mu, t_offset, const1);
+   mir_assert_const_eq(mu, cast2, 1);
+
+   mir_unit_free(mu);
+}
+END_TEST
+
+START_TEST(test_check1)
+{
+   mir_unit_t *mu = mir_unit_new(get_mir(), ident_new("check1"), NULL,
+                                 MIR_UNIT_FUNCTION, NULL);
+
+   mir_type_t t_bool = mir_bool_type(mu);
+   mir_value_t p1 = mir_add_param(mu, t_bool, MIR_NULL_STAMP, ident_new("p1"));
+   mir_value_t locus = mir_build_locus(mu, NULL);
+
+   mir_build_dir_check(mu, p1, mir_const(mu, t_bool, RANGE_TO), locus);
+   mir_build_dir_check(mu, p1, p1, locus);
+
+   static const mir_match_t bb0[] = {
+      { MIR_OP_LOCUS },
+      { MIR_OP_DIR_CHECK, PARAM("p1"), CONST(0) },
+   };
+   mir_match(mu, 0, bb0);
+}
+END_TEST
+
 Suite *get_mir_tests(void)
 {
    Suite *s = suite_create("mir");
@@ -1109,6 +1176,9 @@ Suite *get_mir_tests(void)
    tcase_add_test(tc, test_record1);
    tcase_add_test(tc, test_alias1);
    tcase_add_test(tc, test_cast1);
+   tcase_add_test(tc, test_vec1);
+   tcase_add_test(tc, test_vec2);
+   tcase_add_test(tc, test_check1);
    suite_add_tcase(s, tc);
 
    return s;
