@@ -1,5 +1,5 @@
 //
-//  Copyright (C) 2022-2024 Nick Gasson
+//  Copyright (C) 2022-2025 Nick Gasson
 //
 //  This program is free software: you can redistribute it and/or modify
 //  it under the terms of the GNU General Public License as published by
@@ -27,6 +27,8 @@
 #include "vlog/vlog-number.h"
 #include "vlog/vlog-phase.h"
 
+#include <inttypes.h>
+
 START_TEST(test_dff)
 {
    input_from_file(TESTDIR "/vlog/dff.v");
@@ -137,6 +139,7 @@ START_TEST(test_ports)
 {
    const error_t expect[] = {
       {  4, "duplicate declaration of y" },
+      { 11, "no visible declaration for y" },
       { 19, "duplicate declaration of x" },
       { 22, "duplicate declaration of y" },
       { 31, "'o3' cannot be assigned in a procedural block" },
@@ -148,7 +151,7 @@ START_TEST(test_ports)
 
    input_from_file(TESTDIR "/vlog/ports.v");
 
-   for (int i = 0; i < 7; i++) {
+   for (int i = 0; i < 8; i++) {
       vlog_node_t m = vlog_parse();
       fail_if(m == NULL);
       fail_unless(vlog_kind(m) == V_MODULE);
@@ -169,9 +172,9 @@ START_TEST(test_parse1)
    vlog_node_t m = vlog_parse();
    fail_if(m == NULL);
    fail_unless(vlog_kind(m) == V_MODULE);
-   ck_assert_int_eq(vlog_stmts(m), 19);
+   ck_assert_int_eq(vlog_stmts(m), 33);
    ck_assert_int_eq(vlog_ports(m), 0);
-   ck_assert_int_eq(vlog_decls(m), 19);
+   ck_assert_int_eq(vlog_decls(m), 28);
 
    vlog_node_t x = vlog_decl(m, 0);
    fail_unless(vlog_kind(x) == V_NET_DECL);
@@ -192,7 +195,7 @@ START_TEST(test_parse1)
    fail_unless(vlog_stmts(s1) == 1);
 
    vlog_node_t s1s0 = vlog_stmt(s1, 0);
-   fail_unless(vlog_kind(s1s0) == V_SEQ_BLOCK);
+   fail_unless(vlog_kind(s1s0) == V_BLOCK);
    fail_unless(vlog_ident(s1s0) == ident_new("foo"));
 
    vlog_node_t s1s0s0 = vlog_stmt(s1s0, 0);
@@ -202,7 +205,13 @@ START_TEST(test_parse1)
 
    vlog_node_t hello = vlog_param(s1s0s0, 0);
    fail_unless(vlog_kind(hello) == V_STRING);
-   ck_assert_str_eq(vlog_text(hello), "hello");
+
+   number_t n1 = vlog_number(hello);
+   ck_assert_int_eq(number_byte(n1, 0), 'o');
+   ck_assert_int_eq(number_byte(n1, 1), 'l');
+   ck_assert_int_eq(number_byte(n1, 2), 'l');
+   ck_assert_int_eq(number_byte(n1, 3), 'e');
+   ck_assert_int_eq(number_byte(n1, 4), 'h');
 
    vlog_node_t s1s0s1 = vlog_stmt(s1s0, 1);
    fail_unless(vlog_kind(s1s0s1) == V_SYS_TCALL);
@@ -262,7 +271,7 @@ START_TEST(test_parse1)
    fail_unless(vlog_kind(s7) == V_GATE_INST);
    ck_assert_int_eq(vlog_subkind(s7), V_GATE_PULLUP);
    ck_assert_int_eq(vlog_params(s7), 1);
-   fail_if(vlog_has_ident(s7));
+   fail_unless(vlog_has_ident(s7));
    fail_unless(vlog_kind(vlog_target(s7)) == V_REF);
 
    vlog_node_t s8 = vlog_stmt(m, 8);
@@ -275,7 +284,7 @@ START_TEST(test_parse1)
    vlog_node_t s9 = vlog_stmt(m, 9);
    fail_unless(vlog_kind(s9) == V_INITIAL);
    vlog_node_t s9b = vlog_stmt(s9, 0);
-   fail_unless(vlog_kind(s9b) == V_SEQ_BLOCK);
+   fail_unless(vlog_kind(s9b) == V_BLOCK);
    fail_unless(vlog_kind(vlog_stmt(s9b, 0)) == V_WHILE);
    fail_unless(vlog_kind(vlog_stmt(s9b, 1)) == V_REPEAT);
    fail_unless(vlog_kind(vlog_stmt(s9b, 2)) == V_DO_WHILE);
@@ -288,6 +297,11 @@ START_TEST(test_parse1)
    ck_assert_int_eq(vlog_ranges(d17), 1);
    fail_unless(vlog_subkind(vlog_range(d17, 0)) == V_DIM_UNPACKED);
 
+   vlog_node_t d20 = vlog_decl(m, 20);
+   fail_unless(vlog_kind(d20) == V_VAR_DECL);
+   vlog_node_t d20t = vlog_type(d20);
+   fail_unless(vlog_subkind(d20t) == DT_EVENT);
+
    vlog_node_t s10 = vlog_stmt(m, 16);
    fail_unless(vlog_kind(s10) == V_ALWAYS);
    fail_unless(vlog_subkind(s10) == V_ALWAYS_COMB);
@@ -300,6 +314,12 @@ START_TEST(test_parse1)
    fail_unless(vlog_kind(s12) == V_ALWAYS);
    fail_unless(vlog_subkind(s12) == V_ALWAYS_LATCH);
 
+   vlog_node_t s19 = vlog_stmt(m, 19);
+   fail_unless(vlog_kind(s19) == V_ALWAYS);
+   vlog_node_t s19b = vlog_stmt(s19, 0);
+   vlog_node_t et = vlog_stmt(s19b, 0);
+   fail_unless(vlog_kind(et) == V_EVENT_TRIGGER);
+
    fail_unless(vlog_parse() == NULL);
 
    fail_if_errors();
@@ -308,31 +328,37 @@ END_TEST
 
 START_TEST(test_number1)
 {
-   number_t x = number_new("1'b1");
-   ck_assert_int_eq(x.common.tag, TAG_INTEGER);
+   number_t x = number_new("1'b1", NULL);
+   ck_assert_int_eq(x.common.tag, TAG_BIGNUM);
    fail_unless(number_is_defined(x));
+   fail_if(number_signed(x));
    ck_assert_int_eq(number_width(x), 1);
    ck_assert_int_eq(number_integer(x), 1);
-   number_free(&x);
 
-   number_t y = number_new("5'b101");
-   ck_assert_int_eq(y.common.tag, TAG_INTEGER);
+   number_t y = number_new("5'b101", NULL);
+   ck_assert_int_eq(y.common.tag, TAG_BIGNUM);
    fail_unless(number_is_defined(y));
+   fail_if(number_signed(y));
    ck_assert_int_eq(number_width(y), 5);
    ck_assert_int_eq(number_integer(y), 5);
-   number_free(&y);
 
-   number_t z = number_new("5'b1xx");
-   ck_assert_int_eq(z.common.tag, TAG_SMALLNUM);
+   number_t z = number_new("5'b1xx", NULL);
+   ck_assert_int_eq(z.common.tag, TAG_BIGNUM);
    fail_if(number_is_defined(z));
+   fail_if(number_signed(z));
    ck_assert_int_eq(number_width(z), 5);
-   number_free(&z);
 
-   number_t n1 = number_new("1'bx");
-   ck_assert_int_eq(n1.common.tag, TAG_SMALLNUM);
+   number_t n1 = number_new("1'bx", NULL);
+   ck_assert_int_eq(n1.common.tag, TAG_BIGNUM);
    fail_if(number_is_defined(n1));
+   fail_if(number_signed(n1));
    ck_assert_int_eq(number_width(n1), 1);
-   number_free(&n1);
+
+   number_t n2 = number_new("456", NULL);
+   ck_assert_int_eq(n2.common.tag, TAG_BIGNUM);
+   fail_unless(number_is_defined(n2));
+   fail_unless(number_signed(n2));
+   ck_assert_int_eq(number_width(n2), 32);
 }
 END_TEST
 
@@ -340,30 +366,38 @@ START_TEST(test_number2)
 {
    struct {
       const char *input;
-      unsigned    tag;
       unsigned    width;
       int64_t     ival;
       const char *string;
    } cases[] = {
-      { "1'b1",          TAG_INTEGER, 1,   1,           "1'b1"       },
-      { "5'b100",        TAG_INTEGER, 5,   4,           "5'd4"       },
-      { "1",             TAG_INTEGER, 32,  1,           "1"          },
-      { "42",            TAG_INTEGER, 32,  42,          "42"         },
-      { "251251",        TAG_INTEGER, 32,  251251,      "251251"     },
-      { "'hffffffff",    TAG_INTEGER, 32,  0xffffffff,  "4294967295" },
-      { "33'h100000000", TAG_BIGNUM,  33,  0x100000000,
-        "33'b100000000000000000000000000000000" },
-      { "64'b0",         TAG_BIGNUM,  64,  0,           "64'b0"      },
-      { "64'b101",       TAG_BIGNUM,  64,  5,           "64'b101"    },
-      { "8'b00_01",      TAG_INTEGER, 8,   1,           "8'd1"       },
-      { "128'bx",        TAG_BIGNUM,  128, 0,           "128'bx"     },
+      { "1'b1",          1,   1,           "1'b1"           },
+      { "5'b100",        5,   4,           "5'd4"           },
+      { "1",             32,  1,           "1"              },
+      { "42",            32,  42,          "42"             },
+      { "251251",        32,  251251,      "251251"         },
+      { "'hffffFFff",    32,  0xffffffff,  "32'd4294967295" },
+      { "33'h100000000", 33,  0x100000000,
+        "33'b100000000000000000000000000000000"             },
+      { "64'b0",         64,  0,           "64'b0"          },
+      { "64'b101",       64,  5,           "64'b101"        },
+      { "8'b00_01",      8,   1,           "8'd1"           },
+      { "128'bx",        128, 0,           "128'bx"         },
+      { "4'h12",         4,   2,           "4'd2"           },
+      { "1'b110",        1,   0,           "1'b0"           },
+      { "1_000",         32,  1000,        "1000"           },
    };
 
+   const error_t expect[] = {
+      { LINE_INVALID, "excess digits in hex constant 4'h12" },
+      { LINE_INVALID, "excess digits in binary constant 1'b110" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
    LOCAL_TEXT_BUF tb = tb_new();
 
    for (int i = 0; i < ARRAY_LEN(cases); i++) {
-      number_t n = number_new(cases[i].input);
-      ck_assert_int_eq(n.common.tag, cases[i].tag);
+      number_t n = number_new(cases[i].input, NULL);
       ck_assert_int_eq(number_width(n), cases[i].width);
 
       if (cases[i].width <= 64)
@@ -372,9 +406,9 @@ START_TEST(test_number2)
       tb_rewind(tb);
       number_print(n, tb);
       ck_assert_str_eq(tb_get(tb), cases[i].string);
-
-      number_free(&n);
    }
+
+   check_expected_errors();
 }
 END_TEST
 
@@ -442,28 +476,6 @@ START_TEST(test_timescale1)
 }
 END_TEST
 
-START_TEST(test_defaultnettype)
-{
-   input_from_file(TESTDIR "/vlog/defaultnettype.v");
-
-   const error_t expect[] = {
-      { 12, "unexpected identifier while parsing default_nettype directive, expecting one of wire, tri, tri0, tri1, wand, triand, wor, trior," },
-      { -1, NULL }
-   };
-   expect_errors(expect);
-
-   vlog_node_t m = vlog_parse();
-   fail_if(m == NULL);
-   fail_unless(vlog_kind(m) == V_MODULE);
-
-   vlog_check(m);
-
-   fail_unless(vlog_parse() == NULL);
-
-   check_expected_errors();
-}
-END_TEST
-
 START_TEST(test_gate1)
 {
    input_from_file(TESTDIR "/vlog/gate1.v");
@@ -529,11 +541,10 @@ START_TEST(test_udp1)
    input_from_file(TESTDIR "/vlog/udp1.v");
 
    const error_t expect[] = {
+      { 13, "no visible declaration for z" },
       { 13, "the first port of a primitive must be an output" },
       { 13, "all ports of a primitive except the first must be inputs" },
-      { 13, "no visible declaration for z" },
-      { 25, "missing symbol for input y" },
-      { 26, "too many symbols in UDP table entry" },
+      { 25, "expected 3 symbols in UDP table entry but have 2" },
       { 58, "a sequential input list may have at most one edge indicator" },
       { -1, NULL }
    };
@@ -599,17 +610,23 @@ START_TEST(test_enum1)
 {
    input_from_file(TESTDIR "/vlog/enum1.v");
 
+   const error_t expect[] = {
+      {  5, "'a' is not a data type" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
    vlog_node_t m = vlog_parse();
    fail_if(m == NULL);
    fail_unless(vlog_kind(m) == V_MODULE);
 
    vlog_check(m);
 
-   ck_assert_int_eq(vlog_decls(m), 2);
+   ck_assert_int_eq(vlog_decls(m), 4);
 
    fail_unless(vlog_parse() == NULL);
 
-   fail_if_errors();
+   check_expected_errors();
 }
 END_TEST
 
@@ -653,6 +670,38 @@ START_TEST(test_param1)
 }
 END_TEST
 
+START_TEST(test_param2)
+{
+   input_from_file(TESTDIR "/vlog/param2.v");
+
+   const error_t expect[] = {
+      { 13, "duplicate declaration" },
+      { 18, "local parameter declaration must have a default value" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m1 = vlog_parse();
+   fail_if(m1 == NULL);
+   fail_unless(vlog_kind(m1) == V_MODULE);
+
+   vlog_check(m1);
+
+   vlog_node_t m2 = vlog_parse();
+   fail_if(m2 == NULL);
+   fail_unless(vlog_kind(m2) == V_MODULE);
+
+   vlog_node_t p2 = vlog_decl(m2, 1);
+   fail_unless(vlog_kind(p2) == V_LOCALPARAM);
+
+   vlog_check(m2);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
 START_TEST(test_pp3)
 {
    input_from_file(TESTDIR "/vlog/pp3.v");
@@ -778,6 +827,209 @@ START_TEST(test_const1)
 }
 END_TEST
 
+START_TEST(test_case1)
+{
+   input_from_file(TESTDIR "/vlog/case1.v");
+
+   const error_t expect[] = {
+      { 10, "no visible declaration for aaa" },
+      { 15, "multiple default statements within a single case statement" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m = vlog_parse();
+   fail_if(m == NULL);
+   fail_unless(vlog_kind(m) == V_MODULE);
+
+   vlog_check(m);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_direct1)
+{
+   input_from_file(TESTDIR "/vlog/direct1.v");
+
+   const error_t expect[] = {
+      { 11, "unexpected identifier while parsing default_nettype directive, "
+        "expecting one of wire, tri, tri0, tri1, wand, triand, wor, trior," },
+      { 15, "no visible declaration for x" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m1 = vlog_parse();
+   fail_if(m1 == NULL);
+   fail_unless(vlog_kind(m1) == V_MODULE);
+
+   vlog_check(m1);
+
+   vlog_node_t m2 = vlog_parse();
+   fail_if(m2 == NULL);
+   fail_unless(vlog_kind(m2) == V_MODULE);
+
+   vlog_check(m2);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_string1)
+{
+   static const struct {
+      const char *input;
+      const char *output;
+   } cases[] = {
+      { "\"hello\"", "hello" },
+      { "\"\"", "" },
+   };
+
+   for (int i = 0; i < ARRAY_LEN(cases); i++) {
+      number_t n = number_new(cases[i].input, NULL);
+      ck_assert(number_is_defined(n));
+
+      const int width = number_width(n);
+      ck_assert_int_eq(width, strlen(cases[i].output) * 8);
+
+      for (int j = 0; j < width / 8; j++)
+         ck_assert_msg(number_byte(n, width/8 - 1 - j) == cases[i].output[j],
+                       "%s != %s", cases[i].input, cases[i].output);
+   }
+}
+END_TEST
+
+START_TEST(test_generate1)
+{
+   input_from_file(TESTDIR "/vlog/generate1.v");
+
+   const error_t expect[] = {
+      { 11, "cannot specify both a label and a name for the same block" },
+      { 12, "'g2' does not match label 'g1'" },
+      { 13, "block does not have a label" },
+      {  3, "cannot reference net 'x' in constant expression" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m1 = vlog_parse();
+   fail_if(m1 == NULL);
+   fail_unless(vlog_kind(m1) == V_MODULE);
+
+   vlog_check(m1);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_tfcall1)
+{
+   input_from_file(TESTDIR "/vlog/tfcall1.v");
+
+   const error_t expect[] = {
+      {  8, "expected 2 arguments for 'sum' but have 1" },
+      {  9, "expected 2 arguments for 'sum' but have 3" },
+      { 13, "'x4' is not a function" },
+      { 27, "'no_args' is not a task" },
+      { 12, "no visible declaration for not_here" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m1 = vlog_parse();
+   fail_if(m1 == NULL);
+   fail_unless(vlog_kind(m1) == V_MODULE);
+
+   vlog_check(m1);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_attr1)
+{
+   input_from_file(TESTDIR "/vlog/attr1.v");
+
+   const error_t expect[] = {
+      { 11, "attribute instance is not allowed here" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m1 = vlog_parse();
+   fail_if(m1 == NULL);
+   fail_unless(vlog_kind(m1) == V_MODULE);
+
+   vlog_check(m1);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_initial1)
+{
+   input_from_file(TESTDIR "/vlog/initial1.v");
+
+   const error_t expect[] = {
+      { 9, "unexpected reg while parsing statement item, expecting one of " },
+      { 14, "initial block does not have a label" },
+      { 16, "'another_label' does not match label 'a_label'" },
+      { 22, "unexpected reg while parsing statement item, expecting one of " },
+      { 28, "fork block does not have a label" },
+      { 30, "'fork5' does not match label 'fork4'" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m1 = vlog_parse();
+   fail_if(m1 == NULL);
+   fail_unless(vlog_kind(m1) == V_MODULE);
+
+   vlog_check(m1);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
+START_TEST(test_nets1)
+{
+   input_from_file(TESTDIR "/vlog/nets1.v");
+
+   const error_t expect[] = {
+      { 10, "charge strength only allowed with the trireg keyword" },
+      { 13, "vectored and scalared keywords are only allowed with at least "
+        "a packed dimension" },
+      { 27, "unexpected highz1 while parsing strength1, expecting one of "
+        "supply1, strong1, pull1 or weak1" },
+      { -1, NULL }
+   };
+   expect_errors(expect);
+
+   vlog_node_t m1 = vlog_parse();
+   fail_if(m1 == NULL);
+   fail_unless(vlog_kind(m1) == V_MODULE);
+
+   vlog_check(m1);
+
+   fail_unless(vlog_parse() == NULL);
+
+   check_expected_errors();
+}
+END_TEST
+
 Suite *get_vlog_tests(void)
 {
    Suite *s = suite_create("vlog");
@@ -792,7 +1044,6 @@ Suite *get_vlog_tests(void)
    tcase_add_test(tc, test_pp1);
    tcase_add_test(tc, test_empty1);
    tcase_add_test(tc, test_timescale1);
-   tcase_add_test(tc, test_defaultnettype);
    tcase_add_test(tc, test_gate1);
    tcase_add_test(tc, test_pp2);
    tcase_add_test(tc, test_specify1);
@@ -802,11 +1053,20 @@ Suite *get_vlog_tests(void)
    tcase_add_test(tc, test_enum1);
    tcase_add_test(tc, test_union1);
    tcase_add_test(tc, test_param1);
+   tcase_add_test(tc, test_param2);
    tcase_add_test(tc, test_pp3);
    tcase_add_test(tc, test_concat1);
    tcase_add_test(tc, test_pp4);
    tcase_add_test(tc, test_casename);
    tcase_add_test(tc, test_const1);
+   tcase_add_test(tc, test_case1);
+   tcase_add_test(tc, test_direct1);
+   tcase_add_test(tc, test_string1);
+   tcase_add_test(tc, test_generate1);
+   tcase_add_test(tc, test_tfcall1);
+   tcase_add_test(tc, test_attr1);
+   tcase_add_test(tc, test_initial1);
+   tcase_add_test(tc, test_nets1);
    suite_add_tcase(s, tc);
 
    return s;
